/* cgiargs.c
**
** read the CGI arguments into a vector of (TAG, VALUE)-pairs
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "tools.h"

#include "cgiargs.h"

static __inline__ int max (int a, int b)
{
    return (a < b ? b : a);
}

static __inline__ int min (int a, int b)
{
    return (a < b ? a : b);
}

static __inline__ int lcmp (char *l, char *r)
{
    return lccmp (l, r, max(strlen (l), strlen (r)));
}

static __inline__ int x2c (char *p)
{
    char *hd = "001122334455566778899AaBbCcDdEeFf";
    char *q;
    int res;
    if (!(q = strchr (hd, *p))) { return -256; }
    res = (q - hd) >> 1;
    if (!(q = strchr (hd, p[1]))) { return -256; }
    res = (res << 8) | ((q - hd) >> 1);
    if (res > 127) { res -= 256; }
    return res;
}

/* compare two CGI arguments by their tags */
static int cgic (cgi_t *l, cgi_t *r)
{
    int res = memcmp (l->key, r->key, min (l->kl, r->kl));
    if (res) return res;
    return l->kl - r->kl;
}

/* scan all CGI arguments - either from stdin (REQUEST_METHOD = POST) or from
** the value of an environment variable (`QUERY_STRING', REQUEST_METHOD = GET).
** return the arguments in a vector of (key, value)-pairs.
*/
cgi_t *CGIargs (void)
{
    char *rm = getenv ("REQUEST_METHOD");
    cgi_t *CGI = NULL, *cp;
    char *cargs, *p, *q;
    size_t nel = 0;
    int ix;
    if (!rm) { return CGI; }
    if (!lcmp (rm, "GET")) {
	if (!(cargs = getenv ("QUERY_STRING"))) { return CGI; }
	cargs = gdupl (__FILE__, __LINE__, cargs);
    } else {
	size_t cl;
	char *p;
	if (!(cargs = getenv ("CONTENT_LENGTH"))) { return CGI; }
	if (!(p = getenv ("CONTENT_TYPE"))
	|| strcmp (p, "application/x-www-form-urlencoded")) {
	    return CGI;
	}
	cl = strtol (cargs, &p, 10);
	if (cl <= 0) { return CGI; }
	cargs = galloc (__FILE__, __LINE__, cl + 1);
	cargs[cl] = '\0';
	if (fread (cargs, sizeof(char), cl, stdin) < cl) {
	    gfree (cargs);
	    return CGI;
	}
    }
    /* calculate the number of CGI arguments */
    nel = 1; p = cargs;
    while ((q = strchr (p, '&'))) { ++nel; p = q + 1; }

    /* allocate the space required for the CGI argument vector */
    CGI = galloc (__FILE__, __LINE__, (nel + 1) * sizeof (cgi_t));

    /* initialize the terminating element */
    CGI[nel].key = NULL; CGI[nel].kl = 0;

    /* convert and insert the CGI arguments into the CGI argument vector */
    p = cargs; q = p;
    ix = 0; cp = &CGI[ix];
    cp->key = p; cp->data = NULL;
    while (*q) {
	switch (*q) {
	    case '&':
		*p = '\0';
		if (cp->data) cp->dl = p - cp->data; else cp->kl = p - cp->key;
		++p; ++q;
		++ix; cp = &CGI[ix];
		cp->key = p; cp->data  = NULL;
		continue;
	    case '+': *p++ = ' '; ++q; break;
	    case '%': *p++ = x2c (++q); q += 2; break;
	    case '=':
		*p = '\0'; cp->kl = p - cp->key;
		cp->data = ++p; ++q;
		break;
	    default:  *p++ = *q++; break;
	}
    }
    *p = '\0';
    if (cp->data) cp->dl = p - cp->data; else cp->kl = p - cp->key;

    /* sort the CGI argument vector */
    qsort (CGI, nel, sizeof(cgi_t), (int (*)(const void *, const void *)) cgic);

    /* return the sorted CGI argument vector */
    return CGI;
}

/* find a CGI argument corresponding to the key (key, kl) in the 
** CGI argument vector; return a pointer to this argument if successful
** or (cgi_t *) 0 otherwise.
*/
cgi_t *CGIget (cgi_t *CGI, char *key, size_t kl)
{
    int l, h, m, cv;
    cgi_t kv;
    if (!CGI) return (cgi_t *) 0;
    kv.key = key; kv.kl = kl;
    l = 0; for (h = 0; CGI[h].key; ++h);
    while (l < h) {
	m = (l + h) / 2;
	cv = cgic (&kv, &CGI[m]);
	if (!cv) return &CGI[m];
	if (cv < 0) h = m; else l = m;
    }
    return (cgi_t *) 0;
}
