/* main.c
**
** main program of `vxadmin'
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "dbif.h"
#include "tools.h"
#include "prefs.h"
#include "conf.h"
#include "vxerror.h"
#include "vxopt.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] [tafile]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile>"
		     "\n       read the configuration from file <configfile>"
		     "\n       (default: %s)"
		     "\n   -h"
		     "\n       show this help"
		     "\n   -V"
		     "\n       show the program's version number and exit\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (1);
}

static cgi_t *CGI = (cgi_t *) 0;

int main (int argc, char *argv)
{
    int cgimode = 1;
    char *tafile, *rm;
    cgi_t *cgap;
    vxopt (&argc, argv, usage);
    if (argc > 1) { cgimode = 0; tafile = argv[1]; }
    if (cgimode) {
	if (!(CGI = CGIargs ())) { usage (); }
	/* Was mu ich an dieser Stelle tun?
	 * Ich mu zuerst mal den Zustand feststellen; dazu bentige ich ein
	 * `hidden field', welches mir den gerade gltigen Transaktionszustand
	 * mitteilt. Wenn dieser Wert leer ist, dann mu eine Authentifikation
	 * stattfinden; wenn die in diesem Wert gespeicherte Uhrzeit zu sehr
	 * von der aktuellen Zeit abweicht, dann mu eine Authentifikation
	 * stattfinden.
	 */
	if (!(cgap = CGIget (CGI, "ACTN", 4))) {
CGI_REDO:
	    /* Keine Aktion definiert ==> Authentifizierung
	     * Wenn die Authentifizierung den Benutzer `::MAILSYS::' ergibt,
	     * dann erzeuge eine noch (weitgehend) leere Transaktionsdatei;
	     * `::MAILSYS::' kann nmlich erst dann nderungen ausfhren, wenn
	     * er eine Domain ausgewhlt hat.
	     * Wenn die Authentifizierung einen Domainadministrator ergibt,
	     * dann trage alle Eintrge innerhalb dieser Domain in die Trans-
	     * aktionsdatei ein. Anschlieend sind ist diese Domain fr weitere
	     * Zugriffe durch `vxadmin' gesperrt, und zwar solange, bis die
	     * Transaktion entweder abgeschlossen oder verworfen wurde. Auer-
	     * dem gilt eine Transaktion als verworfen, wenn sie nicht
	     * innerhalb einer (konfigurierbaren) Zeitspanne seit der letzten
	     * Aktion abgeschlossen wurde; fr jede Aktion mu der Zeitpunkt
	     * innerhalb dessen sie stattgefunden hat, in der Transaktionsdatei
	     * gespeichert werden.
	     * Wenn die Authentifizierung einen normalen Benutzer ergeben hat
	     * oder eine Pawortnderung angestoen wird, dann mu die Zeit-
	     * spanne zwischen der letzten Aktion und dem Beginn der eigent-
	     * lichen Transaktion sehr kurz gehalten werden (ca. 1 min). In
	     * diesem Fall wird lediglich der entsprechende Benutzereintrag
	     * in der Transaktionsdatei vermerkt.
	     */
	    switch (CGIauthenticate ()) {
		case AUTH_MAILSYS:
		    /* Lade die Transaktionsdatei mit den ganzen spezialein-
		     * trgen aus der Datenbank sowie den einzelnen Domain-
		     * eintrgen.
		     */
		    TC_LoadFile (NULL, SPECIAL|DOMAIN);
		    ShowHTML (MAILSYS);
		case AUTH_DOMADM:
		    /* Lade die Transaktionsdatei mit den Benutzereintrgen
		     * innerhalb der Domain.
		     */
		    TC_LoadFile (Domainname, USERS);
		    ShowHTML (DOMADM);
		case AUTH_USER:
		    /* Lade die Transaktionsdatei mit dem Benutzereintrag des
		     * entsprechenden Benutzers
		     */
		    TC_LoadFile (Addr, USER);
		    ShowHTML (USER);
	    }
	} else {
	    /* Lies den letzten Stand aus der Transaktionsdatei - oder
	     * leite eine erneute Authentifizierung ein, falls die Datei
	     * nicht existiert oder die letzte Aktion zulange zurckliegt.
	     */
	    if (!CGI_init (cgap)) { goto CGI_REDO; }
	    switch (actiontag (cgap)) {
		case AC_CHGPASS:
		    if (new_passwd (CGI)) {
			ShowHTML (DELAYED);
		    } else {
			ShowHTML (DONE);
		    }
		case AC_ADDUSER:
		    add_user (CGI);
		    ShowHTML (DOMADM);
		case AC_DELUSER:
		    del_user (CGI);
		    ShowHTML (DOMADM);
		case AC_EDIT:
		    ShowHTML (EDITUSER);
		case AC_ENDEDIT:
		    add_user (CGI);
		    ShowHTML (DOMADM);
		case AC_VXTS:
		    add_vxts (CGI);
		    ShowHTML (MAILSYS);
		case AC_REDIRECT:
		    add_redirect (CGI);
		    ShowHTML (MAILSYS);
		case AC_IPTRANS:
		    add_iptrans (CGI);
		    ShowHTML (MAILSYS);
		case AC_MAILHOST:
		    add_mailhost (CGI);
	    }
    }
