/* tc.c
**
** transaction control system
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "tools.h"
#include "xfile.h"
#include "dbif.h"

/* Was mu ich tun?
**
** Ich bekomme eine Datei, die Transaktionskommandos enthlt; wenn ich 
** sofort Schreibzugriff zur Datenbank bekomme, dann fhre ich alle
** Transaktionskommandos sofort aus; wenn nein, dann spalte ich einen Proze
** ab, der in (mglichst unregelmigen) Zeitabstnden immer wieder versucht,
** einen Schreibzugriff zur Datenbank zu bekommen. Der ursprngliche Proze
** hingegen gibt einen Wert zurck, der Auskubft darber gibt, da die
** Transaktion verzgert ausgefhrt wird.
*/

int TcInit (char *filename, char *dbname)
{
    int rstat;
    DBIF_FILE udb = (DBIF_FILE) 0;
    FILE *tf = fopen (filename, "r");
    if (!tf) { return -1; }
    dbif_rw (&udb, dbname);
    if (!udb) {
	if (dbif_errno() == DBIF_NOWR) { return TcSpawn (tf, dbname); }
	fclose (tf);
	return -2;
    }
    rstat = TcExecute (tf, udb);
    dbif_close (udb);
    fclose (tf);
    unlink (filename);
    return rstat;
}

/* Wie mache ich das eigentlich mit den Transaktionen im Hintergrund?
** Ich werde einfach von folgendem ausgehen:
**    Ich schaue zunchst nach, ob ein Transaktionsproze bereits gestartet
**    wurde. Wenn ja, dann warte ich, bis dieser beendet ist, um selbst einen
**    Transaktionsproze zu starten; alle Transaktionen sollen in einer Datei
**    geloggt werden.
**    Es ist nur ein Zugriff pro Domain zulssig; erst wenn die Transaktion
**    fr diesen Zugriff abgeschlossen ist und die Daten wieder vollstndig in
**    der Datenbank stehen, sind weitere Zugriffe auf diese Domain mglich.
**    Die Zugriffe fr die einzelnen Domains mssen trotzdem sequentialisiert
**    werden; dazu kann ich eine einfache Sperrdatei verwenden, die eben
**    solange gesperrt ist, bis eine Transaktion abgeschlossen ist; anschlie-
**    end wird diese Datei wieder freigegeben; dadurch erhalte ich eine
**    quasi natrliche Sequentialisierung.
**
**    Also in Einzelschritten noch einmal:
**    - Gucke nach, ob eine Domaindatei bereits vorhanden ist.
**    - Wenn ja, dann
**      - gucke in dieser Datei nach, ob sie zu der aktuellen Session gehrt;
**      - Wenn nein, dann melde dem Benutzer, da er momentan keinen Zugang zu
**        dieser Domain hat;
**      - Wenn ja, dann lies den Inhalt der Datei vollstndig ein, fhre die
**        einzelnen Transaktionsschritte (quasi virtuell) aus;
**      - Zeige das Ergebnis als aktuellen Zustand dem Benutzer an.
**    - Wenn nein, dann
**      - erzeuge diese Datei neu,
**      - sperre die neu erzeugte Datei,
**      - schreibe die Session-ID hinein,
**      - hole den aktuellen Zustand aus der Datenbank und bertrage ihn
**        in die Datei.
**      - gib die Datei anschlieend wieder frei und schliee sie;
**      - gib dem Benutzer den aktuellen Datenbankzustand.
**    - Wenn ein Benutzer eine nderung ausfhrt (z.B. ein ADD oder DEL),
**      dann hnge diese Aktion als Transaktionskommando an die Datei an.
**    - Wenn der Benutzer die Transaktion verwirft, dann entleere diese Datei
**    - Wenn der Benutzer die Transaktion abschliet, dann schaue nach, ob
**      ein Schreibzugriff auf die Datenbank mglich ist:
**      - Belege die Sperrdatei und gucke nach, ob Du einen Schreibzugriff
**        auf die Datenbank erhlst;
**      - Wenn ja, dann schreibe die Daten in die Datenbank, lsche
**        anschlieend die Transaktionsdatei und gib die Sperrdatei wieder frei;
**        melde anschlieend dem Benutzer den Vollzug seiner Transaktion.
**      - Wenn nein, dann spalte einen Subproze ab, der die Dateisperre
**        sozusagen erben mu; dann melde dem Benutzer einen verzgerten
**        Vollzug; der Subproze mu nun in unregelmigen Abstnden versuchen
**        einen Schreibzugriff auf die Datenbank zu erlangen; hat er diesen
**        Zugriff erhalten, dann mu er die Transaktion ausfhren; anschlieend
**        lscht er die Transaktionsdatei (oder entleert sie vielmehr) und
**        gibt die Sperrdatei wieder frei.
*/

