/* vxconv.c
**
** Hauptprogramm zu `vxconv'.
**
** Aufgabe dieses Programmes ist es, vorhandene Standardfolder der SMTP-Server
** `smail', `sendmail', ... in das VxMail-Format zu konvertieren. Dazu mu
** lediglich eine Indexdatei zu dem jeweiligen Folder erzeugt werden, die die
** Anfangspositionen der einzelnen E-Mails innerhalb des zu konvertierenden
** Folders enthlt.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <getopt.h>

#include "config.h"
#include "prefs.h"
#include "conf.h"
#include "tools.h"
#include "vxerror.h"
#include "xfile.h"
#include "conv.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] <folderfile> ..."
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile> (alt: --config)"
		     "\n       read configuration from file <configfile>"
		     "\n       (default: %s)"
		     "\n   -h (alt: --help)"
		     "\n       show this help"
		     "\n   -V (alt: --version)"
		     "\n       show the program's version number\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (1);
}

static struct option loptions[] = {
    { "config", required_argument, (int *) 0, 'f' },
    { "help", no_argument, (int *) 0, 'h' },
    { "version", no_argument, (int *) 0, 'V' },
    { NULL, 0, (int *) 0, 0 }
};

int main (int argc, char *argv[])
{
    int opt, optix = 0;
    int ix;

    if ((progname = strrchr(argv[0], '/'))) ++progname; else progname = argv[0];

    for (;;) {
	opt = getopt_long (argc, argv, "f:hV", loptions, &optix);
	if (opt == -1) { break; }
	switch (opt) {
	    case 'f':
		prefs.config_name = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'h':
		usage ();
	    case 'V':
		fprintf (stderr, "%s\n", vxVersion ());
		exit (0);
	    default:
		fprintf (stderr, "%s: unknown option `%s'. Please call `%s -h' "
				 "for help!\n",
				 progname,
				 argv[optind],
				 progname);
		exit (1);
	}
    }
    if (optind >= argc) {
	fprintf (stderr, "%s: no folder specified!\n", progname);
	exit (1);
    }
    /* Lies die Konfigurationsdatei ein. */
    getconf ();
    /* Initialisiere den Konverter. */
    conv_init ();
    /* Konvertiere jede angegebene Folderdatei. */
    for (ix = optind; ix < argc; ++ix) {
	xFile fx = xOpen (argv[ix], "ix", prefs.fperm);
	if (!fx) {
	    fprintf (stderr, "%s: opening folder `%s' failed",
			     progname,
			     argv[ix]);
	    perror ("");
	    continue;
	}
	conv (fx);
	xClose (fx);
    }
    return 0;
}
