/* smtp.c
**
** deliver a message to a non-local recipient
** (through an open SMTP connection).
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "tools.h"
#include "hlsock.h"

#include "userdb.h"

/* Entferne das Zeilenende (`\r', `\n' oder `\n') von einer Zeile und gib
** ein Flag zurck, das darber Auskunft gibt, ob die Zeile ein Zeilenende
** enthielt (also abgeschlossen war) oder nicht.
*/
static __inline__ int cuteol (char *line)
{
    char *p = line + strlen (line);
    if (p == line) { return 0; }
    if (*--p == '\r') {
	*p = '\0'; return 1;
    }
    if (*p == '\n') {
	*p = '\0';
	if (p != line && *--p == '\r') { *p = '\0'; }
	return 1;
    }
    return 0;
}

/* Lies smtliche Zeilen einer SMTP-Statusmeldung ein; gib aber nur den
** Status-Code ser ersten Zeile zurck.
*/
static int smtp_result (FILE *in)
{
    char line[1024], *p;
    int ecode = 0, stv;
    while (fgets (line, 1024, in)) {
	cuteol (line);
	vxlog (4, "%s\n", line);
	stv = (int) strtol (line, &p, 10);
	if (!ecode) { ecode = stv; }
	if (*p == '\r' || *p == '\n' || *p == ' ') { break; }
    }
    return ecode;
}

int smtp_send (HLSock serv, char *sender, char *addr, char *tmpfile, off_t from)
{
    FILE *in, *out, *fp;
    int ifd, ofd;
    int stv, ix;
    struct { char *f, *v; } p[3];
    char line[1024];

    /* Initialize the table */
    p[0].f = "HELO %s\r\n";      p[0].v = mailhost ();
    p[1].f = "MAIL FROM:%s\r\n"; p[1].v = sender;
    p[2].f = "RCPT TO:%s\r\n";   p[2].v = addr;

    /* Open the files */
    if (!(fp = fopen (tmpfile, "r"))) { return 1; }
    vxlog (5, "seeking position %ld\n", from);
    if (fseek (fp, from, SEEK_SET) < 0) { fclose (fp); return 1; }
    if ((ifd = dup (HLSockSD (serv))) < 0) { fclose (fp); return 1; }
    if ((ofd = dup (HLSockSD (serv))) < 0) {
	fclose (fp); close (ifd);
	return 1;
    }
    if (!(in = fdopen (ifd, "r"))) {
	fclose (fp); close (ifd); close (ofd);
	return 1;
    }
    setvbuf (in, NULL, _IOLBF, 0);
    if (!(out = fdopen (ofd, "w"))) {
	fclose (fp); fclose (in); close (ofd);
	return 1;
    }
    setvbuf (out, NULL, _IOLBF, 0);
    stv = smtp_result (in);
    if (stv / 100 != 2) {
	fclose (fp); fclose (in); fclose (out);
	return 2;
    }
    fputs ("\r\n", out);
    (void) smtp_result (in);
    for (ix = 0; ix < 3; ++ix) {
	vxlog (5, ">> "); vxlog (5, p[ix].f, p[ix].v);
	fprintf (out, p[ix].f, p[ix].v);
	stv = smtp_result (in);
	if (stv / 100 != 2) {
	    fclose (fp); fclose (in); fclose (out);
	    return ix + 3;
	}
    }
    fprintf (out, "DATA\r\n");
    vxlog (5, ">> DATA\r\n");
    stv = smtp_result (in);
    if (stv / 100 != 3) {
	fclose (fp); fclose (in); fclose (out);
	return 6;
    }
    while (fgets (line, 1024, fp)) {
	int waseol = cuteol (line);
	fputs (line, out);
	vxlog (5, "%s", line);
	if (waseol) { fputs ("\r\n", out); vxlog (5, "\r\n"); }
    }
    fclose (fp);
    fputs (".\r\n", out);
    vxlog (5, ".\r\n");
    stv = smtp_result (in);
    if (stv / 100 != 2) {
	fclose (in); fclose (out);
	return 7;
    }
    fputs ("QUIT\r\n", out);
    vxlog (5, "QUIT\r\n");
    stv = smtp_result (in);
    fclose (out); fclose (in);
    vxlog (5, "mail delivered, QUIT-status = %d\n", stv);
    return (stv / 100 == 2? 0 : 8);
}
