/* vxdomcheck.c
**
** Domain checker - for use in conjunction with the `smail' configuration
** file `/etc/smail/routers'
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "config.h"
#include "prefs.h"
#include "conf.h"
#include "dbif.h"
#include "tools.h"
#include "vxerror.h"
#include "version.h"
#include "vxopt.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] <hostname>"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile>"
		     "\n      read configuration from file <configfile>"
		     "\n      (default: %s)"
		     "\n   -h"
		     "\n      show this help"
		     "\n   -V"
		     "\n      show the program's version number\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

int main (int argc, char *argv[])
{
    char *vxdomain;
    DBIF_FILE userdb;
    dbif_rec *drec;

    vxopt (&argc, argv, usage);

    if (argc < 2) {
	fprintf (stderr, "%s: missing argument! Please call `%s -h' "
			 "for help!\n",
			 progname,
			 progname);
	exit (EX_USAGE);
    }
    vxdomain = argv[1];
    getconf ();
    setgid (prefs.vxgid);
    seteuid (prefs.vxuid);
    dbif_reset (&userdb, prefs.userdb_name);
    if (!userdb) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
		    "opening user database failed: %s\n",
		    dbif_errormsg ());
    }
    drec = dbif_resolve (userdb, vxdomain, NULL);
    dbif_close (userdb);
    if (!drec) { return 1; }
    if (dbif_rectype(drec) != RT_DOMAIN) {
	fprintf (stderr, "%s: `%s' is no domain in the `VxMail' system!\n",
			 progname, vxdomain);
	exit (EX_USAGE);
    }
    return 0;
}
