#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "prefs.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "dbif.h"
#include "version.h"
#include "vxopt.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] [-o <outputfile>]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile>"
		     "\n      read the configuration from file <configfile>"
		     "\n      (default: %s)"
		     "\n   -h"
		     "\n      show this help"
		     "\n   -V"
		     "\n      show the program's version number\n\n",
		     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

static int is_domain (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);
static int is_domlnk (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);
static int is_transl (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);
static int is_user (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);
static int is_link (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);
static void print_users (DBIF_FILE db, char *domkey, char *defadm, int nsd);
static void print_ulinks (DBIF_FILE db, char *domkey, char *defadm);

int main (int argc, char *argv[])
{
    int oc, o_set = 0, ofd, users, maxusers;
    DBIF_FILE userdb;
    DBIF_CURSOR slct;
    char *defadm, *key;
    dbif_rec *rec;

    vxopt (&argc, argv, usage);

    while ((oc = getopt (argc, argv, "o:")) != -1) {
	/* -f configfile -h help */
	switch (oc) {
	    case 'o':
		if (o_set) {
		    fprintf (stderr, "%s: duplicate `-o' option ignored!\n",
				     progname);
		    continue;
		}
		ofd = open (optarg, O_CREAT|O_TRUNC|O_WRONLY, prefs.fperm);
		if (ofd < 0) {
		    vxerror (EX_OSERR, errno, "opening output file failed");
		}
		dup2 (ofd, 1);
		close (ofd);
		break;
	    default:
		fprintf (stderr,
			 "%s: illegal option `%s'; use `%s -h' for help!\n",
			 progname, argv[optind], progname);
		exit (1);
	}
    }
    /* Alle weiteren Parameter werden ignoriert */
    /* Lies die Konfiguration */
    getconf ();
    dbif_reset (&userdb, prefs.userdb_name);
    if (!userdb) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
		    "Sorry, but i couldn't open the user database: %s\n",
		    dbif_errormsg ());
    }
    /* Lies den `identity'-Datensatz aus der Datenbank */
    if (!(rec = dbif_get (userdb, "::VXMAIL::"))) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "reading database identity failed: %s\n",
		    dbif_errormsg ());
    }
    if (dbif_rectype(rec) != RT_IDENTITY) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "user database identity record has the wrong type!\n");
    }
    /* Gib den Kopf der `Eingabedatei' aus */
    printf ("NEWDB \"%s\"\n", rec->folder);
    printf ("MAILHOST \"%s\"\n", rec->passwd);
    defadm = gdupl (__FILE__, __LINE__, rec->folder);
    dbif_free (rec);
    /* Lies alle Domains aus der Datenbank */
#if 0
    if (!(slct = dbif_select (userdb, NULL, CC_DOMAIN|CC_NSDOMAIN))) {
#else
    if (!(slct = dbif_select (userdb, NULL, is_domain))) {
#endif
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "selecting domain entries from the database failed!\n");
    }
    /* Lies fr jede Domain die normalen Benutzereintrge; ignoriere jedoch
    ** vorerst die Verweise auf Benutzereintrge.
    */
    for (key = dbif_first (slct, &rec); key; key = dbif_next (slct, &rec)) {
	if (dbif_rectype(rec) == RT_DOMAIN) {
	    dbif_getlim (rec, &users, &maxusers);
	    if (maxusers) {
		printf("DOMAIN \"%s\" \"%s\" %d\n", key, rec->folder, maxusers);
	    } else {
		printf("DOMAIN \"%s\" \"%s\"\n", key, rec->folder);
	    }
	} else {
	    printf("NONSTD DOMAIN \"%s\" \"%s\"\n", key, rec->folder);
	}
	print_users (userdb, key, defadm, dbif_rectype(rec) == RT_NSDOMAIN);
	free (key); dbif_free (rec);
    }
    /* In einem zweiten Durchgang mssen die Verweise auf andere Benutzer
    ** gelesen werden.
    */
    for (key = dbif_first (slct, &rec); key; key = dbif_next (slct, &rec)) {
/*	printf ("DOMAIN \"%s\"\n", key);*/
	print_ulinks (userdb, key, defadm);
	free (key); dbif_free (rec);
    }
    /* Die normalen Domains werden nicht lnger bentigt! */
    dbif_end (slct);
    /* Ich bentige nun die Verweise auf Domains */
#if 0
    if (!(slct = dbif_select (userdb, NULL, CC_DOMLINK))) {
#else
    if (!(slct = dbif_select (userdb, NULL, is_domlnk))) {
#endif
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "seleccting domain links from the user database failed!\n");
    }
    /* Gib alle Verweise auf Domain Links aus */
    for (key = dbif_first (slct, &rec); key; key = dbif_next (slct, &rec)) {
	printf ("DOMAIN \"%s\" LINKTO \"%s\"\n", key, rec->folder);
	free (key); dbif_free (rec);
    }
    /* Die Verweise auf die Domains werden nun nicht mehr bentigt */
    dbif_end (slct);
    /* Nun bentige ich die Domain-Umleitungen und IP-bersetzungen */
#if 0
    if (!(slct = dbif_select (userdb, NULL, CC_IPTRANS|CC_REDIRECT))) {
#else
    if (!(slct = dbif_select (userdb, NULL, is_transl))) {
#endif
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "selecting ip translations and domain redirects from "
		    "the user database failed!\n");
    }
    /* Nun mssen diese Eintrge ausgegeben werden */
    for (key = dbif_first (slct, &rec); key; key = dbif_next (slct, &rec)) {
	switch (dbif_rectype(rec)) {
	    case RT_IP4TRANS:
		printf ("IP4TRANS %s \"%s\"\n", key, rec->folder);
		break;
	    case RT_IP6TRANS:
		printf ("IP6TRANS %s \"%s\"\n", key, rec->folder);
		break;
	    case RT_REDIRECT:
		printf ("REDIRECT \"%s\" \"%s\" TO \"%s\"\n",
			key, rec->passwd, rec->folder);
		break;
	}
	free (key); dbif_free (rec);
    }
    /* Auch diese Eintrge werden nun nicht lnger bentigt! */
    dbif_end (slct);
    if ((rec = dbif_get (userdb, "::VXTS::"))) {
	if (dbif_rectype(rec) != RT_VXTS) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"VXTS record has the wrong type!\n");
	}
	printf ("VXTS \"%s\"\n", rec->passwd);
	dbif_free (rec);
    }
    if (!(rec = dbif_get (userdb, "::MAILSYS::"))
    || dbif_rectype(rec) != RT_MAILSYS) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "Mail system administrator record not found!\n");
    }
    printf ("MAILSYS \"%s\" \"%s\"\n", rec->passwd, rec->folder);
    dbif_free (rec);
    fflush (stdout);
    dbif_close (userdb);
    return 0;
}

static void print_users (DBIF_FILE db, char *domkey, char *defadm, int nsd)
{
    char *dakey = (char *) 0, *ukey, *p;
    dbif_rec *darec = (dbif_rec *) 0, *urec;
    DBIF_CURSOR usl;
#if 0
    int utypes = CC_USER|CC_NSUSER|CC_ADMIN|CC_LLINK|CC_INVALID|CC_FILTER|CC_BSMTP;
    if (!(usl = dbif_select (db, domkey, utypes))) {
#else
    if (!(usl = dbif_select (db, domkey, is_user))) {
#endif
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "selecting users from domain `%s' failed!\n");
    }
    for (ukey = dbif_first (usl, &urec); ukey; ukey = dbif_next (usl, &urec)) {
	if (!(p = strchr (ukey, '@'))) {
	    fprintf (stderr, "inconsistent entry in the user database!\n");
	} else {
	    *p = '\0';
	}
	if (dbif_rectype(urec) == RT_ADMIN && !strcmp (ukey, defadm)) {
	    if (dakey) {
		fprintf (stderr, "inconsistent entry in the user database!\n");
	    } else {
		dakey = ukey;
		darec = urec;
		continue;
	    }
	}
	switch (dbif_rectype (urec)) {
	    case RT_LLINK:
		printf ("USER \"%s\" LLTO \"%s\"\n", ukey, urec->folder);
		break;
	    case RT_FILTER:
		printf ("FILTER \"%s\"\n", urec->folder);
		break;
	    case RT_BSMTP:
		printf ("USER \"%s\" \"%s\" BSMTP \"%s\"\n", ukey, urec->passwd,
			urec->folder);
		break;
	    default:
		printf ("USER \"%s\" \"%s\"", ukey, urec->passwd);
		switch (dbif_rectype(urec)) {
		    case RT_ADMIN:   printf (" ADMIN\n"); break;
		    case RT_INVALID: printf (" INVALID\n"); break;
		    default:         printf ("\n");
		}
	}
	free (ukey); dbif_free (urec);
    }
    dbif_end (usl);
    if (!dakey) {
	if (!nsd) {
	    fprintf (stderr, "missing admin entry for domain `%s'!\n", domkey);
	}
	return;
    }
    printf ("ADMIN \"%s\"\n", darec->passwd);
    free (dakey); dbif_free (darec);
}

static void print_ulinks (DBIF_FILE db, char *domkey, char *defadm)
{
    char *lkey, *p;
    dbif_rec *lrec;
    DBIF_CURSOR ull;
    int dh_printed = 0;
#if 0
    if (!(ull = dbif_select (db, domkey, CC_LINK))) {
#else
    if (!(ull = dbif_select (db, domkey, is_link))) {
#endif
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "selecting link entries from domain `%s' failed!\n");
    }
    for (lkey = dbif_first (ull, &lrec); lkey; lkey = dbif_next (ull, &lrec)) {
	if (!dh_printed) { printf ("DOMAIN \"%s\"\n", domkey); dh_printed = 1; }
	if (!(p = strchr (lkey, '@'))) {
	    fprintf (stderr, "inconsistent link entry in the domain `%s'!\n",
			     domkey);
	} else {
	    *p = '\0';
	}
	if (!strcmp (lkey, defadm)) {
	    fprintf (stderr, "inconsistent link entry in domain `%s'!\n",
			     domkey);
	}
	printf ("USER \"%s\" LINKTO \"%s\"\n", lkey, lrec->folder);
	free (lkey); dbif_free (lrec);
    }
    dbif_end (ull);
}

static int is_domain (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_DOMAIN) || (rt == RT_NSDOMAIN);
}

static int is_domlnk (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_DOMLINK);
}

static int is_transl (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    return (rt == RT_IP4TRANS) || (rt == RT_IP6TRANS) || (rt == RT_REDIRECT);
}

static int is_user (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p;
    if ((1 << rt) & (CC_DENTRIES - CC_LINK)){
	if (!lkey) { return 1; }
	if (!(p = strchr (rkey, '@'))) { return 0; }
	++p;
	return !strcmp (lkey, p);
    }
    return 0;
}

static int is_link (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec)
{
    int rt = dbif_rectype (rrec);
    char *p;
    if (rt == RT_LINK) {
	if (!lkey) { return 1; }
	if (!(p = strchr (rkey, '@'))) { return 0; }
	++p;
	return !strcmp (lkey, p);
    }
    return 0;
}
