#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#include "prefs.h"
#include "conf.h"
#include "tools.h"
#include "vxerror.h"
#include "version.h"
#include "vxopt.h"

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] [<confitem>]"
		     "\n       %s -n [<confitem>]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -f <configfile>"
		     "\n      read the configuration from file <configfile>"
		     "\n      (default: %s)"
		     "\n   -n"
		     "\n      don't read any configuration file"
		     "\n   -h"
		     "\n      show this help"
		     "\n   -V"
		     "\n      show the program's version number\n\n",
		     progname, progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

int main (int argc, char *argv[])
{
    struct passwd *pw;
    struct group *gr;
    int opt, f_set, n_set = 0, ix;
    char *optname, *optval;

    f_set = vxopt (&argc, argv, usage);

    while ((opt = getopt (argc, argv, "n")) != -1) {
	switch (opt) {
	    case 'n':
		if (n_set) {
		    fprintf (stderr, "%s: second `-f'-option ignored!\n",
				     progname);
		    continue;
		}
		if (f_set) {
		    fprintf (stderr, "%s: you cannot specify `-f' and `-n' "
				     "together!\n",
				     progname);
		    return 1;
		}
		n_set = 1;
		break;
	    default:
		fprintf (stderr,
			"illegal option `%s'! Please call `%s -h' for help!\n",
			argv[optind],
			progname);
		return 1;
	}
    }
    if (!n_set) { getconf (); }
    pw = getpwuid (prefs.vxuid);
    gr = getgrgid (prefs.vxgid);
    if (optind >= argc) {
	for (ix = 0; confitem (ix, &optname, NULL, &optval); ++ix) {
	    if (optval) {
		printf ("%s %s\n", optname, optval);
	    }
	    gfree (optname); gfree (optval);
	}
    } else {
	for (ix = 0; confitem (ix, &optname, NULL, &optval); ++ix) {
	    if (!strcmp (optname, argv[optind])) {
		printf ("%s\n", optval);
		gfree (optname); gfree (optval);
		break;
	    }
	    gfree (optname); gfree (optval);
	}
    }
    return 0;
}
