#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>

#include "vxmkdb.tab.h"
#include "ParserData.h"
#include "tools.h"
#include "prefs.h"
#include "conf.h"
#include "vxerror.h"
#include "version.h"
#include "vxopt.h"

extern int yyparse (void);

ParserData sd;

static void usage (void);

int main (int argc, char *argv[])
{
    int opt, tabstop = 8;
    char *ifile = NULL, *p;
    long int ovi = -1, sdelay = -1;

    vxopt (&argc, argv, usage);

    opterr = 0;
    heap_init ();
    while ((opt = getopt (argc, argv, "d:t:")) >= 0) {
	switch (opt) {
	    case 'd':
		if (sdelay >= 0) {
		    fprintf (stderr, "%s: ignoring additional `-d' option!\n",
				     progname);
		    continue;
		}
		ovi = strtol (optarg, &p, 0);
		if (*p) {
		    vxerror (EX_USAGE, 0, "`-d' option argument must be "
					  "numeric!\n");
		}
		if (ovi < 0 || ovi > 30) {
		    vxerror (EX_USAGE, 0, "`-d' option argument is out of "
					  "range!\n");
		}
		sdelay = ovi;
		break;
	    case 't':
		ovi = strtol (optarg, &p, 0);
		if (*p) {
		    vxerror (EX_USAGE, 0, "`-t' option argument must be "
					  "numeric!\n");
		}
		if (ovi < 0 || ovi > 30) {
		    vxerror (EX_USAGE, 0, "`-t' option argument is out of "
					  "range!\n");
		}
		tabstop = (int) ovi;
		break;
	    default:
		vxerror (0, 0, "illegal option `-%c'!\n", optopt);
		vxerror (EX_USAGE, 0, "use `%s -h' for help!\n");
	}
    }
    max_sleep_delay = (unsigned int) (sdelay < 0 ? 0 : sdelay);
    if (optind < argc) {
	ifile = gdupl (__FILE__, __LINE__, argv[optind]);
    }
    if (ifile) {
	if (!(freopen (ifile, "r", stdin))) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ -2, "`%s'",ifile);
	}
    }
    getconf ();
    yyinit ((ifile ? ifile : "<stdin>"), tabstop);
    return yyparse ();
}

static void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <cfgfile>] [-t <tabsize>] [-d <maxdt>] "
		     "[<inputfile>]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -d <maxdt>"
		     "\n      specify the maximum delay interval `%s' should "
		     "wait before retrying"
		     "\n      if connecting to the database failed"
		     "\n      (0 deactivates this feature)"
		     "\n   -f <cfgfile>"
		     "\n      read the configuration from file <cfgfile>"
		     "\n      (default: %s)"
		     "\n   -h"
		     "\n      show this help"
		     "\n   -V"
		     "\n      show the program's version number\n\n",
		     progname, progname, progname, progname, CONFIGFILE);
    exit (1);
}
