%{
#include <stdio.h>
#include <strings.h>
#include "ParserData.h"
#include "vxmkdb.tab.h"
#include "config.h"
#include "prefs.h"

/* Declare the external named i need fro parsing */
extern char *yytext;
extern ParserData sd;
extern void ccerror (ParserData *sd, char *fmt, ...);

void cvesc (char *text);
void cvipV4 (char *text);
void cvipV6 (char *text);

%}
IP4STR	"["([0-9]{1,3}\.){3}[0-9]{1,3}"]"
IP6STR	"["([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}"]"
DTSTR	\"([^\"]|\\\")*\"
DTNUM	[0-9]+

%x illegal NO
%%

^[ \t]*#.*\n	{ ++sd.lc; sd.cl = 0; }
^[ \t]*\n	{ ++sd.lc; sd.cl = 0; }
"\n"		{ ++sd.lc; sd.cl = 0; return '\n'; }
"\r\n"		{ ++sd.lc; sd.cl = 0; return '\n'; }
"\r"		{ ++sd.lc; sd.cl = 0; return '\n'; }
[ ]		{ ++sd.cl; }
<NO>[ ]		{ ++sd.cl; }
"\t"		{ sd.cl = (sd.cl + sd.tabsize - 1) % sd.tabsize; }
<NO>"\t"	{ sd.cl = (sd.cl + sd.tabsize - 1) % sd.tabsize; }
    /* Mit NEWDB wird der Name des Benutzers, an den Fehlermeldungen etc.
    ** vereinbart; in diesem Fall wird die Datenbank neu erzeugt.
    */
NEWDB		{ sd.cl += 5; return NEWDB; }
    /* Mit OLDDB wird eine vorhandene Datenbank erzeugt!
    **/
OLDDB		{ sd.cl += 5; return OLDDB; }
    /* Mit DOMAIN wird eine neue DOMAIN definiert bzw. eine vorhandene
    ** selektiert.
    */
<NO>DOMAIN	{ BEGIN(INITIAL); sd.cl += 6; return NO_DOMAIN; }
NODOMAIN	{ sd.cl += 8; return NO_DOMAIN; }
DOMAIN		{ sd.cl += 6; return DOMAIN; }
    /* Mit USER werden neue Benutzereintrge innerhalb einer Domain angelegt
    ** bzw. vorhandene Benutzereintrge abgendert.
    */
<NO>USER	{ BEGIN(INITIAL); sd.cl += 4; return NO_USER; }
NOUSER		{ sd.cl += 6; return NO_USER; }
USER		{ sd.cl += 4; return USER; }
    /* ADMIN ist ein Benutzerflag, welches einem (noch zu schreibenden)
    ** Verwaltungsprogramm mitteilt, da dieser Benutzer diese spezielle
    ** Domain administrieren darf.
    */
ADMIN		{ sd.cl += 5; return ADMIN; }
    /* LINKTO erzeugt einen Verweis auf einen anderen Benutzereintrag.
    ** Dieser Eintrag mu vorhanden sein und sollte kein Verweis sein.
    */
LINKTO		{ sd.cl += 6; return LINKTO; }
    /* Mit dem Flag INVALID wird ein Benutzer gesperrt; dies bedeutet, da
    ** er weder Mail empfangen kann noch empfangene Mail per POP3, ... abholen
    ** kann.
    */
INVALID		{ sd.cl += 7; return INVALID; }
    /* Mit MAILSYS wird der Benutzer festgelegt, der das gesamte Mail-System
    ** (d.h. diese Datenbank) verwalten darf.
    */
MAILSYS		{ sd.cl += 7; return MAILSYS; }
    /* Es gibt auch noch die normal von `sendmail' (bzw. `smail') angelegten
    ** Folder. um diese verwalten zu knnen, bentige ich ein weiteres Flag:
    ** SMTPFOLDER. Wenn dieses Flag gesetzt ist, dann werden alle Benutzer
    ** die fr die betreffende Domain angelegt werden, mit dem Flag
    ** `nonstandard' versehen. Dies bedeutet fr `vxdeliver', da an diese
    ** Benutzer keine Mails direkt zugestellt werden knnen; fr `vxpop'
    ** bedeutet dies, da solche Folder nach einem leicht modifizierten
    ** Verfahren bearbeitet werden mssen.
    */
NONSTD		{ sd.cl += 6; return NONSTD; }
    /* Um Domains in der maximalen Zahl ihrer Benutzer einzuschrnken,
    ** mu man hinter der Folderdir-Angabe die maximale zugelassene Anzahl
    ** an Benutzern fr diese Domain angeben; um diese Einschrnkung wieder
    ** aufheben zu knnen, mu man stattdessen das folgende Schlsselwort
    ** angeben.
    */
NONE		{ sd.cl += 4; return NONE; }
    /* Die Daten (Parameter fr die einzelnen Direktiven) knnen entweder
    ** Flags oder Zeichenketten sein.
    */
    /* Mit `MAILHOST' kann ein Hostname fr das `VxMail'-System festgelegt
    ** werden.
    */
MAILHOST	{ sd.cl += 8; return MAILHOST; }
    /* Eine `IP4TRANS'- bzw. `IP6TRANS'-Direktive bezeichnet eine bersetzung
    ** einer IPv4- bzw. IPv6-Adresse in einen Domainnamen. Diese Eintrge
    ** habe ich ursprnglich wegen der DUMMEN Eigenart smtlicher Netscape-
    ** Browservarianten eingefhrt, die den Domainteil von der Benutzerkennung
    ** abschneiden und damit virtuelle Domains in der POP3-Richtung unmglich
    ** machen; inzwischen habe ich mir etwas anderen einfallen lassen, nmlich
    ** ein Ersatzzeichen fr den `@': den `:'. Dieses Zeichen wird vom
    ** POP3-Server VxPOP automatisch in ein `@' bersetzt, bevor eine Suche
    ** in der Benutzerdatenbank stattfindet - und: (Zumindest die 4.xx-Version)
    ** des Netscape (der Communicator) erkennt dieses Zeichen als normales
    ** Zeichen in einer Benutzerkennung an.
    ** Die `IPxTRANS'-Eintrge bleiben trotzdem in der Datenbank; vielleicht
    ** mchte ja jemand stattdessen lieber mit IP-basierten Mail-Domains
    ** arbeiten. Die IP-Nummer fr die bersetzung wird brigens aus der
    ** Socket-Verbindung gewonnen.
    */
<NO>IP4TRANS	{ BEGIN(INITIAL); sd.cl += 8; return NO_IP4TRANS; }
NOIP4TRANS	{ sd.cl += 10; return NO_IP4TRANS; }
IP4TRANS	{ sd.cl += 8; return IP4TRANS; }
<NO>IP6TRANS	{ BEGIN(INITIAL); sd.cl += 8; return NO_IP6TRANS; }
NOIP6TRANS	{ sd.cl += 10; return NO_IP6TRANS; }
IP6TRANS	{ sd.cl += 8; return IP6TRANS; }
    /* `REDIRECT's sind ein spezielles Feature des POP3-Servers VxPOP; um
    ** nmlich vollstndige Mail-Domains von einem Rechner auf einen anderen
    ** umziehen lassen zu knnen, mten normalerweise smtliche Benutzer,
    ** die in der jeweiligen Domain stehen, vorher informiert werden und die
    ** Einstellungen ihrer Mail-Programme (Name des POP-Hosts) entsprechend
    ** ndern; mit einem `REDIRECT' ist es mglich, Domains fr den Benutzer
    ** (nahezu) transparent umziehen zu lassen; findet nmlich VxPOP ein
    ** `REDIRECT', dann versucht es, zu dem neuen POP3-Server eine
    ** POP3-Verbindung aufzubauen; glckt dies, dann werden Befehle, die vom
    ** Benutzerprogramm kommen (in den allermeisten Fllen) auf den anderen
    ** Server umgeleitet und Ausgaben von diesem an den Benutzer
    ** weitergereicht. Hat man den VxPOP entsprechend konfiguriert, dann ist
    ** dieser POP3-Server zustzlich in der Lage, jedem Benutzer dieser Domain
    ** eine zustzliche Mail (on oberster Stelle) unterzuschieben, z.B. um
    ** ihn darber zu informieren, da er seine POP3-Einstellungen ndern mu.
    ** Das `REDIRECT'-Feature ist bislang jedoch noch nicht getestet; es
    ** sollten zumindest erst ein paar Testlufe mit diesem Feature gemacht
    ** werden, bevor es in der Praxis eingesetzt wird.
    */
<NO>REDIRECT	{ BEGIN(INITIAL); sd.cl += 8; return NO_REDIRECT; }
NOREDIRECT	{ sd.cl += 10; return NO_REDIRECT; }
REDIRECT	{ sd.cl += 8; return REDIRECT; }
    /* Dies sollte die `REDIRECT'-Direktive etwas leichter lesbar machen:
    **
    **   REDIRECT "<Domainname>" "<VXTS-Pawort>" TO "<Hostname>"
    **
    ** Ansonsten dient das Schlsselwort `TO' keinem besonderen Zweck.
    */
TO		{ sd.cl += 2; return TO; }
    /* `VXTS' dient einem VxPOP-Server zur Entschlsselung der in einem
    ** `VXTS'-Befehl hereinkommenden Daten; diese dienen im Falle einer
    ** `REDIRECT'ion dazu, dem VxPOP-Server einen anderen APOP-Zeitstempel
    ** zu geben; nur mit diesem neuen Zeitstempel, der von dem POP3-Server
    ** stammt, der eine Verbindung auf den aktuellen umgeleitet hat, ist ein
    ** `APOP'-Kommando ber eine umgeleitete Verbindung mglich.
    */
<NO>VXTS	{ BEGIN(INITIAL); sd.cl += 4; return NO_VXTS; }
NOVXTS		{ sd.cl += 6; return NO_VXTS; }
VXTS		{ sd.cl += 4; return VXTS; }
    /* `LLTO' ist dem `LINKTO' sehr hnlich; es funktioniert jedoch nur im
    ** Zusammenhang mit `USER'. Damit sollen Mailinglisten bzw. (sehr einfache
    ** Aliases mglich sein. Im Gegensatz zu `LINKTO'-Eintrgen ist die
    ** Existenz der Zieladresse nicht garantiert!
    */
LLTO		{ sd.cl += 4; return LLTO; }
    /* `BSMTP' kennzeichnet eine spezielle Variante des `LLTO'-Kommandos */
BSMTP		{ sd.cl += 5; return BSMTP; }
    /* Mit FILTER wird ein einfacher Mail-Filter fr die jeweils aktuelle
    ** Domain eingefhrt.
    */
<NO>FILTER	{ BEGIN(INITIAL); sd.cl += 6; return NO_FILTER; }
NOFILTER	{ sd.cl += 8; return NO_FILTER; }
FILTER		{ sd.cl += 6; return FILTER; }
    /* Mit DELETE werden Eintrge aus der Datenbank gelscht; dabei mu jedoch
    ** beachtet werden, da im Falle Benutzer- und Domaineintrgen nicht nur
    ** diese Eintrge, sondern auch alle auf (bzw. in) sie verweisenden bzw.
    ** in ihnen stehenden Eintrge mitgelscht werden.
    */
DELETE		{ sd.cl += 6; BEGIN(NO); }
    /* NO ist einfach eine krzere Alternative zu DELETE */
NO		{ sd.cl += 2; BEGIN(NO); }
    /* Die folgenden Angaben stellen Daten dar, die mit den einzelnen
    ** Direktiven bergeben werden.
    */
{IP4STR}	{ sd.cl += yyleng; cvipV4 (yytext); return IP4NUM; }
{IP6STR}	{ sd.cl += yyleng; cvipV6 (yytext); return IP6NUM; }
{DTSTR}		{ sd.cl += yyleng; cvesc (yytext); return DATA; }
{DTNUM}		{ sd.cl += yyleng; return NUMERIC; }
    /* Alle anderen Eingabesymbole sind illegal! */
<NO>.		{   /* BEGIN the skipping of an illegal token */
		    BEGIN(illegal);
		}
.		{   /* BEGIN the skipping of an illegal token */
		    BEGIN(illegal);
		}
<illegal>.	{   /* CONTINUE the skipping of an illegal token */
		    ;
		}
<illegal>"\n"	{   /* END the skipping of an illegal token! */
		    BEGIN(INITIAL);
		    return ILLEGAL;
		}
%%

/*  Mit `cvesc' wird die Zeichenkette `text' umgewandelt; hierbei werden
    alle `\'-Sequenzen entsprechend den fr C blichen Regeln konvertiert
    und die `"' am Anfang und am Ende der Zeichenkette entfernt.
*/
void cvesc (char *text)
{
    long l;
    char *p = text, *q = text + 1, *r;
    while (*q && *q != '"') {
	if (*q == '\\') {
	    switch (*++q) {
		case 'x':
		    l = strtol (++q, &r, 16);
		    if (q == r) {
			*p++ = q[-1];
		    } else {
			*p++ = (char) l; q = r;
		    }
		    break;
		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		    l = strtol (q, &r, 8); *p++ = (char) l; q = r; break;
		case '"': case '\\': *p++ = *q++; break;
		case 'n':            *p++ = '\n'; ++q; break;
		case 'r':            *p++ = '\r'; ++q; break;
		case 't':            *p++ = '\t'; ++q; break;
		default:             *p++ = '\\'; *p++ = *q++;
	    }
	} else {
	    *p++ = *q++;
	}
    }
    *p = '\0';
}

void cvipV4 (char *text)
{
    unsigned long ov;
    char *p, *txp = text + 1;
    ov = strtoul (txp, &p, 10);
    *((unsigned char *) text++) = (unsigned char) ov;
    txp = p + 1;
    ov = strtoul (txp, &p, 10);
    *((unsigned char *) text++) = (unsigned char) ov;
    txp = p + 1;
    ov = strtoul (txp, &p, 10);
    *((unsigned char *) text++) = (unsigned char) ov;
    txp = p + 1;
    ov = strtoul (txp, &p, 10);
    *((unsigned char *) text) = (unsigned char) ov;
}

void cvipV6 (char *text)
{
    unsigned long ov;
    char *p, *q;
    unsigned char txb[16], *txp = txb;
    ov = strtoul (text, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    q = p + 1;
    ov = strtoul (q, &p, 16);
    *txp++ = (unsigned char) (ov >> 8);
    *txp++ = (unsigned char) (ov & 0xff);
    memcpy (text, (char *) txb, 16);
}
