%{
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/utsname.h>

#include "tools.h"
#include "prefs.h"

#include "procs.h"
#include "delete.h"

/* `lcc' seems not to support `alloca', so we must use `malloc' instead */
#ifdef __LCC__
#define alloca malloc
#endif

/* For a better error reporting */
#define YYERROR_VERBOSE 1

char *yytext;
ParserData sd;

int lasttok;

#define YYMAXDEPTH 10000

%}

%union {
    char *sval;
    unsigned char *uval;
    int flag;
}
%token Y_EOL
%token NEWDB OLDDB
%token MAILHOST
%token DOMAIN NO_DOMAIN NONSTD
%token USER NO_USER
%token FILTER NO_FILTER
%token IP4TRANS NO_IP4TRANS
%token IP6TRANS NO_IP6TRANS
%token REDIRECT NO_REDIRECT
%token VXTS NO_VXTS
%token MAILSYS
%token ADMIN INVALID LINKTO LLTO BSMTP TO NONE
%token DATA IP4NUM IP6NUM NUMERIC
%token ILLEGAL
%type  <sval> Username Password MailAddr AdminName Domainname Folderdir
%type  <sval> Hostname Folderpath ALPname Filename Filtername FilenameOpt
%type  <uval> IPv4 IPv6
%type  <flag> Flag
%start DBData
%%

/* Was ist zu tun?
**   Es gibt zwei Startzustnde:
**      - es soll eine neue Datenbank angelegt werden
**      - eine vorhandene Datenbank soll erweitert werden.
**   Damit sich das ganze mit dieser `yacc'-Syntax machen lt, mu
**   der Erffnungsvorgang von der Routine, die die erste Schreiboperation
**   vornimmt, vorgenommen werden.
**   Dies knnen `new_domain', `get_domain', `ins_mailadm' sein.
*/
DBData		: DBHead Mailhost Domains IPTransList VXTSopt MailAdmin
		    { yyclose (&sd); }
		;

DBHead		: NEWDB AdminName '\n'
		    { sd.adminname = $2; sd.newdb = 1; }
		| OLDDB '\n'
		    { sd.adminname = NULL; sd.newdb = 0; }
		|   { sd.adminname = NULL; sd.newdb = 0; }
		| error '\n'
		;

Mailhost	: MAILHOST Hostname '\n'
		    {   /* Gib dem Mailsystem einen Hostnamen */
			sd.mailhost = $2;
		    }
		|   {   /* Wenn kein Mailhost angegeben wurde, dann vergib
			** den `nodename' des Systems als Mailhost
			*/
			struct utsname uts;
			uname (&uts);
			sd.mailhost =
			    gdupl (__FILE__, __LINE__,
				   *uts.nodename ? uts.nodename : "localhost");
		    }
		| error '\n'
		    {   /* Wenn ein Fehler aufgetreten ist, dann vergib
			** den `nodename' des Systems als Mailhost
			*/
			struct utsname uts;
			uname (&uts);
			sd.mailhost =
			    gdupl (__FILE__, __LINE__,
				   *uts.nodename ? uts.nodename : "localhost");
		    }
		;


/* An dieser Stelle soll keine weitere Aktion erfolgen! */
Domains		: DomainDef Domains
/*		| DomainDelete Domains*/
		|
		;

/* Auch hier erfolgt noch keine Aktion */
DomainDef	: DomainHead Users1 DomainAdmin
		| DomainLink
		| NonStdDomHead Users
		    { insert_domain (&sd); }
		| DomainDelete
		;

DomainDelete	: NO_DOMAIN Domainname '\n'
		    {   /* delete a complete domain - including
			**   - all it's user entries
			**   - all domain links pointing to it
			**   - all user links pointing into it.
			*/
			delete_domain (&sd, $2);
			gfree ($2);
			if (sd.domainname && !strcmp (sd.domainname, $2)) {
			    gfree (sd.domainname);
			    sd.domainname = NULL;
			}
		    }
		;

DomainAdmin	: ADMIN Password '\n'
		    {
			ins_user (&sd, sd.adminname, $2, RT_ADMIN); gfree ($2);
			insert_domain (&sd);
		    }
		| ADMIN LLTO ALPname '\n'
		    {
			ins_alp (&sd, sd.adminname, $3); gfree ($3);
			insert_domain (&sd);
		    }
		| ADMIN Password BSMTP FilenameOpt '\n'
		    {   /* insert/update a BSMTP entry */
			ins_bsmtp (&sd, sd.adminname, $2, $4);
			gfree ($2); gfree ($4);
			insert_domain (&sd);
		    }
		|
		    {   /* Wenn kein DomainAdmin gesetzt ist, dann darf
			** die Domain nicht neu sein!
			*/
			/* A NONSTD-Domain doesn't need an ADMIN */
			if (sd.state == NEW_DOMAIN) {
			    ccerror (&sd, "missing domain admin for `%s'!",
					  sd.domainname);
			    YYERROR;
			}
			/* set the restrictions for this domain */
			if (sd.domainname) { insert_domain (&sd); }
		    }
		| error '\n'
		;

/* Wenn ich alle fr eine Domain bentigten Daten zusammen habe, dann kann
 * ich die neue Domain einleiten.
 */
DomainHead	: DOMAIN Domainname Folderdir Restriction '\n'
		    {   /* Begin a new domain */
			sd.domainname = $2;
			sd.folderdir = $3;
			sd.nonstd = 0;
			sd.users = 0;
			if (new_domain (&sd)) { YYERROR; }
			sd.state = NEW_DOMAIN;
			printf ("%s %s\n", $2, $3);
		    }
		| DOMAIN Domainname Restriction '\n'
		    {   /* Try to find an old domain */
			int maxusers = sd.maxusers;
			sd.maxusers = 0;
			sd.domainname = $2;
			if (get_domain (&sd)) { YYERROR; }
			sd.state = OLD_DOMAIN;
			/* A NON-STANDARD domain never has a restriction,
			** so the following part is skipped; if a STANDARD
			** domain has no restriction, the following part is
			** skipped, too; if the restriction is set and a new
			** restriction is required, check if the new
			** restriction is valid (i.e. not already more users
			** in the domain than the restriction requires);
			** if not, write an error message and skip this
			** part.
			*/
			if (sd.maxusers && maxusers && sd.users > maxusers) {
			    ccerror (&sd, "sorry, already more users in the "
					  "domain `%s' than allowed by the "
					  "restriction (%d)!",
					  sd.domainname,
					  maxusers);
			    YYERROR;
			}
			/* if the domain is a NON-STANDARD domain, the
			** restriction is ignored; otherwise it is set to the
			** new value.
			*/
			if (!sd.nonstd && maxusers) {
			    sd.maxusers = (maxusers < 0 ? 0 : maxusers);
			}
		    }
		| error '\n'
		;

Restriction	: NUMERIC
		    /* Restrict a domain to a maximum number of users */
		    { sd.maxusers = atoi (yytext); }
		|   /* Don't change an existing restriction */
		    { sd.maxusers = 0; }
		| NONE
		    /* Disable a restriction */
		    { sd.maxusers = -1; }
		;

DomainOpt	: DOMAIN
		|
		;

NonStdDomHead	: NONSTD DomainOpt Domainname Folderdir '\n'
		    {   /* Begin a new domain with non-standard folders */
			sd.domainname = $3;
			sd.folderdir = $4;
			sd.nonstd = 1;
			sd.maxusers = 0;
			sd.users = 0;
			if (new_domain (&sd)) {
			    ccerror (&sd, "the domain `%s' already exists!",
					  sd.domainname);
			    YYERROR;
			}
			sd.state = NEW_DOMAIN;
			printf ("%s %s\n", $3, $4);
		    }
		;

DomainLink	: DOMAIN Domainname LINKTO Domainname '\n'
		    { ins_domlink (&sd, $2, $4); }
		;

Users1		: UserDef Users
		| UserDelete Users
		;

Users		: UserDef Users
		| UserDelete Users
		|
		;

/* Eine USER-Zeile kann zwei verschiedene Aktionen auslsen:
** a) die Eintragung eines normalen E-Mail-Benutzers
** b) die Eintragung eines Links (Verweises auf einen Benutzer)
**
** Im Fall
** a) mu sichergestellt werden, da der Eintrag nicht bereits als Link
**    existiert.
** b) mu sichergestellt werden, da der Eintrag nicht bereits als User
**    existiert.
** Auerdem mu in beiden Fllen sichergestellt sein, da der Eintrag nicht
** den Namen des Domainadministrators hat.
** 
*/
UserDef		: USER Username Password Flag '\n'
		    {   /* insert/update an e-mail user */
			int rt = $4;
			if (sd.nonstd && rt == RT_USER) {
			    rt = RT_NSUSER;
			} else if (sd.nonstd && rt == RT_ADMIN) {
			    ccerror (&sd, "nonstandard domains never "
					  "can have admin users!");
			    gfree ($2); gfree ($3);
			    YYERROR;
			}
			if (ins_user (&sd, $2, $3, rt)) {
			    ++sd.users;
			}
			gfree ($2); gfree ($3);
		    }
		| USER Username LINKTO MailAddr '\n'
		    {   /* insert/update a link (an alias user) */
			/* A link to a user does not count as an ordinary
			** user; so, no restriction checking is required!
			*/
			ins_link (&sd, $2, $4);
			gfree ($2); gfree ($4);
		    }
		| USER Username LLTO ALPname '\n'
		    {   /* insert/update a simple Alias|Listname|cmdref */
			ins_alp (&sd, $2, $4);
			gfree ($2); gfree ($4);
		    }
		| USER Username Password BSMTP FilenameOpt '\n'
		    {   /* insert/update a BSMTP entry */
			ins_bsmtp (&sd, $2, $3, $5);
			gfree ($2); gfree ($3); gfree ($5);
		    }
		| FILTER Filtername '\n'
		    {   /* insert/update a mail filter */
			ins_filter (&sd, $2);
			gfree ($2);
		    }
		| error '\n'
		;

Flag		: ADMIN		{ $$ = RT_ADMIN; }
		| INVALID	{ $$ = RT_INVALID; }
		|		{ $$ = RT_USER; }
		;

UserDelete	: NO_USER Username '\n'
		    { /* delete a single user and all user links pointing to */
			if (delete_user (&sd, $2)) {
			    if (sd.users > 0) { --sd.users; }
			}
			gfree ($2);
		    }
		| NO_FILTER '\n'
		    { /* delete an e-mail filter */
			delete_user (&sd, "::FILTER::");
		    }
		;

MailAdmin	: MAILSYS Password Folderpath '\n'
		    { ins_mailadm (&sd, $2, $3); gfree ($2); gfree ($3); }
		| MAILSYS LLTO ALPname '\n'
		    { ins_alp (&sd, "::MAILSYS::", $3); gfree ($3); }
		|   {   /* Die Datenbank darf nicht neu sein! */
			if (sd.newdb) {
			    ccerror (&sd,
				     "in a new database, you must "
				     "specify a mail admin!");
			    YYERROR;
			}
		    }
		;

IPTransList	: IPTrans IPTransList
		| DeleteIPTrans IPTransList
		|
		;

IPTrans		: IP4TRANS IPv4 Domainname '\n'
		    {   /* insert a POP3 IPv4 translation */
			char tag[30];
			sprintf(tag, "[%u.%u.%u.%u]", *$2, $2[1], $2[2], $2[3]);
			ins_iptrans (&sd, tag, $3, 0);
			gfree ($2); gfree ($3);
		    }
		| IP6TRANS IPv6 Domainname '\n'
		    {	/* insert a POP3 IPv6 translation */
			char tag[30];
			sprintf (tag, "[%x:%x:%x:%x:%x:%x:%x:%x]",
				      (*$2 << 8)   | $2[1],
				      ($2[2] << 8) | $2[3],
				      ($2[4] << 8) | $2[5],
				      ($2[6] << 8) | $2[7],
				      ($2[8] << 8) | $2[9],
				      ($2[10] << 8)| $2[11],
				      ($2[12] << 8)| $2[13],
				      ($2[14] << 8)| $2[15]);
			ins_iptrans (&sd, tag, $3, 1);
			gfree ($2); gfree ($3);
		    }
		| REDIRECT Domainname Password TO Hostname '\n'
		    {   /* insert a POP3 redirection to another HOST */
			ins_redirect (&sd, $2, $3, $5);
			gfree ($2); gfree ($3); gfree ($5);
		    }
		| error '\n'
		;

DeleteIPTrans	: NO_IP4TRANS IPv4 '\n'
		    { /* delete an IPv4 translation */
			delete_special (&sd, $2, RT_IP4TRANS);
			gfree ($2);
		    }
		| NO_IP6TRANS IPv6 '\n'
		    { /* delete an IPv6 translation */
			delete_special (&sd, $2, RT_IP6TRANS);
			gfree ($2);
		    }
		| NO_REDIRECT Domainname '\n'
		    { /* delete an domain redirection */
			delete_special (&sd, $2, RT_REDIRECT);
			gfree ($2);
		    }
		;

VXTSopt		: VXTS Password '\n'
		    {   /* insert a VXTS password */
			ins_vxts (&sd, $2);
			gfree ($2);
		    }
		| NO_VXTS '\n'
		    { /* delete the VXTS password */
			delete_special (&sd, "::VXTS::", RT_VXTS);
		    }
		|
		| error '\n'
		;

MailAddr	: DATA
		    {
			if (macheck (&sd, yytext)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

Username	: DATA
		    {	/* A user name will now be checked for consistency
			** before used in any environment which requires it
			*/
			if (prefs.lcuser && strcmp (yytext, "<OTHER>")) {
			    lcconv(yytext, -1);
			}
			if (uncheck (&sd, yytext, 0)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

Password	: DATA
		    { $$ = gdupl (__FILE__, __LINE__, yytext); }
		;

Domainname	: DATA
		    {	/* A domain name will now be checked for consistency
			** before used in any environment which requires it
			*/
			if (prefs.lcdomain) { lcconv(yytext, -1); }
			if (dncheck (&sd, yytext)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

Folderdir	: DATA
		    {   /* Simple consistency check for folder directories */
			if (fdcheck (&sd, yytext)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

Folderpath	: DATA
		    {   /* Simple consistency check for MAILSYS' folder path */
			if (fpcheck (&sd, yytext)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

AdminName	: DATA
		    {   /* Consistency check for the admin name; this consists
			** a) of a consistency check for a normal user name and
			** b) of a check which guarantees that `<OTHER>' will
			**    never be used as admin name
			*/
			if (prefs.lcuser && strcmp (yytext, "<OTHER>")) {
			    lcconv(yytext, -1);
			}
			if (uncheck (&sd, yytext, 1)) { YYERROR; }
			if (!strcmp (yytext, "<OTHER>")) {
			    ++sd.lc;
			    ccerror (&sd, "illegal admin name `%s'!\n", yytext);
			    --sd.lc;
			    YYERROR;
			}
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

IPv4		: IP4NUM
		    {   $$ = (unsigned char *) galloc (__FILE__, __LINE__, 4);
			memcpy ($$, yytext, 4);
		    }
		;

IPv6		: IP6NUM
		    {   $$ = (unsigned char *) galloc (__FILE__, __LINE__, 8);
			memcpy ($$, yytext, 8);
		    }
		;

Hostname	: DATA
		    {
			if (hncheck (&sd, yytext)) { YYERROR; }
			$$ = gdupl (__FILE__, __LINE__, yytext);
		    }
		;

ALPname		: DATA
		    { $$ = gdupl (__FILE__, __LINE__, yytext); }
		;

Filename	: DATA
		    { $$ = gdupl (__FILE__, __LINE__, yytext); }
		;

FilenameOpt	: Filename
		    { $$ = $1; }
		| /* empty */
		    { $$ = NULL; }
		;

Filtername	: DATA
		    { $$ = gdupl (__FILE__, __LINE__, yytext); }
		;

%%

/* Die Routinen, die ursprnglich hier unten standen, sind in eine eigene
** Datei gewandert: procs.c
** Weiter oben wird lediglich noch die Interfacedatei (procs.h) eingefgt.
*/
