/* main.c
**
** Main program of vxpop
** a pop3 server which handles virtual mail domains
** (c) 1998, Boris Jakubith
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/wait.h>
/* i had to include the following conditional, because the GLIBC2 header
** file `socketbits.h' (version 2.0.7v) makes it otherwise impossible
** to compile this file with any other than the GNU C-compiler.
*/
#if defined(linux) && defined(__GLIBC__) && __GLIBC__ >= 2 && !defined(__GCC__)
#define __cmsg_nxthdr main__cmsg_nxthdr
#endif
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
/*#include <arpa/inet.h>*/

#include "conf.h"
#include "prefs.h"
#include "dbif.h"
#include "exitcodes.h"
#include "tools.h"
#include "vxerror.h"
#include "config.h"
#include "conv.h"
#include "version.h"

#include "userdb.h"
#include "uf.h"
#include "popmsg.h"
#include "apop.h"
#include "cli.h"
#include "vxhost.h"

void pre_exit (void)
{
    if (dbif_isopen (userdb)) {
	vxlog (2, "pre_exit: closing user database\n");
	dbif_close (userdb);
    }
    if (uf_folder) {
	vxlog (2, "pre_exit: closing user folder\n");
	xClose (uf_folder);
    }
}

static struct option loptions[] = {
    { "config", required_argument, (int *) 0, 'f' },
    { "nodetach", no_argument, (int *) 0, 'd' },
    { "help", no_argument, (int *) 0, 'h' },
    { "port", required_argument, (int *) 0, 'p' },
    { "server", no_argument, (int *) 0, 's' },
    { "log", required_argument, (int *) 0, 'l' },
    { "verbose", no_argument, (int *) 0, 'v' },
    { "version", no_argument, (int *) 0, 'V' },
    { NULL, no_argument, (int *) 0, '\0' }
};

/* Diese Variablen bentige ich fr die `vxpop'-Konfiguration im
** Servermodus
*/
static int servermode = 0, logging = 0, serverport = LISTEN_TO, port_set = 1;

/* Diese Routine beginnt den eigentlichen POP3-Service
** main() dagegen dient hier (nebst aufgerufenen Funktionen) nur zur
** Initialisierung des Gesamtsystems.
*/
static void start_pop3 (void);

/* Hier arbeite ich mit `SIGCHLD' um die terminierten Subprozesse abzuholen. */
static void cleanup_child (int sig)
{
    int cldstat;
    pid_t child;
    while ((child = waitpid ((pid_t) -1, &cldstat, WNOHANG)) > 0) {
	if (WIFEXITED (cldstat)) {
	    fprintf (stderr, "%s(%d) exited with status %d!\n!",
			     progname,
			     child,
			     WEXITSTATUS (cldstat));
	} else if (WIFSIGNALED (cldstat)) {
	    fprintf (stderr, "%s(%d) caught signal %d!\n",
			    progname,
			    child,
			    WTERMSIG (cldstat));
	}
    }
    signal (sig, cleanup_child);
}

/* `finished' sagt `vxpop' im Servermodus, ob es terminieren soll */
static int finished = 0;
static pid_t pgroup = 0, pgold = 0;

/* Im Servermodus wird diese Funktion zum Terminieren aufgerufen */
static void cleanup_server (int sig)
{
    setpgid (getpid (), pgold);
    killpg (pgroup, SIGHUP);
    setpgid (getpid (), pgroup);
    finished = 1;
}

/* Setze den tcp-Port abhngig vom Namen des Service; das bentige ich, weil
** anscheinend der Servicename nicht genormt ist, und daher in den
** `/etc/services' verschiedener Distributionen auch verschiedene Namen
** fr den POP3-Service stehen knnen.
*/
static void set_port (void)
{
    static struct servent *service = (struct servent *) 0;
    if (!service) {
	if (!(service = getservbyname (prefs.service, "tcp"))) {
	    return;
	}
    }
    serverport = (int) ntohs ((short) service->s_port);
}

/* Bekommt der Server (nicht die Subprozesse) ein `SIGHUP', dann soll er
** seine Konfiguration neu laden.
*/
static void reload_conf (int sig)
{
    getconf ();
    if (port_set) { set_port (); }
    signal (sig, reload_conf);
}

/* Meine - wie blich ziemliche ausschweifende ;-) Hilfe */
static void usage (void)
{
    fprintf (stderr, "\nUsage: %s [option ...]"
		     "\n       %s -h"
		     "\n       %s -V\n"
		     "\n   -d (alt: --nodetach)"
		     "\n      don't detach from terminal (implies `-s')"
		     "\n   -f <configfile> (alt: --config=<configfile>)"
		     "\n      read the configuration from file <configfile>"
		     "\n      (default: %s)"
		     "\n   -l <file> (alt: --log=<file>)"
		     "\n      log error messages to file <file>"
		     "\n   -p <portnum> (alt: --port=<portnum>)"
		     "\n      listen on port <portnum> (default = %d)"
		     "\n   -s (alt: --server)"
		     "\n      start as an independant server process"
		     "\n      (default: start as an `inetd' sub-server)"
		     "\n   -v (alt: --verbose)"
		     "\n      increase the logging level (log more info)"
		     "\n   -h (alt: --help)"
		     "\n      show this help"
		     "\n   -V (alt: --version)"
		     "\n      show the program's version number\n\n",
		     progname, progname, progname, CONFIGFILE, LISTEN_TO);
    exit (EX_USAGE);
}

/* Hauptprogramm:
** - Initialisiere einige Daten;
** - Hole die Argumente (Optionen; alle Argumente, die keine Optionen sind
**   werden ignoriert);
** - Lade die Konfiguration;
** - Wenn der Servermodus verlangt ist, dann gehe in den Servermodus und
**   starte den POP3-Service fr jede zustandenegekommene Verbindung;
** - Wenn kein Servermodus verlangt ist, dann starte den POP3-Service auf
**   der bereits bestehenden Verbindung (`inetd'-Subproze).
*/
int main (int argc, char *argv[])
{
    int ac, logfd, detach = 1;
    char **av, *logname = "/dev/null";
    FILE *pidfp;
    if ((progname = strrchr(argv[0], '/'))) ++progname; else progname = argv[0];
    opterr = 0;
    servermode = 0;
    logging = 0;
    loglevel = 0;
    finished = 0;
    port_set = 1;
    for (;;) {
	int optix = 0;
	int loptix = optix ? optix : 1;
	int c;
	opterr = 0;
	if ((c=getopt_long(argc, argv, "df:hp:sl:vV", loptions, &optix)) == -1) {
	    break;
	}
	switch (c) {
	    case 'd':
		detach = 0;
		servermode = 1;
		break;
	    case 'f':
		prefs.config_name = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'h':
		usage ();
	    case 'p':
		serverport = atoi (optarg);
		port_set = 0;
		break;
	    case 's':
		servermode = 1;
		break;
	    case 'l':
		logname = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'v':
		++loglevel;
		break;
	    case 'V':
		fprintf (stderr, "%s\n", vxVersion ());
		exit (0);
	    case '?':
		vxerror_at (EX_USAGE, 0, __FILE__, __LINE__,
			    "missing option arg for option `%s'\n",
			    argv[loptix]);
	    default:
		vxerror_at (EX_USAGE, 0, __FILE__, __LINE__, 
			    "illegal option `%s'!\n", argv[loptix]);
	}
    }
    ac = argc - optind;
    av = argv + optind;
    getconf ();
    if (port_set) { set_port (); }

    /* ffne eine LOG-Datei und trage sie als Standardfehlerausgabe ein */
    if ((servermode && detach) || !servermode) {
	if ((logfd = open (logname, O_CREAT|O_APPEND|O_WRONLY, LPERM)) < 0) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
			"opening logfile `%s' failed", logname);
	}
	fflush (stdout);
	dup2 (logfd, 2);
	close (logfd);
    }

    if (servermode) {
	/* Erst einmal die fr den Servermodus bentigten Variablen */
	int servsock, clisock;
	pid_t child;
	size_t addrlen;
	struct sockaddr_in server, client;
	/* Im Servermodus mu sich dieser Proze erst einmal vom Terminal
	** abkoppeln. Auerdem sollte der abgespaltene Serverproze eine
	** Datei mit der Prozenummer anlegen, damit ein einfaches Bearbeiten
	** mittels des Sendens von Signalen spter mglich ist.
	*/
	if (detach) {
	    switch ((child = fork ())) {
		case -1: /* ERROR */
		    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
				"fork() failed for server");
		case 0:  /* Child process */
		    if ((pidfp = fopen (VXPOPPIDFILE, "w"))) {
			fprintf (pidfp, "%d\n", getpid ());
			fclose (pidfp);
		    }
		    break;
		default: /* Parent process */
		    exit (0);
	    }
	}
	/* Im Servermodus mu dieser Proze auf Verbindungswnsche warten, dann
	** einen Sub-Proze starten, der die Bearbeitung einer hergestellten
	** Verbindung bernimmt und sich anschlieend weiter auf Verbindungen
	** warten.
	*/
	if ((servsock = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
			"building socket failed");
	}
	memset (&server, '\0', sizeof (server));
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = htonl(INADDR_ANY);
	server.sin_port = htons(serverport);
	if (bind (servsock, (struct sockaddr *) &server, sizeof (server)) < 0) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
			"binding local address failed");
	}
	listen (servsock, prefs.maxclients);
	signal (SIGCHLD, cleanup_child);
	signal (SIGTERM, cleanup_server);
	signal (SIGINT, cleanup_server);
	signal (SIGHUP, reload_conf);
	pgold = getpgrp ();
	if ((pgroup = setpgid (0, 0)) < 0) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
			"setpgrp () failed!");
	}

	/* Hier mu auf Verbindungsaufnahmen gewartet werden */
	for (;;) {
	    if (finished) {
		close (servsock);
		fprintf (stderr, "finished - terminating %s\n", progname);
		exit (0);
	    }
	    addrlen = sizeof (client);
	    clisock = accept (servsock, (struct sockaddr *) &client,
			      (int *) &addrlen);
	    if (clisock < 0) {
		if (errno == EINTR) { continue; }
		vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
			    "server sccept error");
	    }
	    switch ((child = fork ())) {
		case -1:
		    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__,
				"fork error");
		case 0:
		    close (servsock);
		    dup2 (clisock, 0);
		    dup2 (clisock, 1);
		    close (clisock);
		    setgid (prefs.vxgid);
		    seteuid (prefs.vxuid);
		    start_pop3 ();
		default:
		    close (clisock);
	    }
	}
    }
    /* Im inetd-Modus mu auf der bereits bestehenden Verbindung weiter
    ** gearbeitet werden.
    */
    start_pop3 ();
    return 0;
}

/* Die verschiedenen Signalbehandlungsroutinen */
static void conn_timeout (int sn);
static void conn_terminate (int sn);
static void fatal_error (int sn);
static void broken_pipe (int sn);

/* Ab hier steht folgendes fest:
 * fd == 0 (stdin) ist die Leserichtung der Verbindung zum Client;
 * fd == 1 (stdin) ist die Schreibrichtung der Verbindung zum Client;
 * fd == 2 (stderr) zeigt entweder in eine als Argument angegebene LOG-Datei
 *         oder nach `/dev/null' (also in die universelle Datensenke).
 */
static void start_pop3 (void)
{
    /* Wir beginnen damit, die Standardsignale abzufangen und auf eigene
    ** Routinen umzuleiten.
    */
    vxlog (2, "pop subprocess started!\n");
    vxlog (2, "installing signal handlers ");
    signal (SIGALRM, conn_timeout); vxlog (2, ".");
    signal (SIGTERM, conn_terminate); vxlog (2, ".");
    signal (SIGINT, conn_terminate); vxlog (2, ".");
    signal (SIGHUP, conn_terminate); vxlog (2, ".");
    signal (SIGBUS, fatal_error); vxlog (2, ".");
    signal (SIGSEGV, fatal_error); vxlog (2, ".");
    signal (SIGIOT, fatal_error); vxlog (2, ".");
    signal (SIGILL, fatal_error); vxlog (2, ".");
    signal (SIGPIPE, broken_pipe); vxlog (2, ".");
    vxlog (2, " done.\n");
    /* Nun noch die zustzlichen Aufgaben, die bein einem `exit' ausgefhrt
    ** werden mssen.
    */
    vxlog (2, "installing exin handler ...");
    atexit (pre_exit);
    vxlog (2, " done.\n");
    /* Nun mu die Benutzerdatenbank geffnet werden */
    vxlog (2, "connecting user database ...");
    userdb_init ();
    vxlog (2, " done.\n");
    /* Als nchstes mu der Hostname des Rechners bestimmt werden.
    ** Da dies ein `virtuelles' Mail-System ist, wird dieser Name aus
    ** der soeben geffneten Datenbank durch das Lesen einer Passenden
    ** IPv4-bersetzung (IP4TRANS) (oder spter einer IP6TRANS)
    ** bestimmt.
    */
    vxlog (2, "setting up virtual hostname ...");
    vxhost_setup ();
    vxlog (2, " done.\n");
    /* Nun erfolgt noch die Bestimmung des APOP-Timestamps */
    vxlog (2, "setting up apop-timestamp ...");
    apop_setup ();
    vxlog (2, " done.\n");
    /* Nun beginnt der eigentliche Lauf nach POP3-Protokoll */
    /* Zunchst einmal die Willkommensmeldung mit einem Timestamp, der
    ** anzeigt, da APOP-Kommandos mglich sind.
    */
    vxlog (2, "sending welcome message ...");
    pop_msg (OK, "VxPOP (Version %s) on %s ready; %s.",
	    VXMAIL_VERSION,
	    (vxhost_name ? vxhost_name : "localhost"),
	    apop_timestamp);
    vxlog (2, " done.\n");
    /* Als nchstes folgt die Initialisierung des Kommandointerpreters */
    vxlog (2, "preparing command line interpreter ...");
    cli_setup ();
    vxlog (2, " done.\n");
    /* Als nchstes mu das Modul fr die Konvertierung von Nicht-Standard-
    ** Mailfoldern initialisiert werden.
    */
    vxlog (2, "initializing conversion function ...");
    conv_init ();
    vxlog (2, " done.\n");
    /* Zuletzt die Eingabeschleife */
    vxlog (2, "accepting POP3 commands.\n");
    while (get_state () < 4) {
	char line[1024];
	size_t llen;
	alarm (VXPOP_CLITIMEOUT);
	if (!fgets (line, 1024, stdin)) {
	    vxlog (0, "error in input; terminating!\n");
	    pop_msg (ERR, "error in input - closing connection");
	    fclose (stdin);
	    fclose (stdout);
	    fclose (stderr);
	    exit (1);
	}
	vxlog (3, "CLIENT>> %s\n", line);
	llen = strlen (line);
	if (line[llen - 1] != '\n' && line[llen - 1] != '\r') {
	    char rest[1024];
	    size_t rlen;
	    while (fgets (rest, 1024, stdin)) {
		rlen = strlen (rest);
		if (rest[rlen - 1] == '\n' || rest[rlen - 1] == '\r') break;
	    }
	    alarm (0);
	    vxlog (0, "line too long; ignored!\n");
	    pop_msg (ERR, "input line too long");
	    continue;
	}
	alarm (0);
	vxlog (2, "Executing POP3 command\n");
	interpret (line);
	vxlog (2, "Completed.\n");
    }
    vxlog (2, "POP operations completed; terminating subserver!\n");
    exit (0);
}

/* Von `alarm()' ausgelst */
static void conn_timeout (int sn)
{
    vxlog (0, "connection timed out - terminating subprocess.\n");
    pop_msg (ERR, "connection timed out");
    /* Lsche alle bis dato als gelscht markierten Mails */
    cli_command (quit_cmd, NULL);
    exit (sn);
/*
    signal (sn, SIG_DFL);
    kill (getpid (), sn);
*/
}

/* durch `SIGTERM' ausgelst */
static void conn_terminate (int sn)
{
    vxlog (0, "signal %d caught - terminating subprocess.\n", sn);
    pop_msg (ERR, "server caught signal %d - exiting", sn);
    /* Lsche alle bis dato als gelscht markierten Mails */
    cli_command (quit_cmd, NULL);
    exit (sn);
/*
    signal (sn, SIG_DFL);
    kill (getpid (), sn);
*/
}

/* durch `SIGSEGV' oder `SIGBUS' oder `SIGIOT' ausgelst */
static void fatal_error (int sn)
{
    vxlog (0, "fatal error (signal %d) - terminating subprocess.\n", sn);
    pop_msg (ERR, "fatal error - exiting");
    /* Beende ohne die als gelscht markierten Mails zu lschen */
    cli_command (rset_cmd, NULL);
    cli_command (quit_cmd, NULL);
    exit (sn);
/*
    signal (sn, SIG_DFL);
    kill (getpid (), sn);
*/
}

/* durch `SIGPIPE' ausgelst */
static void broken_pipe (int sn)
{
    vxlog (0, "connection broken - terminating subprocess.\n", sn);
    pop_msg (ERR, "connection broken");
    /* Beende ohne die als gelscht markierten Mails zu lschen */
    cli_command (rset_cmd, NULL);
    cli_command (quit_cmd, NULL);
    exit (sn);
/*
    signal (sn, SIG_DFL);
    kill (getpid (), sn);
*/
}
