/* apop.c
**
** APOP-timestamp and authentication
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <time.h>

#include "dbif.h"
#ifdef __STDC__
#define PROTOTYPES 1
#endif
#include "md5.h"
#include "userdb.h"
#include "vxhost.h"
#include "tools.h"
#include "popmsg.h"
#include "cmi.h"
#include "cli.h"
#include "redirect.h"

/* Woraus setze ich meinen APOP-Stamp zusammen?
** Ich denke, ich werde die Unix-Zeit in Sekunden (seit dem 1.1.1970)
** whlen; auerdem die Proze-Nummer des aktuellen Prozesses und
** schlielich den aktuellen Hostname.
*/

typedef unsigned char uchar;

char *apop_timestamp = (char *) 0;

void apop_setup (char *ts)
{
    time_t utc;
    struct tm *tm;
    char pidbuf[15], smask[30], *hostname = (char *) vxhost_name;
    size_t reslen;
    /* Wenn ein Zeitstempel bergeben wurde, dann installiere diesen anstatt
    ** einen neuen zu erzeugen. Ich bentige dies fr den Fall, da eine
    ** APOP-Authentifizierung ber eine Umleitung geschehen soll.
    */
    if (ts) {
	if (apop_timestamp) { gfree (apop_timestamp); }
	apop_timestamp = gdupl (__FILE__, __LINE__, ts);
	return;
    }
    /* <yyyymmddhhmmss.pid@hostname> */
    if (!hostname) { hostname = ""; }
    time (&utc);
    tm = localtime (&utc);
    strftime (smask, 30, "<%Y%m%d%H%M%S.%%s@%%s>", tm);
    sprintf (pidbuf, "%d", getpid());
    reslen = strlen (smask) + strlen (pidbuf) + strlen (hostname) - 6;
    apop_timestamp = galloc (__FILE__, __LINE__, reslen + 1);
    sprintf (apop_timestamp, smask, pidbuf, hostname);
}

static int hdig (char c)
{
    char *hdigs = "00112233445566778899aAbBcCdDeEfF", *p;
    if ((p = strchr (hdigs, c))) {
	return (p - hdigs) >> 1;
    }
    return -1;
}

int apop_authenticate (char **_user, char *digest, dbif_rec **rec)
{
    unsigned char md5sum[16], vdigest[16];
    MD5_CTX ckcontext;
    int ix, rt;
    dbif_rec *res;
    char *newdom, *user = *_user;
    /* check if the digest is valid */
    if (strlen (digest) != 32) {
	pop_msg (ERR, "invalid MD5-digest for `%s'!", user);
	return 0;
    }
    for (ix = 0; ix < 16; ++ix) {
	int h1 = hdig (digest[2*ix]), h2 = hdig (digest[2*ix+1]);
	if (h1 < 0 || h2 < 0) {
	    pop_msg (ERR, "invalid MD5-digest for `%s'!", user);
	    return 0;
	}
	vdigest[ix] = (unsigned char) ((h1 << 4) | h2);
    }
    if (!(res = dbif_resolve (userdb, user, NULL))) {
	/* Wenn der Benutzer im ersten Durchgang nicht gefunden wurde, dann
	** mu nachgeschaut werden, ob der Domainteil vielleicht einen Verweis
	** auf eine andere Domain oder eine Umleitung bezeichnet; falls ja,
	** dann mu das ganze mit einem neuen Benutzernamen - zusammengesetzt
	** aus dem Namensteil des alten und dem neuen Domainnamen - noch einmal
	** probiert werden.
	*/
	char *dom = strchr (user, '@');
	dbif_rec *domrec;
	if (!dom) { pop_msg (ERR, "%s: no such user!", user); return 0; }
	++dom;
	if (!(domrec = dbif_resolve (userdb, dom, &newdom))) {
	    pop_msg (ERR, "%s: no such domain!", dom); return 0;
	}
	rt = dbif_rectype(domrec);
	if (rt == RT_NSDOMAIN || rt == RT_DOMAIN) {
	    dbif_rec *newrec;
	    size_t nl = (dom - user), dl = strlen (newdom);
	    char *newuser = galloc (__FILE__, __LINE__, nl + dl + 1);
	    memcpy (newuser, user, nl);
	    memcpy (newuser + nl, newdom, dl);
	    newuser[nl + dl] = '\0';
	    if ((newrec = dbif_resolve (userdb, user, NULL))) {
		dbif_free (domrec); free (newdom);
		res = newrec; gfree (user); user = newuser;
		goto DO_MD5CHK;
	    }
	} else if (rt == RT_REDIRECT) {
	    /* Hier die Behandlung fr eine Umleitung stehen. Diese macht
	    ** (quasi) den POP3-Server zum POP3-Client und sorgt dafr
	    ** (falls dies so konfiguriert wurde), da der Benutzer eine
	    ** Nachricht ber den Umzug seiner Domain auf einen neuen HOST
	    ** erhlt.
	    ** Dieser Teil ist leider NOCH NICHT GETESTET!
	    */
	    /* Leider ist das ganze etwas schief geraten; ich mu das ganze
	    ** irgendwann noch berarbeiten.
	    */
	    size_t nl = (dom - user), dl = strlen (newdom);
	    char *newuser = galloc (__FILE__, __LINE__, nl + dl + 1);
	    char *args[3];
	    int ec;
	    memcpy (newuser, user, nl);
	    memcpy (newuser + nl, newdom, dl);
	    newuser[nl + dl] = '\0';
	    vxlog (2, "   initialising redirection.\n");
	    if ((ec = init_redirect (domrec->folder, newuser))) {
		vxlog (1, "   redirecting POP3 request to host `%s' failed!\n",
			  domrec->folder);
		/*pop_msg (ERR, "   initialising redirection failed!");*/
		goto APOP_ERR;
	    }
	    if ((ec = send_vxts (domrec))) {
		vxlog (1, "   sending VXTS command to %s failed; APOP is "
			  "impossible!\n", domrec->folder);
		pop_msg (ERR, "   sending VXTS command failed; APOP is "
			      "impossible!");
		goto APOP_ERR;
	    }
	    args[0] = newuser;
	    args[1] = digest;
	    args[2] = NULL;
	    if ((ec = redirect_apop (2, args))) {
		vxlog (1, "   APOP command on host %s failed!\n",
			  domrec->folder);
	    } else {
		vxlog (2, "   APOP command on host %s succeeded!\n",
			  domrec->folder);
	    }
	APOP_ERR:
	    gfree (newuser);
	    dbif_free (domrec); domrec = (dbif_rec *) 0;
	    gfree (newdom); newdom = NULL;
	    dbif_free (res); *rec = (dbif_rec *) 0;
	    gfree (user); *_user = NULL;
	    /* Ich will die Datenbank nur mglichst kurz offen halten, d.h.
	    ** nach einer erfolgreichen Anmeldung mu sie sofort beendet
	    ** werden. Im Falle einer Umleitung bedeutet an dieser Stelle
	    ** ein Wert `ec = 0', da die Umleitung bereits etabliert wurde;
	    ** es wird also kein weiterer Datenbankzugriff mehr bentigt.
	    */
	    if (!ec) { userdb_exit (); }
	    return (ec ? 0 : 1);
	}
	dbif_free (domrec); free (newdom);
	pop_msg (ERR, "%s: no such user or host!", user);
	gfree (*_user); *_user = (char *) 0;
	return 0;
    }
DO_MD5CHK:
    /* `res' zeigt auf den Datensatz des (gefundenen!!!) Benutzers,
    ** `user' auf den (evtl. genderten!!!) Namen
    */
    rt = dbif_rectype(res);
    if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_NSUSER|CC_MAILSYS|CC_BSMTP))) {
	vxlog (0, "   APOP command failed - invalid user name.\n");
	pop_msg (ERR, "%s: invalid user", user);
	dbif_free (res); *rec = (dbif_rec *) 0;
	gfree (user); *_user = NULL;
	return 0;
    }
    MD5Init (&ckcontext);
    MD5Update (&ckcontext, (uchar *)apop_timestamp, strlen (apop_timestamp));
    MD5Update (&ckcontext, (uchar *)(res)->passwd, strlen ((res)->passwd));
    MD5Final (md5sum, &ckcontext);
    if (memcmp (md5sum, vdigest, 16)) {
	pop_msg (ERR, "MD5-digest doesn't match for `%s'!", user);
	dbif_free (res); res = (dbif_rec *) 0;
	/* Gib die Ergebnisse fr eine erfolglose Authentifizierung zurck! */
	gfree (user); *_user = (char *) 0;
	return 0;
    }
    /* Gib die Ergebnisse fr eine erfolgreiche Authentifizierung zurck! */
    *_user = user; *rec = res;
    return 1;
}
