/* cli.c
**
** command interpreter for `vxpop3d'.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>

#ifdef NEED_CRYPT_H
#include <crypt.h>
#endif

#include "dbif.h"
#include "vxerror.h"
#include "tools.h"
#include "prefs.h"
#include "rc4.h"
#include "popmsg.h"
#include "apop.h"
#include "userdb.h"
#include "uf.h"
#include "vxhost.h"
#include "config.h"
#include "sysauth.h"
#include "uidl.h"

#define _CLI_C
#include "cli.h"
#include "cmi.h"
#include "redirect.h"
#include "drac.h"
#include "poprelay.h"

/* Intialisiere eine Umleitung; diese Initialisierung besteht im wesentlichen
** aus einer nderung der Funktionszeiger in der weiter unten angegebenen
** Kommandoliste `cmds'
*/
int init_redirect (char *hostname, char *username);

/* Smtliche Aktionen fr den Interpreter;
** Die Argumente werden bereits vom Interpreter aufgeteilt;
** Das Ergebnis der Auswertung wird als Funktionswert zurckgegeben.
*/
static int do_apop (int nargs, char *argv[]); /*.DONE*/
static int do_dele (int nargs, char *argv[]); /*DONE*/
static int do_last (int nargs, char *argv[]); /* not ready yet */
static int do_list (int nargs, char *argv[]); /*DONE*/
static int do_noop (int nargs, char *argv[]); /*DONE*/
static int do_pass (int nargs, char *argv[]); /*DONE*/
static int do_quit (int nargs, char *argv[]); /*DONE*/
static int do_retr (int nargs, char *argv[]); /*DONE*/
static int do_rset (int nargs, char *argv[]); /*DONE*/
static int do_stat (int nargs, char *argv[]); /*DONE*/
static int do_top  (int nargs, char *argv[]); /*.DONE*/
static int do_uidl (int nargs, char *argv[]); /*DONE*/
static int do_user (int nargs, char *argv[]); /*DONE*/
static int do_vxts (int nargs, char *argv[]); /*DONE*/

/* Spezielle Argumentparser fr die Kommandos, bei denen auch Argumente mit
** darin enthaltenet Leerzeichen mglich sind
*/ 
static int arg_sngl (char *line, int *nargs, char *argv[]);
static int arg_apop (char *line, int *nargs, char *argv[]);

void set_sumsize (int *msgs, size_t *allsize)
{
    int ix;
    int msgn = 0;
    size_t size = 0;
    for (ix = 0; msgtbl[ix].length >= 0; ++ix) {
	if (msgtbl[ix].deleted) { continue; }
	++msgn;
	size += (size_t) msgtbl[ix].length;
    }
    *msgs = msgn;
    *allsize = size;
}

struct cstruct cmds[] = {
    { "APOP", 2, 2, 1, 3, do_apop, arg_apop },
    { "DELE", 1, 1, 3, 3, do_dele, (argparser) 0 },
    { "LAST", 0, 0, 3, 3, do_last, (argparser) 0 },
    { "LIST", 0, 1, 3, 3, do_list, (argparser) 0 },
    { "NOOP", 0, 0, 3, 3, do_noop, (argparser) 0 },
    { "PASS", 1, 1, 2, 3, do_pass, arg_sngl },
    { "QUIT", 0, 0, 0, 4, do_quit, (argparser) 0 },
    { "RETR", 1, 1, 3, 3, do_retr, (argparser) 0 },
    { "RSET", 0, 0, 3, 3, do_rset, (argparser) 0 },
    { "STAT", 0, 0, 3, 3, do_stat, (argparser) 0 },
    { "TOP",  2, 2, 3, 3, do_top, (argparser) 0 },
    { "UIDL", 0, 1, 3, 3, do_uidl, (argparser) 0 },
    { "USER", 1, 1, 1, 2, do_user, arg_sngl },
    { "VXTS", 1, 1, 1, 1, do_vxts, (argparser) 0 },
    { NULL,   0, 0, 0, 0, (cmdfunc) 0, (argparser) 0 }
};

static char *statename[] = {
    "unknown",
    "authentication",
    "authorization",
    "transaction",
    "update"
};

int cli_command (cmd_t cmd, ...)
{
    va_list ap;
    int nargs;
    char **av;
    va_start (ap, cmd);
    av = (char **) ap;
    for (nargs = 0; av[nargs]; ++nargs);
    va_end (ap);
    switch (cmd) {
	case apop_cmd: return cmds[0].action (nargs, av);
	case dele_cmd: return cmds[1].action (nargs, av);
	case last_cmd: return cmds[2].action (nargs, av);
	case list_cmd: return cmds[3].action (nargs, av);
	case noop_cmd: return cmds[4].action (nargs, av);
	case pass_cmd: return cmds[5].action (nargs, av);
	case quit_cmd: return cmds[6].action (nargs, av);
	case retr_cmd: return cmds[7].action (nargs, av);
	case rset_cmd: return cmds[8].action (nargs, av);
	case stat_cmd: return cmds[9].action (nargs, av);
	case top_cmd:  return cmds[10].action (nargs, av);
	case uidl_cmd: return cmds[11].action (nargs, av);
	case user_cmd: return cmds[12].action (nargs, av);
	case vxts_cmd: return cmds[13].action (nargs, av);
    }
    return -1;
}

static int cmd_args (char *line, char *argv[], int maxargs, int *nargs);

static int current_state = 0;

void cli_setup (void)
{
    current_state = 1;
}

int get_state (void)
{
    return current_state;
}
static void cut_eol (char *line)
{
    char *p = line + strlen (line);
    /* Schneide das bzw. die Zeilenendezeichen ab. */
    if (p != line && *--p == '\n') {
	*p = '\0';
	if (p != line && *--p == '\r') { *p = '\0'; }
    } else if (p != line && *p == '\r') {
	*p = '\0';
    }
}

int interpret (char *line)
{
    int ix, len, nargs, retcode;
    char *cmd, *argv[10];
    for (ix = 0; cmds[ix].cmdname; ++ix) {
	cmd = cmds[ix].cmdname;
	len = strlen (cmd);
	if (!lccmp(line, cmd, len) && (!cmd[len] || cmd[len] == ' ')) {
	    struct cstruct *cs = &cmds[ix];
	    cut_eol (line);
	    if (cs->ap) {
		if (cs->ap (line + len + 1, &nargs, argv)) { return 1; }
	    } else if (cmd_args (line + len + 1, argv, 10, &nargs)) {
		pop_msg (ERR, "syntax error in comand `%s'", cmd);
		return 1;
	    }
	    if (nargs < cmds[ix].minargs || nargs > cmds[ix].maxargs) {
		pop_msg (ERR, "wrong number of aruments for `%s'", cmd);
		return 2;
	    }
	    if (cmds[ix].validstate && cmds[ix].validstate != current_state) {
		pop_msg (ERR, "command `%s' is illegal in %s-state",
			    cmd, statename[current_state]);
		return 3;
	    }
	    retcode = cmds[ix].action (nargs, argv);
	    if (retcode) { return retcode; }
	    current_state = cmds[ix].newstate;
	    return 0;
	}
    }
    pop_msg (ERR, "unknown command `%s'", (line[4] = '\0', line));
    return 1;
}

static int cmd_args (char *line, char *argv[], int maxargs, int *nargs)
{
    int argc = 0;
    char *p;
    *nargs = 0;
    if (!*line) { argv[0] = NULL; return 0; }
    while ((p = strchr (line, ' '))) {
	if (argc < maxargs) {
	    argv[argc++] = line;
	}
	*p++ = '\0';
	line = p;
	++*nargs;
    }
    if (*line) {
	if (argc < maxargs) {
	    argv[argc++] = line;
	}
	++*nargs;
    }
    return 0;
}

static int arg_sngl (char *line, int *nargs, char *argv[])
{
    if (!*line) {
	*nargs = 0; argv[0] = NULL;
    } else {
	*nargs = 1; argv[0] = line; argv[1] = NULL;
    }
    return 0;
}

static int arg_apop (char *line, int *nargs, char *argv[])
{
    char *p;
    if (!*line) { *nargs = 1; argv[0] = NULL; return 0; }
    if (!(p = strrchr (line, ' ')) || (p - line) < 0) {
	*nargs = 1; argv[0] = line; argv[1] = NULL; return 0;
    }
    *nargs = 2; argv[0] = line; *p++ = '\0'; argv[1] = p; argv[2] = NULL;
    return 0;
}

static dbif_rec *userrec;
static char *username;
static int nmsgs = 0;
static int cryptauth = 0;
static size_t sumlen = 0;
static int lastix = 0;

static char *make_addr (char *user)
{
    char *res, *p, *hostname;
    /* Erzeuge die Benutzeradresse fr die Authentifizierung: */
    if (!strcmp (user, "::MAILSYS::")) {
	/* Die Adresse des Mail-Systemadministrators wird ohne Umschweife
	** sofort zurckgegeben.
	*/
	return gdupl (__FILE__, __LINE__, user);
    } else if (strchr (user, '@') || !strcmp (user, "::MAILSYS::")) {
	/* Eine bereits vollstndige E-Mail-Adresse wird einfach dupliziert */
	res = gdupl (__FILE__, __LINE__, user);
    } else if (strchr (user, ':')) {
	/* Dieser Teil stellt einen Workaround fr ein Problem mit allen
	** Versionen des Netscape-Navigator bzw. Communicator dar.
	** Dieser schneidet leider vom in der Konfiguration angegebenen
	** Benutzernamen den Teil ab dem ersten `@' ab. Um dieses Problem
	** zu umgehen, mu man stattdessen an dieser Stelle ein `:' angeben.
	** Dieses wird von `VxPOP' (an genau dieser Stelle) in ein `@'
	** umgewandelt.
	*/
	res = gdupl (__FILE__, __LINE__, user);
	p = strchr (res, ':'); *p = '@';
    } else {
	/* Enthlt der Benutzername kein `@' oder `:', dann ist er (noch)
	** unvollstndig; aus der IP-Nummer der Verbindung wurde bereits
	** (mittels eines IP4TRANS- oder IP6TRANS-Eintrages) ein Hostname
	** bestimmt; dieser wird einfach hinten an den Benutzernamen angehngt.
	*/
	hostname = vxhost_name ? (char *) vxhost_name : "localhost";
	res = gconc (__FILE__, __LINE__, user, "@", hostname, NULL);
    }
    if (prefs.lcuser && strncmp (res, "<OTHER>", 7)) {
	lcconv (res, strchr(res, '@') - res);
    }
    if (prefs.lcdomain) { lcconv (strchr(res, '@'), -1); }
    return res;
}

/* Eine APOP-Authentifizierung erfolgt mit einem Befehl:
**     APOP <username> <MD5 digest>
** Wenn diese Authentifizierung fehlschlgt, dann mu der Benutzer sich noch
** einmal vollstndig (etwa mittels USER/PASS) anmelden.
*/
static int do_apop (int nargs, char *argv[])
{
    vxlog (1, "   APOP command found.\n");
    /* Erzeuge die Benutzeradresse fr die Authentifizierung: */
    username = make_addr (argv[0]);
    vxlog (1, "   authenticating using `%s' and `%s'.\n", username, argv[1]);
    /* APOP-Autorisation ist mit /etc/passwd (bzw. /etc/shadow) nicht mglich;
    ** existiert jedoch bereits ein Benutzereintrag - auch wenn dieser keinen
    ** VxMail-spezifischen Folder hat, dann wird die APOP-Authentifizierung
    ** dadurch mglich, da dieser Benutzereintrag sein eigenes - unverschls-
    ** seltes Pawort hat.
    */
    if (apop_authenticate (&username, argv[1], &userrec)) {
	/* username = NULL indicates a redirected connection */
	if (!username) { return 0; }
	/* register the connection IP address for SMTP relay */
	DRACsend (mailhost ());
	POPRELAYsend (username);
	/* Die eigentliche Konvertierungsarbeit soll in `uf_setup' geschehen. */
	if (uf_setup (userrec)) {
	    vxlog (0, "   APOP setup failed.\n");
/*	    pop_msg (ERR, "setup failed, closing connection");*/
	    gfree (username); username = NULL;
	    /* Stelle die Benutzer- und Gruppen-ID's auf die konfigurierten
	     * Werte um */
	    seteuid (getuid ()); setgid (prefs.vxgid); seteuid (prefs.vxuid);
	    return 1;
	}
	/* Berechne die Summe der Lngen aller vorhandenen E-Mails */
	set_sumsize (&nmsgs, &sumlen);
	vxlog (1, "   APOP authentication and setup succeeded.\n");
	pop_msg (OK, "%d message%s in mail folder (%ld octet%s)",
		    nmsgs, (nmsgs != 1 ? "s" : ""),
		    sumlen, (sumlen != 1 ? "s" : ""));
	/* Ich will die Datenbank nur mglichst kurz offen halten, d.h.
	** nach einer erfolgreichen Anmeldung mu sie sofort beendet
	** werden. Im Falle einer normalen APOP-Authentifikation ist dies
	** an dieser Stelle der Fall.
	*/
	userdb_exit ();
	return 0;
    } else {
	vxlog (0, "   APOP authentication failed.\n");
	pop_msg (ERR, "authentication failed for `%s'", argv[0]);
	seteuid (getuid ());
	setgid (prefs.vxgid);
	seteuid (prefs.vxuid);
	return 1;
    }
}

static int do_user (int nargs, char *argv[])
{
    int rt;
    vxlog (1, "   USER command found.\n");
    /* Erzeuge die Benutzeradresse fr die Authentifizierung: */
    username = make_addr (argv[0]);
    vxlog (1, "   authentication name is `%s'.\n", username);
    /* USER/PASS-Autorisation ist auch mit /etc/passwd (bzw. /etc/shadow)
    ** mglich; jedoch reicht ein einfacher Vergleich nicht aus, da die
    ** Passwrter in dieser Datei (diesen Dateien) verschlsselt abgelegt
    ** sind. Ich mu also ein spezielles Flag setzen, das im PASS-Kommando
    ** dafr sorgt, Passwortvergleiche mit Benutzereintrgen, die aus den
    ** Systemdateien (/etc/passwd, ...) gewonnen wurden, anders verglichen
    ** werden als solche, die direkt aus der Datenbank kommen.
    ** Auerdem gilt volgende EINSCHRNKUNG:
    **     Wenn ein Benutzerdatensatz bereits in der Datenbank enthalten ist,
    **     dann wird - gleichgltig, ob Systemautorisation eingeschaltet
    **     ist oder nicht - immer die VxMail-spezifische Autorisation
    **     verwendet.
    */
    userrec = dbif_resolve (userdb, username, NULL);
TRY_AGAIN:
    if (userrec) {
	/* Wenn ein Benutzerdatensatz gefunden wurde, dann ist an dieser
	** Stelle alles OK - auch in dem Fall, da er einen `nonstandard'-
	** Folder hat.
	*/
	rt = dbif_rectype(userrec);
	if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_NSUSER|CC_MAILSYS|CC_BSMTP))) {
	    vxlog (0, "   USER command failed - invalid user name.\n");
	    pop_msg (ERR, "%s: invalid user", username);
	    gfree (username); username = NULL;
	    dbif_free (userrec); userrec = (dbif_rec *) 0;
	    seteuid (getuid ());
	    setgid (prefs.vxgid);
	    seteuid (prefs.vxuid);
	    return 1;
	}
	cryptauth = 0;
	vxlog (1, "   USER command succeeded - waiting for PASS or QUIT.\n");
	pop_msg (OK, "user name OK, please specify PASS command!");
	return 0;
    } else {
	/* An dieser Stelle IST ein `@' im Benutzernamen enthalten! */
	char *dom = strchr (username, '@') + 1, *newdom;
	/* Wenn der Benutzer im ersten Schritt nicht gefunden wurde, dann
	** schaue nach, ob die Domain existiert.
	*/
	userrec = dbif_resolve (userdb, dom, &newdom);
	if (userrec) {
	    rt = dbif_rectype(userrec);
	    if (rt == RT_NSDOMAIN) {
		free (newdom);
		/* Falls das der Fall ist und Systemautorisation erlaubt ist,
		** dann besorge die Benutzerdaten vom System.
		*/
		vxlog (1, "   User with nonstandard domain `%s'\n", dom);
		if (prefs.sysauth) {
		    dbif_rec *folderrec = userrec;
		    vxlog (1, "   system authentication enabled!\n");
		    userrec = sysauth (folderrec, username);
		    dbif_free (folderrec);
		    if (userrec) {
			vxlog (1, "   System user `%s' found!\n", username);
			cryptauth = 1;
			vxlog (1, "   USER command succeeded (non-VxMail "
				  "user) - waiting for PASS or QUIT.\n");
			pop_msg (OK, "user name OK, please specify PASS "
				     "command");
			return 0;
		    }
		}
	    } else if (rt == RT_DOMAIN) {
		/* Andernfalls untersuche, ob eine Domain gefunden wurde.
		** Falls ja, dann versuche es erneut (Domain-Link); in
		** diesem Fall enthlt `newdom' den Domainnamen, der mit
		** dem Benutzernamen zu der neuen E-Mail-Adresse zusammen-
		** gesetzt werden mu.
		*/
		size_t nl = (dom - username), dl = strlen (newdom);
		char *newaddr = galloc (__FILE__, __LINE__, nl + dl + 1);
		memcpy (newaddr, username, nl);
		memcpy (newaddr + nl, newdom, dl);
		newaddr[nl + dl] = '\0';
		/* Gib den Speicherplatz fr den alten Benutzerdatensatz frei */
		dbif_free (userrec);
		/* Gib ebenso den Speicherplatz fr den alten Benutzernamen
		** frei; belege anschlieend die entsprechende Variable mit
		** der neu erzeugten E-Mail-Adresse
		*/
		gfree (username); username = newaddr;
		/* Lies einen neuen Datensatz */
		if ((userrec = dbif_resolve (userdb, username, NULL))) {
		    /* Da es jetzt an den Anfang zurckgeht, mu zunchst der
		    ** Speicherplatz fr den Domainnamen - der aus dem letzten
		    ** vorherigen Aufruf von `dbif_resolve' stammt, freige-
		    ** geben werden.
		    */
		    free (newdom); goto TRY_AGAIN;
		}
	    } else if (rt == RT_REDIRECT) {
		/* Behandle eine Umleitung auf einen anderen HOST; dazu
		** mu der POP3-Server selbst zum POP3-Client werden, unter
		** der gegebenen Benutzerkennung die Nachrichten abholen und
		** ggf. (wenn dies so konfiguriert ist(!!!)) dem Benutzer eine
		** zustzliche Mail unterschieben, die ihn darber informiert,
		** da seine Domain umgezogen ist.
		** NAHEZU FERTIG; lediglich ein test des APOP-Kommandos ber
		** eine Umleitung fehlt noch.
		*/
		size_t nl = (dom - username), dl = strlen (newdom);
		char *newaddr = galloc (__FILE__, __LINE__, nl + dl + 1);
		char *args[2];
		int res;
		vxlog (2, "   found a domain redirection.\n");
		memcpy (newaddr, username, nl);
		memcpy (newaddr + nl, newdom, dl);
		newaddr[nl + dl] = '\0';
		/* setup the connection to the target host and setup the
		** redirection commands
		*/
		vxlog (2, "   initializing redirection.\n");
		if (init_redirect (userrec->folder, newaddr)) {
		    vxlog (1, "   redirecting POP3 request to host `%s' "
			      "failed!\n",
			      userrec->folder);
		    pop_msg (ERR, "redirecting request to host %s failed!",
				  userrec->folder);
		    gfree (newaddr); 
		    dbif_free (userrec); userrec = (dbif_rec *) 0;
		    gfree (username); username = (char *) 0;
		    return 1;
		}
		vxlog (2, "   OK, redirection ready!\n");
		/* try to enable APOP authentication */
		if (send_vxts (userrec)) {
		    vxlog (0, "   sending VXTS command to %s failed; APOP "
			      "is impossible!\n", userrec->folder);
		}
		args[0] = newaddr;
		args[1] = NULL;
		if ((res = redirect_user (1, args))) {
		    vxlog (0, "   USER command on host %s failed!\n",
			      userrec->folder);
		} else {
		    vxlog (2, "   USER command on host %s succeeded!\n",
			      userrec->folder);
		}
		/* Nach einer Umleitung werden die normalen Daten nicht
		** mehr bentigt - also weder der Benutzername noch der
		** Benutzerdatensatz.
		*/
		gfree (newaddr);
		dbif_free (userrec); userrec = (dbif_rec *) 0;
		gfree (username); username = (char *) 0;
		/* Ich will die Datenbank nur mglichst kurz offen halten, d.h.
		** nach einer erfolgreichen Anmeldung mu sie sofort beendet
		** werden. Im Falle einer Umleitung bedeutet an dieser Stelle
		** ein Wert `res = 0', da die Umleitung bereits etabliert
		** wurde es wird also kein weiterer Datenbankzugriff mehr
		** bentigt.
		*/
		if (!res) { userdb_exit (); }
		return res;
	    }
	    /* In jedem anderen Fall wurde der Benutzer nicht gefunden!
	    ** Der Speicherplatz fr den Namen der zuletzt gefundenen Domain
	    ** mu freigegeben werden.
	    */
	    free (newdom);
	    if (userrec) { dbif_free (userrec); userrec = (dbif_rec *) 0; }
	}
	vxlog (0, "   USER command failed - unknown user name.\n");
	pop_msg (ERR, "unknown user");
	gfree (username); username = NULL;
	seteuid (getuid ());
	setgid (prefs.vxgid);
	seteuid (prefs.vxuid);
	return 1;
    }
}

static int do_pass (int nargs, char *argv[])
{
    size_t pwlen = strlen (userrec->passwd);
    char *pass = galloc (__FILE__, __LINE__, pwlen + 1);
    int cres;
    vxlog (1, "   PASS command found.\n");
    memcpy (pass, userrec->passwd, pwlen);
    pass[pwlen] = '\0';
    /* Wenn `cryptauth' gesetzt ist, dann mssen die verschlsselten
    ** Varianten verglichen werden - sonst die unverschlsselten.
    */
    if (cryptauth) {
	cres = strcmp (crypt (argv[0], pass), pass);
    } else {
	cres = strcmp (argv[0], pass);
    }
    /* Wenn der Vergleich keine bereinstimmung ergeben hat, dann ist die
    ** Authorisation fehlgeschlagen!
    */
    if (cres) {
	vxlog (0, "   PASS command failed - passwords don't match.\n");
	pop_msg (ERR, "authentication failed for `%s'", username);
	gfree (username); username = NULL;
	dbif_free (userrec); userrec = (dbif_rec *) 0;
	current_state = 1;
	cryptauth = 0;
	seteuid (getuid ());
	setgid (prefs.vxgid);
	seteuid (prefs.vxuid);
	return 1;
    }
    /* Und wieder wird der E-Mail-Folder des Benutzers initialisiert; wenn
    ** es sich um einen `nonstandard'-Folder handelt, dann mu eine Indexdatei
    ** temporr angelegt werden.
    */
    DRACsend (mailhost ());
    POPRELAYsend (username);
    if (uf_setup (userrec)) {
	/*        ^^^^^^^  Value-bergabe ist hier richtig */
	vxlog (0, "   PASS command failed - opening user folder failed.\n");
/*	pop_msg (ERR, "setup failed, closing connection");*/
	gfree (username); username = NULL;
	dbif_free (userrec); userrec = (dbif_rec *) 0;
	cli_command (quit_cmd, "force", NULL);
	seteuid (getuid ());
	setgid (prefs.vxgid);
	seteuid (prefs.vxuid);
	return 1;
    }
    set_sumsize (&nmsgs, &sumlen);
    vxlog (1, "   USER/PASS authentication succeeded!\n");
    pop_msg (OK, "%d message%s in mail folder (%ld octet%s)",
		 nmsgs, (nmsgs != 1 ? "s" : ""),
		 sumlen, (sumlen != 1 ? "s" : ""));
    /* Ich will die Datenbank nur mglichst kurz offen halten, d.h.
    ** nach einer erfolgreichen Anmeldung mu sie sofort beendet
    ** werden. Im Falle einer USER/PASS-Authentifikation ist dies
    ** an dieser Stelle der Fall.
    */
    userdb_exit ();
    return 0;
}

static int cvckmsgn (char *nstrg)
{
    int n = atoi (nstrg);
    if (n < 1 || n > nmsgs) {
	pop_msg (ERR, "message number is out of range");
	return 0;
    }
    if (msgtbl[n - 1].deleted) {
	pop_msg (ERR, "no such message", n);
	return 0;
    }
    return n;
}

static int do_dele (int nargs, char *argv[])
{
    int n = cvckmsgn (argv[0]);
    if (!n) return 1;
    vxlog (1, "   DELE command found.\n");
    msgtbl[n - 1].deleted = 1;
    if (n > lastix) { lastix = n; }
    pop_msg (OK, "message %d deleted", n);
    return 0;
}
/* I implemented the LAST command only for backwards compatibility (with
** RFC1081); it is NOT part of the POP3-standard (RFC1957 -> STD0053)!!!
*/
static int do_last (int nargs, char *argv[])
{
    vxlog (1, "   LAST command found.\n");
    pop_msg (OK, "%d", lastix);
    return 0;
}

static int do_list (int nargs, char *argv[])
{
    int n = 0;
    int msgs;
    size_t allsize;
    vxlog (1, "   LIST command found.\n");
    if (nargs == 1) {
	n = cvckmsgn (argv[0]);
	if (!n) return 1;
	pop_msg (OK, "%d %ld", n, msgtbl[n - 1].length);
	return 0;
    }
    set_sumsize (&msgs, &allsize);
    if (msgs == 0) { pop_msg (ERR, "no messages found"); return 1; }
    pop_msg (OK, "%d messages (%ld octets)", msgs, allsize);
    for (n = 0; n < nmsgs; ++n) {
	if (msgtbl[n].deleted) { continue; }
	pop_msg (ML, "%d %ld", n + 1, msgtbl[n].length);
    }
    pop_msg (ML, ".");
    if (lastix < 1) { lastix = 1; }
    return 0;
}

static int do_noop (int nargs, char *argv[])
{
    vxlog (1, "   NOOP command found.\n");
    pop_msg (OK, "");
    if (lastix < 1) { lastix = 1; }
    return 0;
}

static int do_quit (int nargs, char *argv[])
{
    vxlog (1, "   QUIT command initiated - closing connection.\n");
    /* Hier verwende ich einen (bsen) kleinen Trick: wenn `do_quit' vom
    ** POP3-Interpreter aufgerufen wird, dann ist die Anzahl der bergebenen
    ** Parameter immer 0. Wenn ich es jedoch von anderer Seite aus aufrufe
    ** (eben die Stelle in `main'), dann knnen sehr wohl zustzliche Parameter
    ** bergeben werden; in letzterem Fall soll nicht immer eine Erfolgsmeldung
    ** mit ausgegeben werden. Daher diese `if'-Anweisung.
    */
    if (nargs < 1) { pop_msg (OK, "closing connection"); }
    fclose (stdout);
    fclose (stdin);
    if (current_state == 3) {
	uf_update ();
    }
    if (uf_folder) {
	xClose (uf_folder);
	uf_folder = (xFile) 0;
    }
    return 0;
}

static int do_retr (int nargs, char *argv[])
{
    char line[1024];
    int n = cvckmsgn (argv[0]);
    ssize_t llen, byte_stuff;
    vxlog (1, "   RETR command found.\n");
    if (!n) {
	vxlog (0, "   illegal message number %d.\n", n);
	return 1;
    }
    vxlog (5, "   Retrieving message %d ", n);
    From_Skip (uf_folder, n - 1); vxlog (5, ".");
    pop_msg (OK, "message %d follows.", n);
    byte_stuff = 1;
    vxlog (5, ".");
    while ((llen = xGets (line, 1024, uf_folder)) > 0) {
	pop_sstrg (line, byte_stuff);
	byte_stuff = (line[llen - 1] == '\n');
	vxlog (5, ".");
    }
    vxlog (5, ".");
/*    pop_sstrg ("\r\n", 1);*/
    pop_msg (ML, ".");
    if (n > lastix) { lastix = n; }
    vxlog (5, " done.\n");
    return 0;
}

static int do_rset (int nargs, char *argv[])
{
    int ix;
    vxlog (1, "   RSET command initiated.\n");
    for (ix = 0; ix < nmsgs; ++ix) {
	msgtbl[ix].deleted = 0;
    }
    /* Hier verwende ich einen (bsen) kleinen Trick: wenn `do_rset' vom
    ** POP3-Interpreter aufgerufen wird, dann ist die Anzahl der bergebenen
    ** Parameter immer 0. Wenn ich es jedoch von anderer Seite aus aufrufe
    ** (eben die Stelle in `main'), dann knnen sehr wohl zustzliche Parameter
    ** bergeben werden; in letzterem Fall soll nicht immer eine Erfolgsmeldung
    ** mit ausgegeben werden. Daher diese `if'-Anweisung.
    */
    if (nargs < 1) { pop_msg (OK, ""); }
    lastix = 0;
    return 0;
}

static int do_stat (int nargs, char *argv[])
{
    int msgs;
    size_t allsize;
    vxlog (1, "   STAT command found.\n");
    set_sumsize (&msgs, &allsize);
    pop_msg (OK, "%d %ld", msgs, allsize);
    if (lastix < 1) { lastix = 1; }
    return 0;
}

static int do_top  (int nargs, char *argv[])
{
    char line[1024];
    int n = cvckmsgn (argv[0]);
    int lc = atoi (argv[1]);
    int ix, byte_stuff;
    ssize_t llen;
    vxlog (1, "   TOP command found.\n");
    if (!n) { return 1; }
    pop_msg (OK, "message %d's header and %d line%s of body follow.",
		 n, lc, (lc == 1 ? "" : "s"));
    vxlog (5, "   Retrieving message %d ", n);
    From_Skip (uf_folder, n - 1);
/*##was:    xSeek (uf_folder, n - 1);*/
    byte_stuff = 1;
    vxlog (5, ".");
    while ((llen = xGets (line, 1024, uf_folder)) > 0) {
	pop_sstrg (line, byte_stuff);
	byte_stuff = (line[llen - 1] == '\n');
	if (*line == '\r' && (!line[1] || (line[1] == '\n' && !line[2]))) break;
	if (*line == '\n' && !line[1]) break;
	vxlog (5, ".");
    }
    ix = 0;
    byte_stuff = 1;
    vxlog (5, ".");
    while (ix < lc) {
	if ((llen = xGets (line, 1024, uf_folder)) <= 0) break;
	pop_sstrg (line, byte_stuff);
	if ((byte_stuff = (line[llen - 1] == '\n' || line[llen - 1] == '\r'))) {
	    ++ix;
	}
	vxlog (5, ".");
    }
/*    pop_sstrg ("\r\n", 1);*/
    pop_msg (ML, ".");
    vxlog (5, " done.\n");
    if (n > lastix) { lastix = n; }
    return 0;
}

static int hex_decode (char *out, const char *in)
{
    char *hstr = "00112233445566778899AaBbCcDdEeFf", *p = out, *q;
    int ix, ll = strlen (in), c;
    for (ix = 0; ix < ll; ix += 2) {
	if (!(q = strchr (hstr, in[ix]))) { return 1; }
	c = (q - hstr) >> 1;
	if (!(q = strchr (hstr, in[ix + 1]))) { return 1; }
	c = (c << 4) | ((q - hstr) >> 1);
	*p++ = (char) c;
    }
    *p = '\0';
    return 0;
}

static int do_vxts (int nargs, char *argv[])
{
    /* Was ist zu tun?
    ** Der Client ist in diesem Fall ein anderer VxPOP-Server, der seinen
    ** Zeitstempel mit dem gemeinsamen Geheimnis (Pawort) und dem Zeit-
    ** stempel dieses VxPOP-Servers verschlsselt bertrgt.
    ** Kann dieser POP3-Server diese Daten (die als Zeichenkette von
    ** Hexadezimalziffern angegeben sein mu) entschlsseln und erhlt
    ** dabei die IP-Nummer, unter der die Verbindung luft, so ist das Datum
    ** nach der IP-Nummer der Zeitstempel des anderen Systems und mu als der
    ** eigene Zeitstempel akzeptiert werden.
    */
    char *coded, *key;
    unsigned char sbox[256];
    size_t rl = strlen (argv[0]);
    dbif_rec *vxts_rec;
    if (rl & 1) {
	vxlog (0, "   the VXTS-argument doesn't have an even length!\n");
	pop_msg (ERR, "VXTS argument must be of even length!");
	return 1;
    }
    vxts_rec = dbif_get (userdb, "::VXTS::");
    if (!vxts_rec || dbif_rectype(vxts_rec) != RT_VXTS) {
	if (vxts_rec) { dbif_free (vxts_rec); }
	vxlog (0, "   no VXTS record found in database!\n");
	pop_msg (ERR, "no VXTS record! Sorry!");
	return 1;
    }
    if (!(*IPstr())) {
	dbif_free (vxts_rec);
	vxlog (0, "   IP string for comparison is empty!\n");
	pop_msg (ERR, "internal error! Sorry!");
	return 1;
    }
    key = gconc (__FILE__, __LINE__, vxts_rec->passwd, apop_timestamp, NULL);
    dbif_free (vxts_rec);
    rl >>= 1;
    coded = galloc (__FILE__, __LINE__, rl + 1);
    if (hex_decode (coded, argv[0])) {
	gfree (key); gfree (coded);
	vxlog (0, "   VXTS expects a string of hex-digits!");
	pop_msg (ERR, "VXTS argument MUST consist of hexadecimal digits!");
	return 1;
    }
    rc4_init ((unsigned char *) key, sbox);
    rc4_decrypt ((unsigned char *) coded, rl, sbox);
    if (strncmp (coded, IPstr(), strlen (IPstr()))) {
	gfree (key); gfree (coded);
	vxlog (0, "   IP string prefix and local IP string differ!\n");
	pop_msg (ERR, "VXTS authentication failed!");
	return 1;
    }
    apop_setup (coded + strlen (IPstr()));
    gfree (key); gfree (coded);
    pop_msg (OK, "VXTS command succeeded - new APOP timestamp set!");
    return 0;
}

static int do_uidl (int nargs, char *argv[])
{
    int n = 0;
    int msgs;
    size_t allsize;
    char uidlbuf[33];
    vxlog (1, "   UIDL command found.\n");
    if (nargs == 1) {
	n = cvckmsgn (argv[0]);
	if (!n) { return 1; }
	uidl_calc (uf_folder, n - 1, uidlbuf, 33);
	pop_msg (OK, "%d %s", n, uidlbuf);
	return 0;
    }
    set_sumsize (&msgs, &allsize);
    if (msgs == 0) { pop_msg (ERR, "no messages found"); return 1; }
    pop_msg (OK, "UIDL command accepted");
    for (n = 0; n < nmsgs; ++n) {
        if (msgtbl[n].deleted) { continue; }
	uidl_calc (uf_folder, n, uidlbuf, 33);
        pop_msg (ML, "%d %s", n + 1, uidlbuf);
    }
    pop_msg (ML, ".");
    return 0;
}

/* Initialize a redirection; this function may be called from
** apop_authenticate or from do_user.
** In both cases a VXTS command should be sent to the remote host after
** initializing the redirection
*/
int init_redirect (char *hostname, char *username)
{
    return redirect_setup (hostname, username, cmds);
}
