/* redirect.c
**
** redirect a POP3-request to another host
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>

#include "config.h"
#include "prefs.h"
#include "dbif.h"
#include "vxerror.h"
#include "tools.h"
#include "rc4.h"
#include "hlsock.h"
#include "exithl.h"

#include "userdb.h"
#include "cmi.h"
#include "popmsg.h"
#include "vxhost.h"
#include "apop.h"
#include "uidl.h"

/* Was mu ich tun?
** Zuerst mal mu eine POP3-Verbindung zu dem Zielrechner aufgebaut werden;
** wenn dies funktioniert hat, dann mu die Kommandoliste umgeschrieben werden
** - und zwar so, da die in diesem Modul enthaltenen Funktionen die Aufgaben
** der ursprnglichen Kommandos bernehmen; Rckgaben einzelner Kommandos
** mssen an den Benutzer weitergereicht werden; im Falle der Kommandos
** `STAT', `LIST', `TOP' und `RETR' jedoch nicht ohne diese vorher so
** verndert zu haben, da eine zustzliche Meldung eingefgt wird wenn dies
** so konfiguriert wurde.
*/

/* Folgende Variablen mssen exportiert werden, da - im Fall eines Abbruchs -
** die entsprechenden Resourcen wieder freigegeben werden mssen:
*/
char *rdfile = (char *) 0;

/* Ich bentige globale Variablen fr:
** - die Verbindungsdaten
** - den Namen der aus der in der Konfiguration angegebenen Textdatei erzeugten
**   Mail-Datei.
*/

static int rdmsg_deleted = 0, vxts_succeeded = 0;
static int clientport = LISTEN_TO;
static FILE *in, *out;
static char *ServSockname = NULL;
static char *redirect_timestamp = NULL;
static off_t rmsg_size = 0;
static int lastix = 0;

static __inline__ void cut_eol (char *s)
{
    char *p = s + strlen (s);
    if (p != s && p[-1] == '\n') {
	*--p = '\0';
	if (p != s && p[-1] == '\r') {
	    *--p = '\0';
	}
    } else if (p != s && p[-1] == '\r') {
	*--p = '\0';
    }
}

static __inline__ int has_eol (char *s)
{
    char *p = s + strlen (s);
    if (p == s) return 0;
    if (p[-1] == '\n' || p[-1] == '\r') { return 1; }
    return 0;
}

/* get a (line) from the redirected POP3 connection */
char *pop_gline (char *line, size_t lsz, FILE *in)
{
    char *r;
    alarm (REDIR_TIMEOUT);
    r = fgets (line, lsz, in);
    if (r) { vxlog (4, ">>> %s", line); }
    alarm (0);
    return r;
}

/* get a response message from the redirected POP3 connection;
** return a non-zero value for a positive response and zero for
** a negative one or a NULL response string or a response string
** not beginning with `+' or `-'; this function removes the state
** indicator from the from the beginning of the line
*/
int get_pop (char *line, size_t lsz, FILE *in)
{
    char c, *p;
    if (!pop_gline (line, lsz, in)) { return 0; }
    c = *line;
    /* Wenn ein Leerzeichen nach dem Statusindikator gefunden wird,
    ** dann soll alles ab dem ersten Zeichen danach an den Anfang der
    ** Zeichenkette gefunden werden.
    ** Wird kein Leerzeichen gefunden, dann wird nachgeschaut, ob ein
    ** Zeilenende vorliegt; wenn ja, dann wird dieses an den Anfang der
    ** Zeichenkette geschoben. Sonst wird lediglich das die Zeichenkette
    ** terminierende NUL-Zeichen an ihren Anfang verschoben.
    */
    if (!(p = strchr (line, ' '))) {
	p = line + strlen (line) - 1;
	if (*p == '\n') {
	    --p; if (*p == '\r') { --p; }
	} else if (*p == '\r') {
	    --p;
	}
    }
    /* Die eigentliche Verschiebung luft hier ab */
    while ((*line++ = *++p));
    /* Gib das passende Ergebnis zurck */
    if (c == '-') { return 0; }
    if (c == '+') { return 1; }
    return 0;
}

void rd_send (char *line, ...)
{
    va_list ap;
    va_start (ap, line);
    vfprintf (out, line, ap); fputs("\r\n", out); fflush (out);
    vxlog (4, "<<< "); vvxlog (4, line, ap); vxlog (4, "\\r\\n\n");
    va_end (ap);
}

static int tr_line (FILE *rmsg, FILE *tfp, char *line, char *hostname)
{
    char *hn;
    /* Solange das Tag `<HOST>' vollstndig gefunden wird, schreibe alles
    ** bis zum letzten Zeichen vor dem Tag und danach den Hostnamen; setze
    ** anschlieend direkt hinter dem Tag wieder auf
    */
    while ((hn = strstr (line, "<HOST>"))) {
	*hn = '\0';
	fputs (line, tfp);
	fputs (hostname, tfp);
	line = hn + 6;
    }
    /* Wenn der Anfang des Tags am Ende des Puffers gefunden wird, dann
    ** lies noch soviele Zeichen ein, wie bentigt werden, um das Tag zu
    ** identifizieren; handelt es sich um das Tag, dann schreibe alles
    ** bis zum letzten Zeichen vor dem Tag und danach den Hostnamen;
    ** handelt es sich nicht um das Tag, dann schreibe den gesamten
    ** Pufferrest.
    ** Danach sind wir FERTIG!
    */
    if ((hn = strrchr(line, '<')) && !strncmp(hn, "<HOST>", strlen(hn))) {
	if (!fgets (line + strlen(line), 6 - strlen(hn), rmsg)) {
	    fputs (line, tfp);
	} else if (!strcmp (hn, "<HOST>")) {
	    *hn = '\0'; fputs (line, tfp); fputs (hostname, tfp);
	} else {
	    fputs (line, tfp);
	}
    } else {
	fputs (line, tfp);
    }
    return has_eol (line);
}

void remove_rdfile (void)
{
    if (rdfile) { unlink (rdfile); }
}

/* Erzeuge aus der in der Konfiguration angegebenen Textdatei eine
** vollstndige E-Mail, die per POP3 wie jede andere E-Mail abgeholt werden
** kann. Diese E-Mail wird in einer Datei abgelegt, deren Name hinterher in
** der globalen Variablen `rdfile' zur Verfgung steht.
*/

void init_file (char *hostname, char *username)
{
    char tfbuf[40], line[1024 + 6];
    int tmpfd;
    time_t gmt;
    FILE *tfp, *rmsg;
    struct stat sb;
    if (!prefs.redir_info) { return; }
    sprintf (tfbuf, "/tmp/vxpop-redirect.%d", getpid ());
    /* Wenn beim ffnen ein Fehler auftritt, dann wird ohne zustzliche
    ** E-Mail weitergemacht.
    */
    if ((tmpfd = open (tfbuf, O_CREAT|O_TRUNC|O_WRONLY, 0600)) < 0) { return; }
    if (!(tfp = fdopen (tmpfd, "w"))) { close (tmpfd); unlink (tfbuf); return; }
    if (!(rmsg = fopen (prefs.redir_info, "r"))) {
	close (tmpfd); unlink (tfbuf);
	return;
    }
    (void) time (&gmt);
    fprintf (tfp, "Return-Path: <::MAILSYS::@%s>\r\n", mailhost());
    fprintf (tfp, "From: VxMail POP3-Server <VxPOP@%s>\r\n", mailhost());
    fprintf (tfp, "To: %s\r\n", username);
    (void) strftime (line, 1024, "Date: %a, %d %b %Y %H:%M:%S %z",
		     localtime (&gmt));
    fprintf (tfp, "%s\r\n", line);
    if (fgets (line, 1024, rmsg)) {
	fprintf (tfp, "Subject: ");
	/* Lies die Zeile bis zum Ende und bertrage sie nach Ersetzung aller
	** `<HOST>'-Tags in die Zieldatei.
	*/
	while (!tr_line (rmsg, tfp, line, hostname)) {
	    /* Wenn noch kein Zeilenende gefunden wurde, jedoch keine Daten
	    ** mehr vorhanden sind, dann mu ein abschlieendes Zeilenende
	    ** eingefgt und die Schleife sofort beendet werden.
	    */
	    if (!(fgets (line, 1024, rmsg))) { fputs ("\n", tfp); break; }
	}
    }
    /* Fge eine Leerzeile ein, um den Mail-Header vom eigentlichen Text
    ** zu trennen.
    */
    fputs ("\n", tfp);
    /* Nachdem der Mail-Header erzeugt bzw. bertragen wurde, geschieht dies
    ** nun mit dem eigentlichen Text
    */
    while (fgets (line, 1024, rmsg)) {
	if (*line == '.') { (void) putc ('.', tfp); }
	while (!tr_line (rmsg, tfp, line, hostname)) {
	    /* Wenn noch kein Zeilenende gefunden wurde, jedoch keine Daten
	    ** mehr vorhanden sind, dann mu ein abschlieendes Zeilenende
	    ** eingefgt und die Schleife sofort beendet werden.
	    */
	    if (!fgets (line, 1024, rmsg)) { fputs ("\n", tfp); goto TO_EOD; }
	}
    }
TO_EOD: ;
    /* Den Pufferinhalt - falls noch vorhanden - in die Zieldatei schreiben */
    fflush (tfp);
    /* Die Lnge der Zieldatei bestimmen */
    if (fstat (fileno(rmsg), &sb) >= 0) {
	/* Eigentlich sollte dieser Fall immer eintreten. Da dies jedoch
	** generell nicht ausgeschlossen werden kann und ich es hasse,
	** stundenlang nach `segmentation fault's zu suchen, frage ich
	** einen eventuellen Fehler trotzdem ab!
	*/
	rmsg_size = sb.st_size;
	rdfile = gdupl (__FILE__, __LINE__, tfbuf);
    } else {
	/* Unix erlaubt das Lschen einer Datei auch wenn sie noch offen ist.
	** Da dieses Tool mit groer Wahrscheinlichkeit nur unter Unix (oder
	** eben Linux) eingesetzt werden wird (eine reine POSIX-Implemen-
	** tierung ist nicht mglich - unter anderem wegen des unter POSIX
	** fehlenden `seteuid'-Aufrufs), werde ich hier mit Sicherheit
	** keinerlei nderungen vornehmen.
	*/
	unlink (tfbuf);
    }
    /* Zum Schlu mssen Quell- und Zieldatei noch geschlossen werden */
    fclose (rmsg);
    fclose (tfp);
    add_exithl ("redirect", remove_rdfile);
}


static int init_connection (char *newhost)
{
    char line[1024];
    int client_in, client_out;
    HLSock hlc;
    sprintf (line, "%d", clientport);
    hlc = HLConnect ("tcp", newhost, line);
    if (!hlc) { return 1; }
    client_out = dup (HLSockSD (hlc));
    client_in = dup (HLSockSD (hlc));
    ServSockname = HLSockname (hlc);
    HLClose (hlc);
    if (!(out = fdopen (client_out, "w"))) {
	close (client_out); close (client_in);
	return 1;
    }
    if (!(in = fdopen (client_in, "r"))) {
	close (client_in); fclose (out);
	return 1;
    }
    if (get_pop (line, 1024, in)) {
	char *tsb, *tse;
	if ((tsb = strchr (line, '<')) && (tse = strchr (tsb, '>'))) {
	    ++tse;
	    redirect_timestamp = galloc (__FILE__, __LINE__, (tse - tsb) + 1);
	    memcpy (redirect_timestamp, tsb, (tse - tsb));
	    redirect_timestamp[tse - tsb] = '\0';
	}
	while (!has_eol (line)) {
	    if (!pop_gline (line, 1024, in)) { break; }
	}
	vxlog (2, "done.\n");
	return 0;
    }
    vxlog (2, "failed.\n");
    fclose (out); fclose (in);
    return 1;
}

static int do_apop (int nargs, char *argv[]); /* DONE */
static int do_dele (int nargs, char *argv[]); /* DONE */
static int do_last (int nargs, char *argv[]); /* DONE */
static int do_list (int nargs, char *argv[]); /* DONE */
static int do_noop (int nargs, char *argv[]); /* DONE */
static int do_pass (int nargs, char *argv[]); /* DONE */
static int do_quit (int nargs, char *argv[]); /* DONE */
static int do_retr (int nargs, char *argv[]); /* DONE */
static int do_rset (int nargs, char *argv[]); /* DONE */
static int do_stat (int nargs, char *argv[]); /* DONE */
static int do_top  (int nargs, char *argv[]); /* DONE */
static int do_uidl (int nargs, char *argv[]); /* DONE */
static int do_user (int nargs, char *argv[]); /* DONE */
static int do_vxts (int nargs, char *argv[]); /* DONE */

/* Hex-Kodierung einer Zeichenkette */
static void hex_encode (char *in, size_t ll, char *out)
{
    char *hexstr = "0123456789abcdef";
    int hd, ld;
    while (ll-- > 0) {
	hd = *in++;
	ld = hd & 0x0f;
	hd = (hd >> 4) & 0x0f;
	*out++ = hexstr[hd];
	*out++ = hexstr[ld];
    }
    *out = '\0';
}

/* Ich bentige eine zustzliche Funktion, um APOP zu ermglichen; diese
** Funktion soll ein `VXTS'-Kommando an den Zielrechner absetzen; dieses
** Kommando beherrscht jedoch nur ein `vxpop'-artiger Server.
** Was soll `VXTS' nun eigentlich tun?
**   Die Aufgabe eines `VXTS'-Kommandos ist es, den POP3-Server auf der
**   anderen Seite darum zu bitten, den eigenen Zeitstempel zu bernehmen.
**/
int send_vxts (dbif_rec *domrec)
{
    char *coded, *hex, *key, line[1024];
    unsigned char sbox[256];
    size_t msgl;
    if (!redirect_timestamp) {
	vxlog (1, "   the remote server doesn't have the APOP capability!\n");
	return 1;
    }
    /* Der Schlssel, mit dem die Nachricht verschlsselt werden soll */
    key = gconc (__FILE__, __LINE__, domrec->passwd, redirect_timestamp, NULL);
    /* Die zu verschlsselnde Nachricht */
    coded = gconc (__FILE__, __LINE__, ServSockname, apop_timestamp, NULL);
    msgl = strlen (coded);
    hex = galloc (__FILE__, __LINE__, 2 * msgl + 1);
    rc4_init ((unsigned char *) key, sbox);
    rc4_encrypt ((unsigned char *) coded, strlen (coded), sbox);
    hex_encode (coded, msgl, hex);
    rd_send ("VXTS %s", hex);
    if (get_pop (line, 1024, in)) {
	while (!has_eol (line)) {
	    if (!pop_gline (line, 1024, in)) { break; }
	}
	vxts_succeeded = 1;
	vxlog (2, "   vxpop compatible remote server - APOP is possible!\n");
	gfree (hex); gfree (coded); gfree (key);
	return 0;
    } else {
	vxts_succeeded = 0;
	vxlog (1, "   the remote server doesn't have the VXTS capability!\n");
	gfree (hex); gfree (coded); gfree (key);
	return 1;
    }
}

int redirect_user (int nargs, char *argv[])
{
    return do_user (nargs, argv);
}

int redirect_apop (int nargs, char *argv[])
{
    return do_apop (nargs, argv);
}

/* Die folgende Liste dient lediglich der Zuordnung der Kommandos zu den
** richtigen Pltzen in der Kommandotabelle des Interpreters.
*/
static struct {
    char *cmd;
    cmdfunc func;
} newcmds[] = {
    { "APOP", do_apop },
    { "DELE", do_dele },
    { "LAST", do_last },
    { "LIST", do_list },
    { "NOOP", do_noop },
    { "PASS", do_pass },
    { "QUIT", do_quit },
    { "RETR", do_retr },
    { "RSET", do_rset },
    { "STAT", do_stat },
    { "TOP",  do_top  },
    { "UIDL", do_uidl },
    { "USER", do_user },
    { "VXTS", do_vxts },
    { (char *) 0, (cmdfunc) 0 }
};

int redirect_setup (char *hostname, char *user, struct cstruct *cmdtbl)
{
    int ix, iy;
    char *ccmd;
    cmdfunc cfunc;
    vxlog (2, "   init_connection (\"%s\") ... ", hostname);
    if (init_connection (hostname)) { return 1; }
    vxlog (2, "   init_file (\"%s\") ... ", hostname);
    init_file (hostname, user);
    vxlog (2, "done.\n");
    vxlog (2, "   setting up a new command set ... ");
    for (ix = 0; newcmds[ix].cmd; ++ix) {
	ccmd = newcmds[ix].cmd;
	cfunc = newcmds[ix].func;
	for (iy = 0; cmdtbl[iy].cmdname; ++iy) {
	    if (!strcmp (cmdtbl[iy].cmdname, ccmd)) {
		vxlog (5, "   NEW function for `%s'\n", cmdtbl[iy].cmdname);
		cmdtbl[iy].action = cfunc;
		break;
	    }
	}
    }
    vxlog (2, "done!\n");
    return 0;
}

/* Hier beginnen Implementationen der einzelnen Kommandos.
** Mit Ausnahme von STAT, LIST, RETR, TOP und DELE mssen alle Funktionen
** die Kommandos, die sie implementieren in Textform an den Zielrechner
** bertragen und die Ergebnisse von diesem an den Benutzer weiterleiten.
*/

static int do_stat (int nargs, char *argv[])
{
    char line[1024], ok_msg[1024];
    /* Was mu ich hier tun?
    ** Ich mu das STAT-Kommando an den Zielrechner schicken;
    ** ich mu anschlieend auf Antwort warten; kommt eine Antwort
    ** zurck, dann mu ich diese interpretieren; ist sie positiv,
    ** dann mu ich die zwei Zahlen die mit der Antwort kommen auslesen
    ** und um 1 (eine zustzliche Meldung) bzw. die Lnge der zustzlichen
    ** Meldung erhhen und die so vernderte Erfolgsmeldung an den Benutzer
    ** weiterversenden. Letzteres darf aber nur dann geschehen, wenn das auch
    ** so konfiguriert wurde, d.h. wenn die temporr erzeugte Mail-Datei
    ** existiert.
    */
    rd_send ("STAT");
    if (get_pop (line, 1024, in)) {
	char *p, *q;
	strcpy (ok_msg, line);
	cut_eol (ok_msg);
	/* Skip a possible rest of the line */
	while (!has_eol (line)) {
	    if (!pop_gline (line, 1024, in)) { break; }
	}
	if (rdfile) {
	    int l1 = 0;
	    off_t l2 = 0;
	    l1 = (int) strtol (ok_msg, &p, 10);
	    if (p == ok_msg || *p++ != ' ') {
		pop_msg (ERR, "invalid non-numeric first result from STAT");
		return 1;
	    }
	    l2 = (off_t) strtol (p, &q, 10);
	    if (q == p || *q) {
		pop_msg (ERR, "invalid non-numeric second result from STAT");
		return 1;
	    }
	    if (rdmsg_deleted) {
		pop_msg (OK, "%d %ld", l1, l2);
	    } else {
		pop_msg (OK, "%d %ld", l1 + 1, l2 + rmsg_size);
	    }
	} else {
	    pop_sstrg ("+OK ", 0);
	    pop_sline (line, 0);
	}
    } else {
	pop_sstrg ("-ERR ", 0);
	pop_sline (line, 0);
	while (!has_eol (line)) {
	    if (!pop_gline (line, 1024, in)) { break; }
	}
	return 1;
    }
    if (!lastix) { lastix = 1; }
    return 0;
}

static int do_retr (int nargs, char *argv[])
{
    FILE *rd;
    char *p, line[1024];
    int mn = (int) strtol (argv[0], &p, 10);
    /* if the conversion failed, write an error message and return! */
    if (p == argv[0] || *p) {
	pop_msg (ERR, "invalid numeric argument for RETR");
	return 1;
    }
    /* if a redirection file is present
     *     and the message number is 1, then send the redirection message;
     *     otherwise decrease the message number by one and retrieve the
     *     corresponding message from the remote host and send it untrans-
     *     lated to the user.
     * if no redirection message is present then
     *     retrieve the corresponding message from the remote host and send
     *     it untranslated to the user.
     * if the message request to the remote host leads to an error message
     * send from this host, then send this error untranslated to the user.
     */
    if (rdfile) {
	if (mn == 1) {
	    if (rdmsg_deleted) {
		pop_msg (ERR, "no such message");
		return 1;
	    }
	    /* send the redirection message in this case */
	    if (!(rd = fopen (rdfile, "r"))) {
		pop_msg (ERR, "transmission failed - internal error");
		return 1;
	    }
	    /* send a positive response message */
	    pop_msg (OK, "message %d follows.", 1);
	    /* append the message */
	    while (fgets (line, 1024, rd)) {
		pop_sstrg (line, 0);
	    }
	    /* terminate the message with `.', followed by CR/LF */
	    pop_msg (ML, ".");
	    fclose (rd);
	    if (lastix < mn) { lastix = mn; }
	    return 0;
	} else {
	    rd_send ("RETR %d", --mn);
	}
    } else {
	rd_send ("RETR %d", mn);
    }
    if (get_pop (line, 1024, in)) {
	pop_msg (OK, "message %d follows.", mn);
	while (!has_eol (line) && pop_gline (line, 1024, in));
	while (pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	    if (!strcmp (line, ".\r\n")) { break; }
	}
	if (lastix < mn) { lastix = mn; }
	return 0;
    } else {
	pop_msg (ERR, "no such message");
	while (!has_eol (line) && pop_gline (line, 1024, in));
	return 1;
    }
}

static int do_top (int nargs, char *argv[])
{
    FILE *rd;
    char *p, *q, line[1024];
    int mn = (int) strtol(argv[0], &p, 10), nl = (int) strtol(argv[1], &q, 10);
    /* if the conversion failed, write an error message and return! */
    if (p == argv[0] || q == argv[1] || *p || *q) {
	pop_msg (ERR, "invalid numeric argument(s) for TOP");
	return 1;
    }
    /* if a redirection file is present
     *     and the message number is 1, then send the redirection message;
     *     otherwise decrease the message number by one and retrieve the
     *     corresponding message from the remote host and send it untrans-
     *     lated to the user.
     * if no redirection message is present then
     *     retrieve the corresponding message from the remote host and send
     *     it untranslated to the user.
     * if the message request to the remote host leads to an error message
     * send from this host, then send this error untranslated to the user.
     */
    if (rdfile) {
	if (mn == 1) {
	    if (rdmsg_deleted) {
		pop_msg (ERR, "no such message");
		return 1;
	    }
	    /* send the redirection message in this case */
	    if (!(rd = fopen (rdfile, "r"))) {
		pop_msg (ERR, "transmission failed - internal error");
		return 1;
	    }
	    /* send a positive response message */
	    pop_msg (OK, "message %d's header and %d line%s of body follow.",
			 1, nl, (nl == 1 ? "" : "s"));
	    /* append the message header */
	    while (fgets (line, 1024, rd) && strcmp (line, "\n")) {
		pop_sstrg (line, 0);
	    }
	    /* append `nl' lines of the message body */
	    if (nl-- > 0) {
		pop_sstrg (line, 0);
		while (fgets (line, 1024, rd) && nl > 0) {
		    pop_sstrg (line, 0);
		    if (has_eol (line)) { --nl; }
		}
	    }
	    /* terminate the message with `.', followed by CR/LF */
	    pop_msg (ML, ".");
	    fclose (rd);
	    if (lastix < mn) { lastix = mn; }
	    return 0;
	} else {
	    rd_send ("TOP %d %d", --mn, nl);
	}
    } else {
	rd_send ("TOP %d %d", mn, nl);
    }
    if (get_pop (line, 1024, in)) {
	pop_msg (OK, "message %d's header and %d line%s of body follow.",
		     mn, nl, (nl == 1 ? "" : "s"));
	while (!has_eol (line) && pop_gline (line, 1024, in));
	while (pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	    if (!strcmp (line, ".\r\n")) { break; }
	}
	if (lastix < mn) { lastix = mn; }
	return 0;
    } else {
	pop_msg (ERR, "no such message");
	pop_sline (line, 0);
	while (!has_eol (line) && pop_gline (line, 1024, in));
	return 1;
    }
}

/* DELE command */
static int do_dele (int nargs, char *argv[])
{
    char *p, line[1024];
    int mn = (int) strtol (argv[0], &p, 10);
    if (p == argv[0] || *p) {
	pop_msg (ERR, "invalid numeric argument for DELE!");
	return 1;
    }
    if (rdfile) {
	if (mn == 1) {
	    if (rdmsg_deleted) { pop_msg (ERR, "no such message"); return 1; }
	    rdmsg_deleted = 1;
	    pop_msg (OK, "redirection message deleted!");
	    if (lastix < mn) { lastix = mn; }
	    return 0;
	}
	rd_send ("DELE %d", mn - 1);
    } else {
	rd_send ("DELE %d", mn);
    }
    if (get_pop (line, 1024, in)) {
	pop_sstrg ("+OK ", 0); pop_sstrg (line, 0);
	while (!has_eol (line) && pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	}
	if (lastix < mn) { lastix = mn; }
	return 0;
    } else {
	pop_msg (ERR, "no such message");
	while (!has_eol (line) && pop_gline (line, 1024, in));
	return 1;
    }
}

/* LAST command - only for backward compatibility with RFC1081, ..., RFC1460 */
static int do_last (int nargs, char *argv[])
{
    /* Ob der POP3-Server am Ziel der Umleitung das LAST-Kommando beherrscht,
    ** kann mir eigentlich egal sein; schlielich mu ich die Zhlung trotzdem
    ** durchfhren; dann kann ich auch gleich das LAST-Kommando auf dieser
    ** Seite (sozusagen virtuell) Implementieren, ohne den POP3-Server auf der
    ** anderen Seite irgendwie weiter damit zu belasten.
    */
    return lastix;
}

/* LIST command - it has an optional argument */
static int do_list (int nargs, char *argv[])
{
    int mn;
    off_t ms;
    char *p, *q, line[1024], cmdbuf[128];
    if (nargs > 0) {
	mn = (int) strtol (argv[0], &p, 10);
	if (p == argv[0] || *p) {
	    pop_msg (ERR, "invalid non-numeric argument for LIST");
	    return 1;
	}
	if (mn == 1) {
	    if (rdmsg_deleted) { pop_msg (ERR, "no such message"); return 1; }
	    pop_msg (OK, "1 %ld", rmsg_size);
	    if (lastix < mn) { lastix = mn; }
	    return 0;
	}
	rd_send ("LIST %d", mn - 1);
	if (get_pop (line, 1024, in)) {
	    pop_sstrg ("+OK ", 0); pop_sstrg (line, 0);
	    while (!has_eol (line) && pop_gline (line, 1024, in)) {
		pop_sstrg (line, 0);
	    }
	    if (lastix < mn) { lastix = mn; }
	    return 0;
	} else {
	    pop_msg (ERR, "no such message");
	    while (!has_eol && pop_gline (line, 1024, in));
	    return 1;
	}
    }
    /* Hmmm ... a LIST command returns an error if there are no messages
    ** in the folder; i must circumvent this, because i may have the
    ** redirection message (virtually) in the folder.
    */
    rd_send ("STAT");
    if (get_pop (line, 1024, in)) {
	memcpy (cmdbuf, line, 127); cmdbuf[127] = '\0';
	cut_eol (cmdbuf);
	while (!has_eol (line) && pop_gline (line, 1024, in));
	mn = (int) strtol (cmdbuf, &p, 10);
	if (p == line || *p != ' ') {
	    pop_msg (ERR, "LIST: invalid non-numeric first result from STAT");
	    return 1;
	}
	++p;
	ms = (off_t) strtol (p, &q, 10);
	if (q == p || *q) {
	    pop_msg (ERR, "LIST: invalid non-numeric second result from STAT");
	    return 1;
	}
	if (rdfile && !rdmsg_deleted) { mn += 1; ms += rmsg_size; }
    } else {
	pop_sstrg ("-ERR ", 0);
	pop_sline (line, 0);
	while (!has_eol (line) && pop_gline (line, 1024, in));
	return 1;
    }
    /* If i don't have any message, i should print an error message and
    ** return immediately
    */
    if (mn == 0) { pop_msg (ERR, "no messages found"); return 1; }
    /* I will print the OK message deferred - in the case that the LIST
    ** command itself fails.
    **/
    sprintf (cmdbuf, "%d message%s (%ld octet%s)",
		     mn, (mn == 1 ? "" : "s"),
		     ms, (ms == 1 ? "" : "s"));
    /* sending a LIST command to the target host would fail if there is
    ** only the redirection message, so i will catch this case and send only
    ** data for this message.
    */

    if (mn == 1 && rdfile && !rdmsg_deleted) {
	pop_msg (OK, cmdbuf);
	pop_msg (ML, "%d %ld", 1, rmsg_size);
	pop_msg (ML, ".");
	if (lastix < 1) { lastix = 1; }
	return 0;
    }
    /* the LIST command should succeed now */
    rd_send ("LIST");
    if (get_pop (line, 1024, in)) {
	/* if the list command itself succeeded, i will replace the OK
	** message by my own - which i stored in `cmdbuf'
	*/
	while (!has_eol (line) && pop_gline (line, 1024, in));
	pop_msg (OK, cmdbuf);
	if (rdfile) {
	    if (!rdmsg_deleted) { pop_msg (ML, "%d %ld", 1, rmsg_size); }
	    while (pop_gline (line, 1024, in)) {
		if (!strcmp (line, ".\r\n")) { break; }
		mn = (int) strtol (line, &p, 10);
		++p; ms = strtol (p, &q, 10);
		pop_msg (ML, "%d %ld", mn + 1, ms);
		while (!has_eol (line) && pop_gline (line, 1024, in));
	    }
	    pop_msg (ML, ".");
	    if (lastix < 1) { lastix = 1; }
	    return 0;
	}
	while (pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	    if (!strcmp (line, ".\r\n")) { break; }
	}
	if (lastix < 1) { lastix = 1; }
	return 0;
    }
    pop_sstrg ("-ERR ", 0);
    pop_sline (line, 0);
    while (!has_eol (line) && pop_gline (line, 1024, in));
    return 1;
}

/* QUIT command */
static int do_quit (int nargs, char *argv[])
{
    char line[1024];
    rd_send ("QUIT");
    get_pop (line, 1024, in);
    if (nargs < 1) { pop_msg (OK, "closing connection"); }
    /* close the redirected connection */
    vxlog (5, "   rQUIT - closing redirected connection.\n");
    fclose (out); fclose (in);
    vxlog (5, "   rQUIT - done.\n");
    /* close the normal connection */
    vxlog (5, "   rQUIT - closing user connection.\n");
    fclose (stdout); fclose (stdin);
    vxlog (5, "   rQUIT - done.\n");
    /* NO ERRORS */
    return 0;
}

/* PASS command */
static int do_pass (int nargs, char *argv[])
{
    /* Hmm, APOP sollte - ebenso wie PASS noch ein STAT-Kommando
    ** hinterherschicken, um die Anzahl und die Gesamtlnge aller Mails
    ** zu bekommen; ich meine, da zwar die Texte hinter einer positiven
    ** Antwort auf ein APOP- oder PASS-Kommando nur einen informationellen
    ** Wert haben (siehe RFC1939); es wre jedoch trotzdem schner, die
    ** richtigen Antworten (auch im Falle einer existierenden virtuell einge-
    ** fgten Mail - die Umleitung betreffend) zu bekommen.
    ** Vielleicht ist das ganze aber auch zuviel Aufwand - schlielich mu
    ** ich auerdem ja noch die Implementation des LAST-Kommandos bedenken,
    ** die einen etwas vorsichtigeren Umgang mit Transaktionskommandos erfor-
    ** dert.
    */
    char line[1024];
    int stat;
    rd_send ("PASS %s", argv[0]);
    stat = get_pop (line, 1024, in);
    pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
    pop_sstrg (line, 0);
    while (!has_eol (line) && pop_gline (line, 1024, in)) {
	pop_sstrg (line, 0);
    }
    return (stat ? 0 : 1);
}

/* NOOP command */
static int do_noop (int nargs, char *argv[])
{
    char line[1024];
    int stat;
    rd_send ("NOOP");
    stat = get_pop (line, 1024, in);
    pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
    pop_sstrg (line, 0);
    while (!has_eol (line) && pop_gline (line, 1024, in)) {
	pop_sstrg (line, 0);
    }
    return (stat ? 0 : 1);
}

/* RSET command */
static int do_rset (int nargs, char *argv[])
{
    char line[1024];
    int stat;
    rd_send ("RSET");
    stat = get_pop (line, 1024, in);
    if (nargs < 1) {
	pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
	pop_sstrg (line, 0);
    }
    while (!has_eol (line) && pop_gline (line, 1024, in)) {
	if (nargs < 1) pop_sstrg (line, 0);
    }
    if (stat) {
	if (rdfile) { rdmsg_deleted = 0; }
	lastix = 0;
    }
    return (stat ? 0 : 1);
}

/* VXTS command */
static int do_vxts (int nargs, char *argv[])
{
    pop_msg (ERR, "VXTS command is illegal in redirection mode!");
    return 1;
}

/* USER command - allowed in the redirection mode */
static int do_user (int nargs, char *argv[])
{
    char line[1024];
    int stat;
    rd_send ("USER %s", argv[0]);
    stat = get_pop (line, 1024, in);
    pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
    pop_sstrg (line, 0);
    while (!has_eol (line) && pop_gline (line, 1024, in)) {
	pop_sstrg (line, 0);
    }
    return (stat ? 0 : 1);
}

/* APOP command - allowed only if the redirect time stamp was set and a
**                preceding VXTS command was successful
*/
static int do_apop (int nargs, char *argv[])
{
    /* Hmm, APOP sollte - ebenso wie PASS noch ein STAT-Kommando
    ** hinterherschicken, um die Anzahl und die Gesamtlnge aller Mails
    ** zu bekommen; ich meine, da zwar die Texte hinter einer positiven
    ** Antwort auf ein APOP- oder PASS-Kommando nur einen informationellen
    ** Wert haben (siehe RFC1939); es wre jedoch trotzdem schner, die
    ** richtigen Antworten (auch im Falle einer existierenden virtuell einge-
    ** fgten Mail - die Umleitung betreffend) zu bekommen.
    ** Vielleicht ist das ganze aber auch zuviel Aufwand - schlielich mu
    ** ich auerdem ja noch die Implementation des LAST-Kommandos bedenken,
    ** die einen etwas vorsichtigeren Umgang mit Transaktionskommandos erfor-
    ** dert.
    */
    char line[1024];
    int stat;
    if (redirect_timestamp && vxts_succeeded) {
	rd_send ("APOP %s %s", argv[0], argv[1]);
	stat = get_pop (line, 1024, in);
	pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
	pop_sstrg (line, 0);
	while (!has_eol (line) && pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	}
	return (stat ? 0 : 1);
    }
    if (redirect_timestamp) {
	pop_msg (ERR, "APOP failed - VXTS command failed!");
    } else {
	pop_msg (ERR, "APOP failed - the redirected server doesn't have APOP "
		      "capabilities!");
    }
    return 1;
}

/* UIDL - Kommando
 *     Wie mache ich das?
 *
*/
static int do_uidl (int nargs, char *argv[])
{
    char line[1024], *p, uidl[33], sbuf[64];
    int stat, mn;
    if (nargs >= 1) {
	mn = (int) strtol (argv[0], &p, 10);
	if (p == argv[0] || *p) {
	    pop_msg (ERR, "invalid non-numeric argument for UIDL");
	    return 1;
	}
	if (rdfile) {
	    if (rdmsg_deleted) { pop_msg (ERR, "no such message"); return 1; }
	    if (mn == 1) {
		rcalc_uidl (rdfile, uidl, 33);
		pop_msg (OK, "%d %s", mn, uidl);
		if (lastix < mn) { lastix = mn; }
		return 0;
	    }
	    rd_send ("UIDL %d", mn - 1);
	} else {
	    rd_send ("UIDL %d", mn);
	}
	stat = get_pop (line, 1024, in);
	pop_sstrg ((stat ? "+OK " : "-ERR "), 0);
	pop_sstrg (line, 0);
	while (!has_eol (line) && pop_gline (line, 1024, in)) {
	    pop_sstrg (line, 0);
	}
	if (stat && lastix < mn) { lastix = mn; }
	return (stat ? 0 : 1);
    }
    /* Was mu ich hier machen?
    ** Ein UIDL-Kommando schlgt ja auch dann fehl, wenn auf der anderen Seite
    ** keine Mails vorhanden sind. Wenn ich jedoch meine virtuelle Mail
    ** eingeschoben habe, dann darf dies nicht geschehen. In diesem Fall
    ** mu(!) ich also ein STAT-Kommando vorausschicken, welches mir die
    ** Anzahl der verfgbaren E-Mails zurckgibt.
    */
    if (rdfile) {
	rd_send ("STAT");
	stat = get_pop (line, 1024, in);
	if (stat) { memcpy (sbuf, line, 63); sbuf[63] = '\0'; }
	while (!has_eol (line) && pop_gline (line, 1024, in));
	if (!stat) {
	    pop_msg (ERR, "redirected STAT command failed");
	    return 1;
	}
	mn = strtol (sbuf, &p, 10);
	if (p == sbuf || *p != ' ') {
	    pop_msg (ERR, "non-numeric STAT-response from redirection");
	    return 1;
	}
	/* Wenn keine Meldungen auf der anderen Seite vorhanden sind, dann
	** darf ich kein UIDL-Kommando senden.
	*/
	if (!mn) {
	    if (rdfile && !rdmsg_deleted) {
		pop_msg (OK, "UIDL command accepted");
		rcalc_uidl (rdfile, uidl, 33);
		pop_msg (ML, "%d %s", 1, uidl);
		if (lastix < 1) { lastix = 1; }
		return 0;
	    } else {
		pop_msg (ERR, "no messages found");
		return 1;
	    }
	}
    }
    /* Sonst sende ich das UIDL-Kommando einfach; bei einer positiven Antwort
    ** wird auch die (Pseudo-)UIDL einer eventuell vorhandenen virtuellen
    ** E-Mail berechnet und ausgegeben.
    */
    rd_send ("UIDL");
    if (get_pop (line, 1024, in)) {
	/* Gib nur bei einer positiven Antwort die UIDL der virtuellen E-Mail
	** aus. Diese Vorgehensweise ist notwendig! Wenn ich nmlich zuerst
	** eine positive Antwort ausgebe und hinterher von der anderen Seite
	** eine Fehlermeldung, dann kann ich schlecht die bereits begonnene
	** Ausgabe der Meldungen unterbrechen, um die Fehlermeldung von der
	** anderen Seite auszugeben.
	*/
	if (rdfile && !rdmsg_deleted) {
	    pop_msg (OK, "UIDL command accepted");
	    if (rcalc_uidl (rdfile, uidl, 33)) {
		vxlog (1, "calculating UIDL failed for virtually inserted "
			  "message!\n");
	    }
	    pop_msg (ML, "%d %s", 1, uidl);
	}
	/* Lies immer die Zeile bis zum Zeilenende! */
	while (!has_eol (line) && pop_gline (line, 1024, in));
	/* Ich beende meine Ausgabe, wenn
	** - entweder kein Text mehr vorhanden ist (Verbindung
	**   zusammengebrochen)
	** - oder die Abschluzeile von der anderen Seite empfangen wurde.
	** In diesem Fall kann ich brigens davon ausgehen, da diese
	** Abschluzeile genau das Format hat wie das zweite Argument der
	** `strcmp'-Funktion; schlielich schreibt das POP3-Protokoll genau
	** dies so vor.
	*/
	while ((p = pop_gline (line, 1024, in)) && strcmp (line, ".\r\n")) {
	    mn = (int) strtol (line, &p, 10);
	    /* `pop_msg' hngt blicherweise ein CR/LF an die auszugebende
	    ** Zeile an. Da jedoch die empfangene Zeile bereits ein CR/LF
	    ** enthlt, darf dies nicht geschehen; daher habe ich `pop_msg' so
	    ** erweitert, da bei bergabe von `NOEOL' als erstem Argument
	    ** kein CR/LF zustzlich angehngt wird.
	    ** Wenn eine virtuelle E-Mail vorhanden war, dann mssen die
	    ** E-Mail-Nummern aus dem Antwortprotokoll des Servers auf der
	    ** anderen Seite immer um eins erhht werden; sonst bleiben diese
	    ** unverndert.
	    */
	    pop_msg (NOEOL, "%d%s", mn + (rdfile ? 1 : 0), p);
	    /* Lies immer die Zeile bis zum Zeilenende! */
	    while (!has_eol (line) && pop_gline (line, 1024, in));
	}
	/* Gib den obligatorischen Punkt (+ Zeilenende) zum Abschlu einer
	** mehrzeiligen Antwort aus.
	*/
	pop_msg (ML, ".");
	if (lastix < 1) { lastix = 1; }
	return 0;
    }
    /* Diese Stelle wird nur im Fehlerfall erreicht. Dann ist nmlich ein
    ** UIDL-Kommando auf der anderen Seite unbekannt.
    */
    pop_sstrg (line, 0);
    while (!has_eol (line) && pop_gline (line, 1024, in)) {
	pop_sstrg (line, 0);
    }
    return 1;
}
