/* uf.c
**
** file handle for the user folder
**
*/

#define tmpfile __XXX_tmpfile
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "vxerror.h"
/*#include "config.h"*/
#include "tools.h"
#include "prefs.h"
#include "conv.h"
#include "exithl.h"

#define _UF_C
#include "uf.h"
#include "popmsg.h"
#include "bsmtp.h"
#include "vxhost.h"
#undef tmpfile

struct msg_s *msgtbl = (struct msg_s *) 0;
xFile uf_folder = (xFile) 0;
char *uf_foldername = (char *) 0;

static char *tmpfile = (char *) 0;
static int tempix = 0;

/* uf_update ()
**     entfernt alle als gelscht markierten Mails aus dem Folder des
**     Benutzers - dem im folgenden als Zielfolder bezeichneten Folder.
**     Da ich hierzu `file locking' dringend bentige, darf ich die
**     ursprngliche Folderdatei nicht schlieen oder lschen; ich mu also
**     folgenden Weg gehen:
**       Ohne den Zielfolder zu schlieen, kopiere ich alle nicht als
**       gelscht markierten Mails aus diesem in einen temporren Folder.
**       Anschlieend ffne ich den temporren Folder zum lesen neu und
**       den Zielfolder mit `u'-Modus - aber mit einem weiteren Dateihandle;
**       letzteres bewirkt ein lschen smtlicher im Folder enthaltener
**       Daten.
**       Nun kopiere ich smtliche Mails aus dem temporren Folder zurck in
**       den Zielfolder.
**       Zum Schlu schliee ich das zustzliche Dateihandle auf den
**       Zielfolder, schliee den temporren Folder und lsche letzteren
**       anschlieend.
*/
void uf_update (void)
{
    xFile tmpf, df;
    ssize_t wlen;
    char *line[1024];
    int ix, no_update = 1;
    /* berprfe zunchst einmal, ob berhaupt Messages gelscht werden
    ** mssen.
    */
    for (ix = 0; msgtbl[ix].length >= 0; ++ix) {
	if (msgtbl[ix].deleted) { no_update = 0; }
    }
    /* Wenn nmlich nicht, dann sind die ganzen Kopieroperationen ohne einen
    ** gesteigerten Nutzen und wir ersparen sie uns!
    */
    if (no_update) {
	if (tempix) { xNoIndex (uf_foldername); }
	return;
    }
    if (!(tmpf = xOpen (tmpfile, "w", 0600))) {
	pop_msg (ERR, "i couldn't create the temporary folder for updating");
	return;
    }
    /* Kopiere zunchst alle nicht als gelscht markierten Mails
    ** in einen temporren Folder.
    */
    for (ix = 0; msgtbl[ix].length >= 0; ++ix) {
	if (msgtbl[ix].deleted) { continue; }
	xSeek (uf_folder, ix);
	while ((wlen = xRead (uf_folder, line, 1024)) > 0) {
	    xWrite (tmpf, line, (size_t) wlen);
	}
	xSet (tmpf);
    }
    xClose (tmpf);
    /* ffne den temporren Folder neu zum lesen */
    if (!(tmpf = xOpen (tmpfile, "r"))) {
	pop_msg (ERR, "temporary folder not found");
	return;
    }
    /* Erzeuge den Zielfolder neu, d.h. entleere ihn */
    if (!(df = xOpen (uf_foldername, "u", 0600))) {
	xClose (tmpf);
	xUnlink (tmpfile);
	pop_msg (ERR, "replacing folder contents failed");
	return;
    }
    /* Kopiere nacheinander alle Mails aus dem temporren Folder in den
     * Zielfolder
     */
    for (ix = 0; xSeek (tmpf, ix) >= 0; ++ix) {
	while ((wlen = xRead (tmpf, line, 1024)) > 0) {
	    xWrite (df, line, (size_t) wlen);
	}
	xSet (df);
    }
    /* Schliee temporren und Zielfolder */
    xClose (df);
    xClose (tmpf);
    /* Entferne den temporren Folder */
    xUnlink (tmpfile);
    /* Wenn die Indexdatei nur temporr angelegt wurde, dann mu sie nun
    ** wieder entfernt werden!
    */
    if (tempix) { xNoIndex (uf_foldername); }
}

void uf_abort (void)
{
    vxlog (5, "   removing index file (if required!)\n");
    if (uf_foldername && uf_folder && tempix) {
	xNoIndex (uf_foldername);
	xClose (uf_folder);
	uf_folder = (xFile) 0;
    }
    vxlog (5, "   index file removed!\n");
}

/* Skip the `From ' line at the top of the message */
ssize_t From_Skip (xFile fx, int ix)
{
    char line[1024];
    size_t ll, res;
    xSeek (fx, ix);
    if ((ll = xGets (line, 1024, fx)) > 0) {
	if (!strncmp (line, "From ", 5)) {
	    res = ll;
	    while (line[ll - 1] != '\n' && line[ll - 1] != '\r') {
		if ((ll = xGets (line, 1024, fx)) <= 0) { break; }
		res += ll;
	    }
	    return res;
	}
    }
    xReSet (fx);
    return 0;
}

/* Initiate the uploading of all messages in a RT_BSMTP-folder via SMTP to
** the host determined by the POP3-connection.
*/
static void uf_smtp (char *batchfolder)
{
    pid_t chld;
    switch ((chld = fork ())) {
	case -1: /* fork failed! */
	    vxlog (0, "POP3-triggered SMTP failed: %s\n", strerror (errno));
	    return;
	case  0: /* child process */
	    clean_exithl ();
	    if (bsmtp (RADDRstr (), batchfolder, mailhost ())) {
		vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1, "(bsmtp)");
	    }
	    vxlog (1, "Exiting `uf_smtp'.\n");
	    exit (0);
	default: /* parent process */
	    return;
    }
}

/* Wann mu der setup stattfinden?
** - genau dann, wenn die Authentifizierung abgeschlossen ist, d.h. nach
**   einem `APOP'-Kommando oder nach einem `PASS'-Kommando.
** Zu diesem Zeitpunkt liegt dann der Datensatz des Benutzers bereits vor.
** 1. Erweiterung (26.4.1999, 3:02:15):
**    Der hier hineingehende Datensatz kann auch ein BSMTP-Eintrag sein;
**    in diesem Fall mu eine SMTP-bertragung zu dem Host stattfinden, dessen
**    POP3-Client gerade mit `vxpop' (ebendiese) Verbindung aufgenommen hat;
**    `vxpop' mu stattdessen mit einer leeren Liste von E-Mails zurckkehren,
**    ohne, da der BSMTP-Folder irgendwie angetastet wird.
**    `vxpop' startet zu diesem Zweck einen neuen Proze, der die Aufgabe der
**    SMTP-bertragung der im Folder stehenden E-Mails im Hintergrund
**    bernimmt, whrend `vxpop' selbst den Folder nicht anrhrt und
**    stattdessen mit einer leeren Liste zurckkehrt.
*/
int uf_setup (dbif_rec *rec)
{
    size_t fnlen = rec->flags - rec->folder, tmplen;
    size_t nmsgs = 0;
    ssize_t cxlen;
    int ix;
    char pidbuf[15];
    /* Check if this is only a trigger for a SMTP-upload of a BSMTP-entry. */
    if (dbif_rectype(rec) == RT_BSMTP) {
	uf_smtp (rec->folder);
	msgtbl = (struct msg_s *) galloc (__FILE__, __LINE__,
					  sizeof (struct msg_s));
	msgtbl[nmsgs].length = -1;
	vxlog (5, "   uf_smtp done, leaving uf_setup\n");
	return 0;
    }
    /* Setup the temporary folder name */
    tmplen = strlen (TMPDIR) + strlen (progname) + 20;
    tmpfile = galloc (__FILE__, __LINE__, tmplen);
    strcpy (tmpfile, TMPDIR);
    strcat (tmpfile, "/");
    strcat (tmpfile, progname);
    strcat (tmpfile, ".");
    sprintf (pidbuf, "%d", getpid ());
    strcat (tmpfile, pidbuf);
    /* setup the user's folder name */
    uf_foldername = galloc (__FILE__, __LINE__, fnlen + 1);
    memcpy (uf_foldername, rec->folder, fnlen);
    uf_foldername[fnlen] = '\0';
    /* So - hier also die zustzliche Arbeit:
    **     Wenn das Flag `nonstandard' im Benutzerdatensatz gesetzt ist,
    **     dann mu der Folder konvertiert werden.
    **     Auerdem mu - da die Indexdatei nur temporr angelegt werden soll -
    **     ein Flag gesetzt werden, da es `uf_update' ermglicht, am Ende der
    **     Update-Aktionen die Indexdatei wieder zu lschen.
    */
    tempix = !prefs.syswithidx && (dbif_rectype(rec) == RT_NSUSER);
    if (tempix) {
       	if (!(uf_folder = xOpen (uf_foldername, "ix", prefs.fperm))) {
	    if (errno == ENOENT) { goto MT_FOLDER; }
	    vxlog (1, "opening non-standard folder `%s' failed: %s (%d,%d)\n",
		      uf_foldername, strerror (errno), geteuid (), getgid ());
	    vxlog (1, " -- you should check the permissions of your "
		      "spool directory\n");
	    pop_msg (ERR, "opening nonstd folder failed!");
	    xNoIndex (uf_foldername);
	    gfree (uf_foldername);
	    return 1;
	}
	/* Setze den Puffer auf einen greren Wert, damit eine mglichst
	** schnelle Konvertierung der Datei stattfinden kann
	*/
	xSetBufSize (262144);
	vxlog (2, "preparing nonstandard folder\n");
	/* Folder ist im Indexmodus geffnet! Konvertiere ihn */
	conv (uf_folder);
	/* ndere den Modus des Folders von Index- auf Lesemodus */
	xReSet (uf_folder);
	vxlog (3, "register uf_abort as an exit-function!\n");
	add_exithl ("user-folder", uf_abort);
	vxlog (2, "nonstandard folder ready\n");
	/* Der Folder ist jetzt bereit! */
    } else if (!(uf_folder = xOpen (uf_foldername, "rx"))) {
	if (errno != ENOENT) {
	    vxlog (1, "accessing folder `%s' failed: %s\n",
		      uf_foldername, strerror (errno));
	    pop_msg (ERR, "opening user folder failed!");
	    gfree (uf_foldername);
	    gfree (tmpfile);
	    return 1;
	}
    MT_FOLDER:
	/* Der Folder konnte nicht geffnet werden. Versuche, ihn anzulegen! */
	vxlog (3, "trying to create folder `%s' with UID=%d, GID=%d.\n",
		  uf_foldername, getuid (), getgid ());
	if (!(uf_folder = xOpen (uf_foldername, "wx", prefs.fperm))) {
	    vxlog (1, "creating standard folder `%s' failed: %s\n",
		      uf_foldername, strerror (errno));
	    vxlog (1, " -- you should check the permissions of your "
		      "spool directory\n");
	    pop_msg (ERR, "creating user folder failed!");
	    gfree (uf_foldername);
	    gfree (tmpfile);
	    return 1;
	}
	/* Versetze den soeben geffneten Folder in den Lesemodus zurck */
	xReSet (uf_folder);
    }
    /* Nachdem das mit der Erffnung nun geklrt ist, gehen wir zur Bestimmung
    ** der Anzahl der Mails und ihrer einzelnen Lngen ber.
    */
    for (nmsgs = 0; (cxlen = xMsgLen (uf_folder, (int) nmsgs)) >= 0; ++nmsgs);
    msgtbl = (struct msg_s *) galloc (__FILE__, __LINE__,
				      (nmsgs + 1) * sizeof (struct msg_s));
    for (ix = 0; (cxlen = xMsgLen (uf_folder, ix)) >= 0; ++ix) {
	msgtbl[ix].length = ((ssize_t) cxlen) - From_Skip (uf_folder, ix);
	msgtbl[ix].deleted = 0;
    }
    msgtbl[nmsgs].length = -1;
    vxlog (5, "   leaving uf_setup\n");
    return 0;
}
