/* uidl.c
**
** (Pseudo-)UIDL calculation
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <time.h>
#include <errno.h>

#include "dbif.h"
#ifdef __STDC__
#define PROTOTYPES 1
#endif
#include "md5.h"
#include "tools.h"
#include "popmsg.h"
#include "cmi.h"
#include "xfile.h"

typedef unsigned char uchar;

void uidl_calc (xFile fx, int msgn, char *resbf, size_t ressz)
{
    MD5_CTX uidlcntxt;
    unsigned char md5sum[16];
    char line[1024], *hex = "0123456789abcdef";
    char outbuf[33], *outp;
    ssize_t llen;
    size_t rlen = 32;
    int ix;
    MD5Init (&uidlcntxt);
    xSeek (fx, msgn);
    vxlog (5, "   calculating MD5 sum of message header ");
    while ((llen = xGets (line, 1024, fx)) > 0) {
	MD5Update (&uidlcntxt, (uchar *) line, llen);
	vxlog (5, ".");
	if (*line == '\r' && (!line[1] || (line[1] == '\n' && !line[2]))) break;
	if (*line == '\n' && !line[1]) break;
    }
    MD5Final (md5sum, &uidlcntxt);
    vxlog (5, " done.\n");
    --ressz;
    if (ressz < rlen) rlen = ressz;
    outp = outbuf;
    for (ix = 0; ix < 16; ++ix) {
	int x = (md5sum[ix] >> 4);
	*outp++ = hex[x];
	x = (md5sum[ix] & 0x0f);
	*outp++ = hex[x];
    }
    *outp = '\0';
    --ressz;
    if (ressz > 32) { ressz = 32; }
    memcpy (resbf, outbuf, ressz);
    resbf[ressz] = '\0';
    vxlog (5, "   (Pseudo-)UIDL = %s\n", outbuf);
}

int rcalc_uidl (char *rdfile, char *resbf, size_t ressz)
{
    MD5_CTX uidlcntxt;
    unsigned char md5sum[16];
    char line[1024], *hex = "0123456789abcdef";
    char outbuf[33], *outp;
    ssize_t llen;
    size_t rlen = 32;
    int ix;
    FILE *rd;
    MD5Init (&uidlcntxt);
    if (!(rd = fopen (rdfile, "r"))) {
	vxlog (1, "opening redirection message file `%s' failed: %s\n",
		  rdfile, strerror (errno));
	return 1;
    }
    vxlog (5, "   calculating MD5 sum of message header ");
    while (fgets (line, 1024, rd)) {
	llen = strlen (line);
	MD5Update (&uidlcntxt, (uchar *) line, llen);
	vxlog (5, ".");
	if (*line == '\r' && (!line[1] || (line[1] == '\n' && !line[2]))) break;
	if (*line == '\n' && !line[1]) break;
    }
    fclose (rd);
    MD5Final (md5sum, &uidlcntxt);
    vxlog (5, " done.\n");
    --ressz;
    if (ressz < rlen) rlen = ressz;
    outp = outbuf;
    for (ix = 0; ix < 16; ++ix) {
	int x = (md5sum[ix] >> 4);
	*outp++ = hex[x];
	x = (md5sum[ix] & 0x0f);
	*outp++ = hex[x];
    }
    *outp = '\0';
    --ressz;
    if (ressz > 32) { ressz = 32; }
    memcpy (resbf, outbuf, ressz);
    resbf[ressz] = '\0';
    vxlog (5, "   (Pseudo-)UIDL = %s\n", outbuf);
    return 0;
}
