#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

struct _tnsz {
    char *t_name;
    size_t t_size;
};

static struct _tnsz sc_types[] = {
    { "char", sizeof (char) },
    { "short int", sizeof (short int) },
    { "int", sizeof (int) },
    { "long int", sizeof (long int) },
    { "long long int", sizeof (long long int) },
    { (char *) 0, 0 }
};

static struct _tnsz uc_types[] = {
    { "unsigned char", sizeof (unsigned char) },
    { "unsigned short int", sizeof (unsigned short int) },
    { "unsigned int", sizeof (unsigned int) },
    { "unsigned long int", sizeof (unsigned long int) },
    { "unsigned long long int", sizeof (unsigned long long int) },
    { (char *) 0, 0 }
};

static struct _tnsz c_ptypes[] = {
    { "void *", sizeof (void *) },
    { (char *) 0, 0 }
};

static struct _tnsz indep_ptypes[] = {
    { "pointer", sizeof (void *) },
    { (char *) 0, 0 }
};

static struct _tnsz sindep_types[] = {
    { "int8_t", 8 },
    { "int16_t", 16 },
    { "int32_t", 32 },
    { (char *) 0, 0 }
};

static struct _tnsz uindep_types[] = {
    { "u_int8_t", 8 },
    { "u_int16_t", 16 },
    { "u_int32_t", 32 },
    { (char *) 0, 0 }
};

static size_t charsize (void)
{
    char c = (char) 1;
    size_t sz = 0;
    do {
	++sz;
	c <<= 1;
    } while (c);
    return sz;
}

static char *progname;

static int try (struct _tnsz t1, struct _tnsz types[]);
static void openoutput (char *outfile);
static void check_oopt (int *argc, char *argv[], char **outfile);

int main (int argc, char *argv[])
{
    int ix, iy, iz, n_errors = 0;
    size_t cz = charsize ();
    char *outfile = (char *) 0;
    if ((progname = strrchr(argv[0], '/'))) ++progname; else progname = argv[0];
    for (ix = 0; sc_types[ix].t_name; ++ix) {
	sc_types[ix].t_size *= cz;
    }
    for (ix = 0; uc_types[ix].t_name; ++ix) {
	uc_types[ix].t_size *= cz;
    }
    fprintf (stderr, "%s: The bit size of the type `char' is %d; this sounds "
		     "%s!\n",
		     progname,
		     cz,
		     (cz == 8 ? "good" : "bad"));
    iz = 0;
    check_oopt (&argc, argv, &outfile);
    openoutput (outfile);
    if (argc < 2) {
	for (ix = 0; sindep_types[ix].t_name; ++ix) {
	    if (try (sindep_types[ix], sc_types)) { ++n_errors; }
	}
	for (ix = 0; uindep_types[ix].t_name; ++ix) {
	    if (try (uindep_types[ix], uc_types)) { ++n_errors; }
	}
	for (ix = 0; indep_ptypes[ix].t_name; ++ix) {
	    if (try (indep_ptypes[ix], c_ptypes)) { ++n_errors; }
	}
    } else {
	for (iz = 1; iz < argc; ++iz) {
	    for (iy = 0; sindep_types[iy].t_name; ++iy) {
		if (!strcmp (argv[iz], sindep_types[iy].t_name)) {
		    if (!try (sindep_types[iy], sc_types)) { goto DONE; }
		    ++n_errors;
		    goto DONE;
		}
	    }
	    for (iy = 0; uindep_types[iy].t_name; ++iy) {
		if (!strcmp (argv[iz], uindep_types[iy].t_name)) {
		    if (!try (uindep_types[iy], uc_types)) { goto DONE; }
		    ++n_errors;
		    goto DONE;
		}
	    }
	    for (iy = 0; indep_ptypes[iy].t_name; ++iy) {
		if (!strcmp (argv[ix], indep_ptypes[iy].t_name)) {
		    if (!try (indep_ptypes[iy], c_ptypes)) { goto DONE; }
		    ++n_errors;
		    goto DONE;
		}
	    }
	    fprintf (stderr, "%s: Sorry, the type name %s is unknown!\n",
			     progname,
			     argv[iz]);
	    ++n_errors;
	DONE:
	    ;
	}
    }
    return (n_errors ? 1 : 0);
}

static int try (struct _tnsz t1, struct _tnsz types[])
{
    int ix;
    for (ix = 0; types[ix].t_name; ++ix) {
	if (t1.t_size == types[ix].t_size) {
	    printf ("typedef %s %s;\n", types[ix].t_name, t1.t_name);
	    return 0;
	}
    }
    fprintf (stderr,
	     "%s: Sorry, but i couldn't find a C-type matching %s!\n",
	     progname,
	     t1.t_name);
    return 1;
}

static void check_oopt (int *argc, char *argv[], char **outfile)
{
    int ix, iy, iz, ac = *argc;
    iz = iy = 0;
    *outfile = (char *) 0;
    for (ix = 1; ix < ac; ++ix) {
	if (!strncmp (argv[ix], "-o", 2)) {
	    iy = ix;
	    if (strlen (argv[ix]) == 2) {
		if (ix + 1 < ac) {
		    *outfile = argv[ix + 1];
		    iz = ix + 2;
		} else {
		    fprintf (stderr,
			     "%s: missing argument for `-o'!\n",
			     progname);
		    exit (1);
		}
	    } else {
		*outfile = &argv[ix][2];
		iz = ix + 1;
	    }
	    break;
	}
    }
    if (iz) {
	while (iz < ac) { argv[iy++] = argv[iz++]; }
	*argc = ac - (iz - iy);
    }
}

static void openoutput (char *outfile)
{
    int fd;
    if (!outfile) { return; }
    if ((fd = open (outfile, O_CREAT|O_TRUNC|O_WRONLY, 0644)) < 0) {
	fprintf (stderr,
		 "%s: Sorry, opening the output file `%s' failed: %s!\n",
		 progname,
		 outfile,
		 strerror (errno));
	exit (1);
    }
    fflush (stdout);
    dup2 (fd, 1);
    close (fd);
}
