#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>


struct cs {
    char *c;
    size_t s;
    struct cs *l;
};

static char *prog;

/* short [int] */
struct cs t_short[] = {
    { "int", sizeof (short int), 0 },
    { NULL, sizeof (short), 0 }
};
/* long long [int] */
struct cs t_longlong[] = {
    { "int", sizeof (long long int), 0 },
    { NULL, sizeof (long long), 0 }
};
/* long double, long int, long long, long */
struct cs t_long[] = {
    { "double", sizeof (long double), 0 },
    { "int", sizeof (long int), 0 },
    { "long", 0, t_longlong },
    { NULL, sizeof (long), 0 }
};
/* unsigned short [int] */
struct cs t_ushort[] = {
    { "int", sizeof (unsigned short int), 0 },
    { NULL, sizeof (unsigned short), 0 }
};
/* unsigned long long [int] */
struct cs t_ulonglong[] = {
    { "int", sizeof (unsigned long long int), 0 },
    { NULL, sizeof (unsigned long long), 0 }
};
/* unsigned long int, unsigned long long */
struct cs t_ulong[] = {
    { "int", sizeof (unsigned long int), 0 },
    { "long", 0, t_ulonglong },
    { NULL, sizeof (unsigned long), 0 }
};
/* unsigned char, unsigned short, unsigned int, unsigned long, unsigned */
struct cs t_unsigned[] = {
    { "char", sizeof (unsigned char), 0 },
    { "short", 0, t_ushort },
    { "int", sizeof (unsigned int), 0 },
    { "long", 0, t_ulong },
    { NULL, sizeof (unsigned), 0 }
};
#ifdef __GNUC__
/* __complex__ short ... */
struct cs t_cshort[] = {
    { "int", sizeof (__complex__ short int), 0 },
    { NULL, sizeof (__complex__ short), 0 }
};
/* __complex__ long long ... */
struct cs t_clonglong[] = {
    { "int", sizeof (__complex__ long long int), 0 },
    { NULL, sizeof (__complex__ long long), 0 }
};
/* __complex__ long ... */
struct cs t_clong[] = {
    { "double", sizeof (__complex__ long double), 0 },
    { "int", sizeof (__complex__ long int), 0 },
    { "long", 0, t_clonglong },
    { NULL, sizeof (__complex__ long), 0 }
};
/* __complex__ unsigned short ... */
struct cs t_cushort[] = {
    { "int", sizeof (__complex__ unsigned short int), 0 },
    { NULL, sizeof (__complex__ unsigned short), 0 }
};
/* __complex__ unsigned long long ... */
struct cs t_culonglong[] = {
    { "int", sizeof (__complex__ unsigned long long int), 0 },
    { NULL, sizeof (__complex__ unsigned long long), 0 }
};
/* __complex__ unsigned long ... */
struct cs t_culong[] = {
    { "int", sizeof (__complex__ unsigned long int), 0 },
    { "long", 0, t_culonglong },
    { NULL, sizeof (__complex__ unsigned long), 0 }
};
/* __complex__ unsigned ... */
struct cs t_cunsigned[] = {
    { "char", sizeof (__complex__ unsigned char), 0 },
    { "short", 0, t_cushort },
    { "int", sizeof (__complex__ unsigned int), 0 },
    { "long", 0, t_culong },
    { NULL, sizeof (__complex__ unsigned), 0 }
};
/* __complex__ ... */
struct cs t_complex[] = {
    { "char", sizeof (__complex__ char), 0 },
    { "short", 0, t_cshort },
    { "int", sizeof (__complex__ int), 0 },
    { "long", 0, t_clong },
    { "unsigned", 0, t_cunsigned },
    { "float", sizeof (__complex__ float), 0 },
    { "double", sizeof (__complex__ double), 0 },
    { NULL, 0, 0 }
};
#endif
struct cs t_void[] = {
    { "*", sizeof (void *), 0 },
    { NULL, 0, 0 }
};
/* base; here all integer and floating point type specifications start */
struct cs t_base[] = {
    { "char", sizeof (char), 0 },
    { "short", 0 , t_short },
    { "int", sizeof (int), 0 },
    { "long", 0, t_long },
    { "unsigned", 0, t_unsigned },
#ifdef __GNUC__
    { "__complex__", 0, t_complex },
#endif
    { "float", sizeof (float), 0 },
    { "double", sizeof (double), 0 },
    { "void", 0, t_void },
    { NULL, 0, 0 }
};
/* base; here all integer and floating point type specifications start
 * but this table doesn't contain the `__complex__' types
 */
struct cs t_base_nc[] = {
    { "char", sizeof (char), 0 },
    { "short", 0 , t_short },
    { "int", sizeof (int), 0 },
    { "long", 0, t_long },
    { "unsigned", 0, t_unsigned },
    { "float", sizeof (float), 0 },
    { "double", sizeof (double), 0 },
    { "void", 0, t_void },
    { NULL, 0, 0 }
};

size_t typesize (struct cs *table, int argc, char *argv[])
{
    int ix;
    if (argc < 2) {
	for (ix = 0; table[ix].c; ++ix);
	return table[ix].s;
    }
    for (ix = 0; table[ix].c; ++ix) {
	if (!strcmp (argv[1], table[ix].c)) {
	    if (table[ix].s) {
		return table[ix].s;
	    } else if (table[ix].l) {
		return typesize (table[ix].l, argc - 1, argv + 1);
	    } else {
		return 0;
	    }
	}
    }
    return 0;
}

#define bitsize(T) \
    ({ T x = (T) 1; int cc = 0; while (x) { x <<= 1; ++cc; }; cc; })

void allsizes (char *pfx, char *sfx, struct cs *table)
{
    int ix;
    int pfxlen = strlen (pfx);
    int beginning = (table == t_base) || (table == t_base_nc);
    for (ix = 0; table[ix].c; ++ix) {
	if (table[ix].s) {
	    printf ("%s%s%s%s%d\n",
		    pfx,
		    (beginning ? "" : " "),
		    table[ix].c,
		    sfx,
		    table[ix].s * bitsize(char));
	} else if (table[ix].l) {
	    if (!beginning) strcat (pfx, " ");
	    strcat (pfx, table[ix].c);
	    allsizes (pfx, sfx, table[ix].l);
	    pfx[pfxlen] = '\0';
	}
    }
    if (table[ix].s) {
	printf ("%s%s%d\n", pfx, sfx, table[ix].s * bitsize(char));
    }
}


int main (int argc, char *argv[])
{
    char pfx[512];
    int opt;
    struct cs *basetable = t_base;
    prog = strrchr (argv[0], '/');
    if (prog) ++prog; else prog = argv[0];
    opterr = 0;
    while ((opt = getopt (argc, argv, "ch")) > 0) {
	switch (opt) {
	    case 'c':
		basetable = t_base_nc;
		break;
	    case 'h':
		fprintf (stderr, "Usage: %s [-c] [<C type>]\n"
				 "       %s -h\n",
				 prog, prog);
		exit (1);
	    default:
		fprintf (stderr, "%s: illegal option `-%c'!\n", prog, optopt);
		exit (1);
	}
    }
    if (optind != argc) {
	size_t bytesize = typesize (basetable, argc-optind+1, &argv[optind-1]);
	if (bytesize == 0) {
	    fprintf (stderr, "%s: illegal type specification!\n", prog);
	    exit (1);
	}
	printf ("%d\n", bytesize * bitsize (char));
    } else {
	strcpy (pfx, "bytesize (");
	allsizes (pfx, ") = ", basetable);
    }
    return 0;
}
