/* intcoder.c
**
** code/decode integers in big endian order in a string
** using c uuencode-like coding mechanism
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#define _INTCODER_C
#include "intcoder.h"
#include "vxerror.h"

static char codes[] = "!\"#$%&'()*+,-./0123456789:;<=>?@"
		      "ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`";

int decode (char **ci)
{
    int ix, res = 0;
    char *p;
    /* A byte contains eight bits but with coding we can only use six of
    ** these eight bits; so we need 4/3 times the size of an integer in
    ** bytes for holding a complete integer plus one byte for the fraction
    ** of 4/3 - the most significant bits.
    */
    size_t bsz = ICBSZ;
    for (ix = 0; ix < bsz; ++ix) {
	if (!(p = strchr (codes, *(*ci)++))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"coded integer is corrupt!\n");
	}
	res = (res << 6) | (p - codes);
    }
    return res;
}

void encode (char **cb, int val)
{
    int ix, v;
    char *p;
    size_t bsz = ICBSZ;
    p = *cb + bsz;
    *cb = p;
    for (ix = 0; ix < bsz; ++ix) {
	*--p = codes[val & 0x3F];
	val >>= 6;
    }
}
