/* md5.c
**
**   create a MD5 digest
*/
#include <stdio.h>
#include <string.h>
#include <endian.h>

#define WORD_A 0x01234567
#define WORD_B 0x89abcdef
#define WORD_C 0xfedcba98
#define WORD_D 0x76543210

typedef unsigned long ulong;
typedef unsigned short ushort;
typedef unsigned char uchar;

#ifndef __GNUC__
#define __inline__
#endif

/* Auxiliary functions as defined in RFC1321 */
__inline__ ulong F(ulong X, ulong Y, ulong Z)
{
    return (X & Y) | (~X & Z);
}

__inline__ ulong G(ulong X, ulong Y, ulong Z)
{
    return (X & Z) | (Y & ~Z);
}

__inline__ ulong H(ulong X, ulong Y, ulong Z)
{
    return (X ^ Y) ^ Z;
}

__inline__ ulong I(ulong X, ulong Y, ulong Z)
{
    return Y ^ (X | ~Z);
}

__inline__ ulong lrot (ulong v, int n)
{
    return ((v << n) | (v >> (32 - n)));
}

__inline__ ulong rrot (ulong v, int n)
{
    return ((v >> n) | (v << (32 - n)));
}

#if BYTE_ORDER == LITTLE_ENDIAN
#define bswap(v)
#else
__inline__ void bswap (ulong *v)
{
    ushort *p = (ushort *) v, *q = (ushort *) v + 1;
    ushort t;
    t = (*p >> 8) | (*p << 8);
    *p = (*q >> 8) | (*q << 8);
    *q = t;
}
#endif

#include "sinetab.c"

#define bF(a,b,c,d,Xk,s,Ti) \
    ((a) = (b) + lrot(((a) + F((b),(c),(d)) + Xk + Ti), s))
#define bG(a,b,c,d,Xk,s,Ti) \
    ((a) = (b) + lrot(((a) + G((b),(c),(d)) + Xk + Ti), s))
#define bH(a,b,c,d,Xk,s,Ti) \
    ((a) = (b) + lrot(((a) + H((b),(c),(d)) + Xk + Ti), s))
#define bI(a,b,c,d,Xk,s,Ti) \
    ((a) = (b) + lrot(((a) + I((b),(c),(d)) + Xk + Ti), s))

__inline__ void md5step (ulong X[], ulong *pA, ulong *pB, ulong *pC, ulong *pD)
{
    ulong A = *pA, B = *pB, C = *pC, D = *pD;
    ulong AA, BB, CC, DD;
    AA = A; BB = B; CC = C; DD = D;
    /* Round 1 */
    bF(A,B,C,D,X[ 0], 7,T[ 0]); bF(D,A,B,C,X[ 1],12,T[ 1]);
    bF(C,D,A,B,X[ 2],17,T[ 2]); bF(B,C,D,A,X[ 3],22,T[ 3]);
    bF(A,B,C,D,X[ 4], 7,T[ 4]); bF(D,A,B,C,X[ 5],12,T[ 5]);
    bF(C,D,A,B,X[ 6],17,T[ 6]); bF(B,C,D,A,X[ 7],22,T[ 7]);
    bF(A,B,C,D,X[ 8], 7,T[ 8]); bF(D,A,B,C,X[ 9],12,T[ 9]);
    bF(C,D,A,B,X[10],17,T[10]); bF(B,C,D,A,X[11],22,T[11]);
    bF(A,B,C,D,X[12], 7,T[12]); bF(D,A,B,C,X[13],12,T[13]);
    bF(C,D,A,B,X[14],17,T[14]); bF(B,C,D,A,X[15],22,T[15]);
    /* Round 2 */
    bG(A,B,C,D,X[ 0], 7,T[ 0]); bG(D,A,B,C,X[ 1],12,T[ 1]);
    bG(C,D,A,B,X[ 2],17,T[ 2]); bG(B,C,D,A,X[ 3],22,T[ 3]);
    bG(A,B,C,D,X[ 4], 7,T[ 4]); bG(D,A,B,C,X[ 5],12,T[ 5]);
    bG(C,D,A,B,X[ 6],17,T[ 6]); bG(B,C,D,A,X[ 7],22,T[ 7]);
    bG(A,B,C,D,X[ 8], 7,T[ 8]); bG(D,A,B,C,X[ 9],12,T[ 9]);
    bG(C,D,A,B,X[10],17,T[10]); bG(B,C,D,A,X[11],22,T[11]);
    bG(A,B,C,D,X[12], 7,T[12]); bG(D,A,B,C,X[13],12,T[13]);
    bG(C,D,A,B,X[14],17,T[14]); bG(B,C,D,A,X[15],22,T[15]);
    /* Round 3 */
    bH(A,B,C,D,X[ 0], 7,T[ 0]); bH(D,A,B,C,X[ 1],12,T[ 1]);
    bH(C,D,A,B,X[ 2],17,T[ 2]); bH(B,C,D,A,X[ 3],22,T[ 3]);
    bH(A,B,C,D,X[ 4], 7,T[ 4]); bH(D,A,B,C,X[ 5],12,T[ 5]);
    bH(C,D,A,B,X[ 6],17,T[ 6]); bH(B,C,D,A,X[ 7],22,T[ 7]);
    bH(A,B,C,D,X[ 8], 7,T[ 8]); bH(D,A,B,C,X[ 9],12,T[ 9]);
    bH(C,D,A,B,X[10],17,T[10]); bH(B,C,D,A,X[11],22,T[11]);
    bH(A,B,C,D,X[12], 7,T[12]); bH(D,A,B,C,X[13],12,T[13]);
    bH(C,D,A,B,X[14],17,T[14]); bH(B,C,D,A,X[15],22,T[15]);
    /* Round 4 */
    bI(A,B,C,D,X[ 0], 7,T[ 0]); bI(D,A,B,C,X[ 1],12,T[ 1]);
    bI(C,D,A,B,X[ 2],17,T[ 2]); bI(B,C,D,A,X[ 3],22,T[ 3]);
    bI(A,B,C,D,X[ 4], 7,T[ 4]); bI(D,A,B,C,X[ 5],12,T[ 5]);
    bI(C,D,A,B,X[ 6],17,T[ 6]); bI(B,C,D,A,X[ 7],22,T[ 7]);
    bI(A,B,C,D,X[ 8], 7,T[ 8]); bI(D,A,B,C,X[ 9],12,T[ 9]);
    bI(C,D,A,B,X[10],17,T[10]); bI(B,C,D,A,X[11],22,T[11]);
    bI(A,B,C,D,X[12], 7,T[12]); bI(D,A,B,C,X[13],12,T[13]);
    bI(C,D,A,B,X[14],17,T[14]); bI(B,C,D,A,X[15],22,T[15]);
    /* Increment registers */
    *pA = A + AA; *pB = B + BB; *pC = C + CC; *pD = D + DD;
}

void md5enc(size_t msglen, char *msg, ulong digest[4])
{
    ulong A = WORD_A, B = WORD_B, C = WORD_C, D = WORD_D;
    ulong X[16];
    int ix;
    ulong blenH = msglen >> 29, blenL = msglen << 3;
    ulong blklen = msglen / 16;
    ulong trlen = 16 * blklen;
    ulong restlen = msglen - trlen;
    ulong *M;
    uchar pad[64];
    char *endmsg = msg + trlen;
    bswap (&blenH); bswap (&blenL);
    memcpy (pad, msg + trlen, restlen);
    pad[restlen] = 0x80;
    memset (pad + restlen + 1, '\0', 64 - restlen - 1);
    memcpy (pad + 56, &blenL, 4);
    memcpy (pad + 60, &blenH, 4);
    while (msg != endmsg) {
	M = (ulong *) msg;
	for (ix = 0; ix < 16; ++ix) {
	    X[ix] = M[ix];
	    bswap (&X[ix]);
	}
	md5step (X, &A, &B, &C, &D);
    }
    M = (ulong *) pad;
    for (ix = 0; ix < 16; ++ix) {
	X[ix] = M[ix];
	bswap (&X[ix]);
    }
    md5step (X, &A, &B, &C, &D);
    digest[0] = A; bswap (&digest[0]);
    digest[1] = B; bswap (&digest[1]);
    digest[2] = C; bswap (&digest[2]);
    digest[3] = D; bswap (&digest[3]);
}
