/* smtp.c
**
** send smtp commands via the `high level' soket interface to a SMTP-Server.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "tools.h"
#include "smtp.h"

/* Entferne das Zeilenende (`\r', `\n' oder `\r\n') von einer Zeile und gib
 * ** ein Flag zurck, das darber Auskunft gibt, ob die Zeile ein Zeilenende
 * ** enthielt (also abgeschlossen war) oder nicht.
 * */
static __inline__ int cuteol (char *line)
{
    char *p = line + strlen (line);
    if (p == line) { return 0; }
	if (*--p == '\r') { *p = '\0'; return 1; }
	if (*p == '\n') {
	    *p = '\0';
	    if (p != line && *--p == '\r') { *p = '\0'; }
	    return 1;
	}
	return 0;
}

static int SMTP_Result (HLSock in)
{
    char line[1024], *p;
    int ecode = 0;
    long stv;
    while (HLGets (line, 1024, in)) {
	cuteol (line);
	vxlog (4, "%s\n", line);
	stv = strtol (line, &p, 10);
	if (!ecode) { ecode = (int) stv; }
	if (*p == '\r' || *p == '\n' || *p == ' ') { break; }
    }
    return ecode;
}

HLSock SMTP_Connect (char *smtp_host)
{
    int stv;
    HLSock serv = HLConnect ("tcp", smtp_host, "smtp");
    if (!serv) { return (HLSock) 0; }
    stv = SMTP_Result (serv);
    if (stv / 100 != 2) { HLClose (serv); return (HLSock) 0; }
    if (HLhfInit (serv)) {
	HLClose (serv); return (HLSock) 0;
    }
    return serv;
}

int SMTP_HELLO (HLSock out, char *sender_host)
{
    int stv;
    HLPuts ("HELO ", out); HLPuts (sender_host, out); HLPuts ("\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 2) { return 1; }
    return 0;
}

int SMTP_MAIL (HLSock out, char *sender)
{
    int stv;
    HLPuts ("MAIL FROM:", out); HLPuts (sender, out); HLPuts ("\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 2) { return 1; }
    return 0;
}

int SMTP_RCPT (HLSock out, char *recipient)
{
    int stv;
    HLPuts ("RCPT TO:", out); HLPuts (recipient, out); HLPuts ("\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 2) { return 1; }
    return 0;
}

int SMTP_DATA (HLSock out)
{
    int stv;
    HLPuts ("DATA\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 3) { return 1; }
    return 0;
}

int SMTP_SEND (HLSock out, char *line)
{
    HLPuts (line, out);
    return 0;
}

int SMTP_EOD (HLSock out)
{
    int stv;
    HLPuts (".\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 2) { return 1; }
    return 0;
}

int SMTP_QUIT (HLSock out)
{
    int stv;
    HLPuts ("QUIT\r\n", out);
    stv = SMTP_Result (out);
    if (stv / 100 != 2) { return 1; }
    return 0;
}
