/* tools.c
**
** Einige ntzliche Routinen
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>

#include "vxerror.h"
#define _TOOLS_C
#include "tools.h"
#include "prefs.h"

static exception_t erraction = (exception_t) 0; 

exception_t setup_exception (exception_t handle)
{
    exception_t tmp = erraction;
    erraction = handle;
    return tmp;
}

typedef struct heapel_s *heapel;

struct heapel_s {
	size_t sz;
	heapel next;
	char vec[1];
};

static heapel gheap = (heapel) NULL;

void heap_init (void)
{
    gheap = (heapel) NULL;
}

static void *lookup (size_t sz)
{
    heapel p, q;
    p = gheap;
    if (p) {
/*##*/fprintf (stderr, "##########lookup(A): p = 0x%08x\n", (char *) p);
	if (p->sz - sizeof (size_t) >= sz) {
	    gheap = p->next;
/*##*/fprintf (stderr, "##lookup(0): allocated 0x%08x\n",((char *)p));
	    return (void *) ((char *) p + sizeof (size_t));
	}
	for (q = p, p = p->next; p; q = p, p = p->next) {
/*##*/fprintf (stderr, "##########lookup(B): p = 0x%08x, q = 0x%08x\n", (char *) p, (char *)q);
	    if (p->sz - sizeof (size_t) >= sz) {
		q->next = p->next;
/*##*/fprintf (stderr, "##lookup(1): allocated 0x%08x\n",((char *)p));
		return (void *) ((char *) p + sizeof (size_t));
	    }
	}
    }
    if (sz < sizeof (heapel)) sz = sizeof (heapel);
    p = (heapel) malloc (sz + sizeof (size_t));
    if (!p) { return (void *) p; }
    p->sz = sz + sizeof (size_t);
/*##*/fprintf (stderr, "##lookup(2): allocated 0x%08x\n",((char *)p));
    return (void *) ((char *) p + sizeof (size_t));
}

char *gdupl (char *file, int line, char *string)
{
    char *res = lookup (strlen (string) + 1);
    if (!res) {
	if (erraction) {
	    erraction ("(gdupl) failed");
	}
	vxerror_at (EX_OSERR, errno, file, line, "(gdupl)");
    }
    strcpy (res, string);
    return res;
}

void *galloc (char *file, int line, size_t size)
{
    void *res = lookup (size);
    if (!res) {
	vxerror_at (EX_OSERR, errno, file, line, "(galloc)");
    }
    return res;
}

void gfree (void *ref)
{
    heapel nr;
/*##*/fprintf(stderr, "##gfree(0): releasing 0x%08x\n", (char *)ref-sizeof(size_t));
    nr = (heapel) ((char *) ref - sizeof (size_t));
    nr->next = gheap;
    gheap = nr;
}

static void *grealloc (void *ref, size_t size)
{
    void *res;
    heapel o = (heapel) ((char *) ref - sizeof (size_t));
    size_t csz = o->sz - sizeof (size_t);
    res = lookup (size);
    if (res) {
	memcpy (res, ref, csz);
	gfree (ref);
    }
    return res;
}

void *gresize (char *file, int line, void *ref, size_t size)
{
    void *res = grealloc (ref, size);
    if (!res) {
	vxerror_at (EX_OSERR, errno, file, line, "(gresize)");
    }
    return res;
}

/* concat a NULL-terminated list of string to a new string which is the
** result of this function.
*/
char *concat (char *_1st, ...)
{
    va_list ap;
    char *res, *op;
    size_t res_sz = 0, op_sz;

    va_start (ap, _1st);
    op_sz = strlen (_1st);
    if (!(res = lookup (op_sz + 1))) { return res; }
    memcpy (res, _1st, op_sz);
    res_sz += op_sz;
    while ((op = va_arg (ap, char *))) {
	char *newres;
	op_sz = strlen (op);
	if (!(newres = grealloc (res, res_sz + op_sz + 1))) {
	    gfree (res);
	    return newres;
	}
	res = newres;
	memcpy (res + res_sz, op, op_sz);
	res_sz += op_sz;
    }
    res[res_sz] = '\0';
    return res;
}

void vxlog (int ll, char *fmt, ...)
{
    va_list ap;
    if (loglevel >= ll) {
	va_start (ap, fmt);
	vfprintf (stderr, fmt, ap);
	va_end (ap);
	fflush (stderr);
    }
}
