/* vxlist.h
**
** Interface file for `vxlist.c'
** (a list datatype for the VxMail-system)
**
*/

/* I intend to implement this data type as list with pairs of references;
** this makes it possible to implement
**   - doubly linked lists
**   - FIFOs
*/
#ifndef _VXLIST_H
#define _VXLIST_H

/* these structures are invisible outside of the implementation
** of `vxlist'; so `vxlist' is an abstract data type.
*/
/* The `vxlist' type structure */
struct _vxlist;
typedef struct _vxlist *vxlist;

/* The type structure for all iterators on `vxlist' */
struct _vxlist_iterator;
typedef struct _vxlist_iterator *vxlist_iterator;

typedef enum { LT_FIFO, LT_DOUBLE, LT_RING } vxlist_type;

typedef enum { VXLD_FORW, VXLD_BACK } vxlist_direction;

#ifndef _VXLIST_C

/* insert a new element into the list `ent'; the new list is the
** result of this operation;
*/
vxlist vxlist_insert (vxlist ent, void *info, vxlist_type lt, int append);

/* create a new list iterator for the list `ent'; the iterator
** has to advance in the specified direction (`dir') throug the list.
*/
vxlist_iterator vxlist_newiterator (vxlist ent, vxlist_direction dir);

/* (Re-)Set the a list iterator to it's initial state */
void *vxlist_first (vxlist_iterator *itr);

/* advance through an iterator */
void *vxlist_advance (vxlist_iterator *itr);

/* Release the memory allocated for a list iterator and reset the
** iterator to an uninitialized state
*/
void vxlist_freeiterator (vxlist_iterator *itr);

#endif /* defined(_VXLIST_C) */

#endif /* defined(_VXLIST_H) */
