/* conv.c
**
** Convert a message folder into the new format (i.e. add an index file for
** that message).
**
** Dieses Modul habe ich in die `libtools.a' integriert, weil es von zwei
** Programmen des VxMail-Systems verwendet werden wird: `vxconv' und `vxpop'.
** Dies hat folgende Grnde:
** a) Wenn ich `vxpop' als vollwertigen Ersatz fr irgendeinen POP3-Server
**    verwenden will, dann mu dieses Programm dazu in der Lage sein, auch
**    normale `smail'- (bzw. `sendmail'-) Folder zu bearbeiten; dies kann am
**    einfachsten realisiert werden, wenn `vxpop' temporr eine Indexdatei zu
**    dem jeweiligen Mail-Folder anlegt; diese Indexdatei kann nach
**    Bearbeitung durch `vxpop' wieder gelscht werden.
** b) Wenn vorhandene Folder (mitsamt ihrem Inhalt) an das VxMail-System
**    angepat werden sollen, dann bentige ich sowieso ein Konvertierungs-
**    programm, welches diese Aufgabe bernimmt.
** c) Da sowohl POP3-Server als auch Konvertierungsprogramm in diesem Punkt in
**    etwa das gleiche tun mssen - nmlich konvertieren, mu ich auch diese
**    Aktion auch nur einmal implementieren.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>

#include "regex.h"
#include "xfile.h"
#include "prefs.h"
#include "vxerror.h"
#include "tools.h"

static regex_t *first = NULL, *last = NULL;
static char *fsub, *lsub;

static char * rcmsg (int rccode)
{
    switch (rccode) {
	case REG_BADRPT:
	    return "invalid use of of a repetition operator";
	case REG_BADBR:
	    return "invalid use of a back reference operator";
	case REG_EBRACE:
	    return "unmatched `{' (or `}')";
	case REG_EBRACK:
	    return "unmatched `[' (or `]')";
	case REG_ERANGE:
	    return "invalid use of the range operator `-'";
	case REG_ECTYPE:
	    return "unknown character class name";
	case REG_EPAREN:
	    return "unmatched `(' (or `)')";
	case REG_ESUBREG:
	    return "invalid back reference to a subexpression";
/*	case REG_EEND:
	    return "invalid regular expression (unknown)";
*/
	case REG_EESCAPE:
	    return "invalid escape sequence";
	case REG_BADPAT:
	    return "invalid use of pattern operators";
/*	case REG_ESIZE:
	    return "regular expression buffer overflow";
*/
	case REG_ESPACE:
	    return "out of memory for regular expression";
	default:
	    return "unknown error";
    }
}

/* Bereite alle regulren Ausdrcke vor, die bentigt werden.
** Dies sind in diesem Fall nur der Ausdruck fr die erste und die letzte
** Zeile einer Meldung.
** Die eigentliche Konvertierung kann NUR dann stattfinden, wenn diese Routine
** zuvor genau einmal aufgerufen wurde.
*/
void conv_init (void)
{
    int rccode;
    if (prefs.msg_scantype) {
	fsub = gdupl (__FILE__, __LINE__, prefs.msg_firstline);
	lsub = gdupl (__FILE__, __LINE__, prefs.msg_lastline);
    } else {
	if (!first) { first = galloc (__FILE__, __LINE__, sizeof (regex_t)); }
	if (!last) { last = galloc (__FILE__, __LINE__, sizeof (regex_t)); }
	if ((rccode = regcomp (first, prefs.msg_firstline, REG_NOSUB))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"regcomp failed: %s\n", rcmsg (rccode));
	}
	if ((rccode = regcomp (last, prefs.msg_lastline, REG_NOSUB))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"regcomp failed: %s\n", rcmsg (rccode));
	}
    }
}

/* conv (fx)
**     konvertiert die Folderdatei `fx' (xFile-handle) vom Standardformat
**     in das VxMail-Format; dies geschieht einfach, indem eine Indexdatei
**     zu der Standardfolderdatei erzeugt wird.
**
** Ich hatte keine Lust, lauter einzelne Routinen zu schreiben; daher
** geschieht der (doch recht komplexe) Einlesevorgang innerhalb dieser
** Routine. Ich kann (meine Meinung!) nicht wissen(!!!), wie lang die Zeilen
** innerhalb eines Mail-Folders sein knnen; daher lege ich einen Zeilenpuffer
** an, der dynamisch immer weiter vergrert wird - so lange, bis auch die
** lngste Zeile im Mail-Folder in ihm Platz findet.
** Um flexibel genug zu sein, habe ich noch zwei weitere Konfigurations-
** optionen hinzugefgt: `msg_firstline' und `msg_lastline'. Beide Optionen
** bekommen als Parameter einen regulren Ausdruck, der in dieser Routine zum
** Auffinden von Anfang und Ende einer Mail innerhalb eines Standardfolders
** benutzt wird. Die Bedingung lautet so:
**   Wenn die zuletzt eingelesene Zeile zu dem Muster in `msg_lastline' gepat
**   hat, dann ist die aktuelle Zeile genau dann der Anfang einer Meldung,
**   wenn sie zu dem Muster in `msg_firstline' pat.
** Habe ich den Anfang einer Mail gefunden, dann trage ich die Position dieses
** Anfangs (dieser ersten Zeile) als neue Indexpotsition in die Indexdatei
** ein. Binn ich am Dateiende angelangt, dann trage ich dieses ebenfalls als
** Indexposition ein.
**
** Ach ja - da wre noch etwas.
** Ich berprfe das in dieser Routine nicht, aber es ist klar, da eine
** Konvertierung nur auf einem mittels `xOpen' im Index-Modus geffneten
** Folder stattfinden kann.
*/

static int rconv (xFile fx);
static int sconv (xFile fx);

int conv (xFile fx)
{
    return (prefs.msg_scantype ? sconv (fx) : rconv (fx));
}

static int rconv (xFile fx)
{
    static size_t lbsz = 0;
    size_t nmatch = 1, llen, lblen = 0;
    static char *lbuf = NULL;
    char line[1024], *lp;
    regmatch_t pmatch[1];
    off_t fp = 0, fp1;
    int lastmatch = 0;
    if (!lbuf) {
	lbsz += 1024;
	lbuf = galloc (__FILE__, __LINE__, lbsz);
    }
    if (xEmpty (fx)) { return 0; }
    while ((llen = xGets (line, 1024, fx)) > 0) {
	memcpy (lbuf, line, llen);
	lblen = llen;
	lp = lbuf + lblen;
	while (lp[-1] != '\r' && lp[-1] != '\n') {
	    if ((llen = xGets (line, 1024, fx)) <= 0) { break; }
	    /* Nur, wenn der Platz im bisherigen Puffer nicht ausreicht, dann
	    ** mu dieser vergrert werden.
	    */
	    if (lblen + llen > lbsz) {
		lbuf = gresize (__FILE__, __LINE__, lbuf, (lbsz += 1024));
	    }
	    /* Nach dem `resize' kann lbuf woanders hinzeigen als lp; daher
	    ** mu `lp' neu gesetzt werden - und zwar ans Ende des bisherigen
	    ** Inhaltes von lbuf.
	    */
	    lp = lbuf + lblen;
	    /* Nun wird der neue Teil der Zeile angehngt, */
	    memcpy (lp, line, llen);
	    /* die neue Pufferlnge berechnet */
	    lblen += llen;
	    /* und `lp' fr die Zeilenendeuntersuchung wieder ans Ende der
	    ** Zeile gelegt.
	    */
	    lp = lbuf + lblen;
	}
	/* So, ab hier sollte eine (beliebig lange) Zeile eingelesen worden
	** sein. Nun wird der Offset-Zhler um die Lnge der Zeile erhht, und
	** anschlieend das Zeilenende abgetrennt - ich kann es fr die
	** Untersuchung auf erste bzw. letzte Zeile nicht gebrauchen.
	*/
	fp1 = fp;
	fp += (off_t) lblen;
	lp = lbuf + lblen;
	if (*--lp == '\r') {
	    *lp = '\0';
	} else if (*lp == '\n') {
	    *lp = '\0';
	    if (*--lp == '\r') { *lp = '\0'; }
	}
	/* Ich gehe von folgender Bedingung aus:
	** Wenn `lastmatch' gesetzt ist und die Zeile auf `msg_firstline' pat,
	** dann handelt es sich um die erste Zeile der neuen Meldung.
	** `lastmatch' ist bei der Initialisierung ungesetzt, da ich die erste
	** Mail immer am Anfang der Folderdatei habe und darum ihre Position
	** auch nicht bentige.
	*/
	if (!regexec (first, lbuf, nmatch, pmatch, 0)) {
	    /* Wenn beide Bedingungen zutreffen, dann setze die Dateiposition
	    ** vom Anfang der Zeile als neue Indexposition.
	    */
	    if (lastmatch) { xSet (fx, fp1); }
	}
	/* Ansonsten wird `lastmatch' immer gesetzt, wenn die eingelesene
	** Zeile auf `msg_lastline' pat.
	** Mit diesen beiden Bedingungen sollte eine Erkennung der
	** Anfangspositionen einer Meldung immer mglich sein.
	*/
	lastmatch = !regexec (last, lbuf, nmatch, pmatch, 0);
    }
    /* Die letzte zu setzende Indexposition ist die Dateiendeposition der
    ** Folderdatei. An dieser Stelle steht zwar keine Meldung mehr; sie wird
    ** jedoch fr eine Lngenberechnung der jeweiligen Meldung bentigt; auer-
    ** ist sie die Anfangsposition einer neuen Mail, die spter evtl. an den
    ** Folder angehngt werden soll.
    */
    xSet (fx, fp);
    return 0;
}

static __inline__ int empty_line (char *line)
{
    char *p = line + strlen (line);
    if (*--p == '\r' && p == line) return 1;
    if (*p == '\n') {
	if (p == line) return 1;
	if (*--p == '\r' && p == line) return 1;
    }
    return 0;
}

static __inline__ int prefix (char *pat, char *where)
{
    return !strncmp (where, pat, strlen (pat));
}

static __inline__ int check_lastline (char *line)
{
    return (prefs.msg_scantype == 1? empty_line(line) :
	    prefix(prefs.msg_lastline, line));
}

static int sconv (xFile fx)
{
    char line[1024], *p;
    off_t fp = 0l;
    ssize_t rlen;
    int try_next = 0;
    if (xEmpty (fx)) { return 0; }
    while ((rlen = xGets (line, 1024, fx)) > 0) {
	if (check_lastline (line)) {
	    fp  += (off_t) rlen;
	    try_next = 1; continue;
	}
	if (try_next && !strncmp (line, fsub, strlen (fsub))) {
	    xSet (fx, fp);
	    try_next = 0;
	}
	fp += (off_t) rlen;
	for (;;) {
	    p = line + rlen - 1;
	    if (*p == '\r' || *p == '\n') break;
	    rlen = xGets (line, 1024, fx);
	    if (rlen <= 0) break;
	    fp  += rlen;
	}
    }
    xSet (fx, fp);
    return 0;
}
