/* dbif.h
**
** Interfacedatei zu `dbif.h'
** (Database Interface for the user database)
*/
#ifndef _DBIF_H
#define _DBIF_H

#include <stdio.h>

/* The content of the database consists of variable length records of
** four elements:
**   - the e-mail address of the user or domain name of special key
**   - the users password or domain's maximum number of users
**   - the pathname of the user's folder / pathname of the directory
**     containing all user folders for this domain
**   - a row of flags, specifying the type of the entry
** A record of this typecontains the latter three entries; the first entry
** is the `primary key' of the database; it is only returned by `dbif_seq',
** so i didn't want it be included here.
*/
typedef struct {
    char *passwd;
    char *folder;
    char *flags;
} dbif_rec;

/* DBIF_FILE is the database handle; the name comes from the three (at the
** moment) implemented types of databases. When using a `SyBase' or
** `PostgreSQL' database, this is really not a file, but a set of tables in
** the corresponding relational database; the programs of the VxMail-system
** should nonetheless NOT recognize nor need the type of database which is
** really used; this is the reason for implementing this module as an abstract
** data type.
*/
struct _DBIF_FILE_S;
typedef struct _DBIF_FILE_S *DBIF_FILE;

struct _DBIF_CURSOR_S;
typedef struct _DBIF_CURSOR_S *DBIF_CURSOR;

typedef int (*rrel_t) (char *lkey, dbif_rec *lrec, char *rkey, dbif_rec *rrec);

/* Die verschiedenen Typen von Datenbankeintrgen */
#define RT_USER		0	/* user */
#define CC_USER		(1 << RT_USER)
#define RT_LINK		1	/* link to a user */
#define CC_LINK		(1 << RT_LINK)
#define RT_NSUSER	2	/* user of a non-standard domain */
#define CC_NSUSER	(1 << RT_NSUSER)
#define RT_ADMIN	3	/* admin user */
#define CC_ADMIN	(1 << RT_ADMIN)
#define RT_DOMAIN	4	/* domain */
#define CC_DOMAIN	(1 << RT_DOMAIN)
#define RT_DOMLINK	5	/* link to a domain */
#define CC_DOMLINK	(1 << RT_DOMLINK)
#define RT_NSDOMAIN	6	/* non-standard domain */
#define CC_NSDOMAIN	(1 << RT_NSDOMAIN)
#define RT_LLINK	7	/* a mail alias */
#define CC_LLINK	(1 << RT_LLINK)
#define RT_MAILSYS	8	/* mail system admin */
#define CC_MAILSYS	(1 << RT_MAILSYS)
#define RT_IP4TRANS	9	/* IPv4 -> domain name translation */
#define CC_IP4TRANS	(1 << RT_IP4TRANS)
#define RT_IP6TRANS	10	/* IPv6 -> domain name translation */
#define CC_IP6TRANS	(1 << RT_IP6TRANS)
#define RT_REDIRECT	11	/* Redirection of a domain to another host */
#define CC_REDIRECT	(1 << RT_REDIRECT)
#define RT_IDENTITY	12	/* The database identity entry */
#define CC_IDENTITY	(1 << RT_IDENTITY)
#define RT_VXTS		13	/* Authorizationentry for the VXTS command */
#define CC_VXTS		(1 << RT_VXTS)
#define RT_INVALID	14	/* an invalid user entry */
#define CC_INVALID	(1 << RT_INVALID)
#define RT_FILTER	15	/* filename of a mail filter */
#define CC_FILTER	(1 << RT_FILTER)
#define RT_BSMTP	16	/* a special user entry for a type of `bsmtp' */
#define CC_BSMTP	(1 << RT_BSMTP)
#define CC_DOMLIKE	(CC_DOMAIN|CC_DOMLINK|CC_NSDOMAIN)
#define CC_IPTRANS	(CC_IP4TRANS|CC_IP6TRANS)
#define CC_SPECIAL	(CC_MAILSYS|CC_IPTRANS|CC_REDIRECT|CC_IDENTITY)
#define CC_RUSERS	(CC_USER|CC_NSUSER|CC_ADMIN|CC_INVALID|CC_BSMTP)
#define CC_DENTRIES	(CC_RUSERS|CC_LINK|CC_LLINK|CC_FILTER)

#ifndef _DBIF_C

/* Connect to the user database, but restrict the access to read-only
** operations
*/
void dbif_reset (DBIF_FILE *dbf, char *dbname);

/* Setup a completely new user database; when using one of the implemented
** database modules based on DB1, DB2 or GDBM, this means, that the file(s)
** containing the database are created or truncated to file position 0;
** when using a `SyBase' or `PostgreSQL' database, this should consist of
** `DROP TABLE ...'-operations, followed by `CREATE TABLE ...'-operations on
** all tables of the user database.
*/
void dbif_create (DBIF_FILE *dbf, char *dbname, int mode);

/* Connect to the user database without restrictions; this access should
** lock the user database or parts of it; how this operation is implemented
** for `SyBase' or `PostgreSQL' databases is really not interesting for the
** VxMail-system.
*/
void dbif_rw (DBIF_FILE *dbf, char *dbname);

/* End all transactions on the user database and disconnect from it.
** When using a DB1, DB2 or GDBM library-like database, this should consist
** of a flush operation, followed by closing the corresponding database
** file(s); when using a `SyBase' or `PostgreSQL' database, this
** should consist of an `END TRANSACTION ...'-operation, followed by the
** operations which are needed for releasing the connection to the user
** database.
*/
void dbif_close (DBIF_FILE dbf);

/* commit all changes on the user database; a simple flush operation
** for DB1, DB2 or GDBM-like databases; a `COMMIT ...' or `END TRANSACTION ...'
** operation for `SyBase' or `PostgreSQL' databases.
*/
void dbif_sync (DBIF_FILE dbf);

/* follow a link-chain to it's origin and return the corresponding entry
** from the user database.
*/
dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr, char **res_addr);

/* get a record from the user database; if there is no record corresponding to
** the key `em_addr', return a NULL-pointer instead.
** The memory allocated for the result is completely dynamically allocated;
** For freeing this memory the function `dbif_free' (see below) MUST be used.
*/
dbif_rec *dbif_get (DBIF_FILE db, char *em_addr);

/* insert a new record into the user database; for simplicity, i don't supply
** a user record; instead, all elements of the database entry are supplied as
** single arguments; the flags are supplied as a NULL-terminated list of
** strings.
*/
int dbif_insert (DBIF_FILE db, char *em_addr, dbif_rec *rec);

/* insert a new record into the user database or update an existing one; for
** simplicity, i don't supply a user record; instead, all elements of the
** database entry are supplied as single arguments; the flags are supplied as
** a NULL-terminated list of strings.
*/
int dbif_set (DBIF_FILE db, char *em_addr, dbif_rec *rec);

/* delete a record from the user database; the system entries and the database
** identity can not be deleted!
** A return value of zero indicates a successful operation; a non-zero return
** value indicates an error.
*/
int dbif_del (DBIF_FILE db, char *em_addr);


/* return the error message corresponding to the last database error */
const char *dbif_errormsg (void);

/* Error numbers */
typedef enum {
    DBIF_NOERR, DBIF_ENOMEM, DBIF_ACFAIL, DBIF_WRFAIL, DBIF_RDFAIL,
    DBIF_SLFAIL, DBIF_NOWR, DBIF_NORD, DBIF_NOIT, DBIF_NODB, DBIF_INT,
    DBIF_DATA, DBIF_CCFAIL
} dbif_error;

/* Error function */
dbif_error dbif_errno (void);

/* The following functions are used for creating user records;
 * the first function can be used for creating records from non-terminated
 * strings, whereas the second one needs NUL-terminated strings.
*/
dbif_rec *dbif_mkrecl (char *passwd, size_t pwl,
		       char *folder, size_t fll,
		       int rtype,
		       ...);

dbif_rec *dbif_mkrec (char *passwd, char *folder, int rtype, ...);

/* This function returns the type of a user record */
int dbif_rectype (dbif_rec *r);

/* The following function sets the user limit and the number of users
** in a domain
*/
int dbif_setlim (dbif_rec *r, int nusers, int maxusers);

/* The following function reads the user limit and the number of users
** from a domain record
*/
int dbif_getlim (dbif_rec *r, int *nusers, int *maxusers);

/* This function releases the memory allocated for a user record completely;
** all elements and the user record structure itself are released.
** For implementation indepency, only this function should be used for
** releasing the dynamically allocated memory for the user record structure.
*/
void dbif_free (dbif_rec *p);

/* The following function implement a (very simple) select/cursor concept.
** I implemented this functions primarily for the programs `vxcheck' and
** `vxdump' but it seems to be usable in a more complex administration
** program, too.
*/

/* dbif_select (db, key, ccflags)
**     selects all records from the database which match the condition
**     constructed fromm the key `key' and the condition flags `ccflags'.
**     the result is a cursor which allows a sequential access to the
**     resulting records.
*/
DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, rrel_t ccfunc);

/* key = dbif_first (dbc, &userrec)
**     resets the database cursor `dbc' to the first record of the selection
**     described by this cursor.
*/
char *dbif_first (DBIF_CURSOR dbc, dbif_rec **userrec);

/* key = dbif_next (dbc, &userrec)
**     reads the next record from the selection described by the database
**     cursor `dbc'.
*/
char *dbif_next (DBIF_CURSOR dbc, dbif_rec **userrec);

/* dbif_end (dbc)
**     removes the database cursor and all resources allocated for it.
*/
int dbif_end (DBIF_CURSOR dbc);

#endif /* defined(_DBIF_C) */

#endif /* defined(_DBIF_H) */
