/* cursor.c
**
** implement a simple database cursor concept for a DB1 based database
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "ifconf.h"
#include "dbif_priv.h"
#include "prefs.h"
#include "tools.h"

static int seq_num = 0;
#if 0
static int ccCheck (DBIF_FILE db,
		    char *key, dbif_rec *rec,
		    char *ikey, dbif_rec *irec,
		    int ccflags);
#endif

/* dbif_select (db, key, ccflags)
**     selects all records from the database which match the condition
**     constructed fromm the key `key' and the condition flags `ccflags'.
**     the result is a cursor which allows a sequential access to the
**     resulting records.
*/
#if 0
DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, int ccflags)
#else
DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, rrel_t ccfunc)
#endif
{
    /* Was mu ich tun?
     * - Lesen des Datensatzes zu dem Argument `key' aus der Datenbank
     * - Konstruktion des Dateinamens fr die Ergebnisdatei
     * - Erzeugung der Ergebnisdatei
     * - Sequentielles Lesen aller Datenstze aus der Datenbank;
     *   bertragen der zu der Bedingung passenden Datenstze in die
     *   Ergebnisdatei; nach jedem Datensatz ein `xSet', damit eine
     *   Indexierung mglich ist.
     * - Umschalten der Ergebnisdatei in den Lesemodus
     * - Erzeugung der Cursor-Datenstruktur; wenn dies unmglich ist, dann
     *   Freigabe aller Resourcen.
     */
    DBIF_CURSOR dbc = (DBIF_CURSOR) 0;
    char *resn, fnbuf[49], *p;
    size_t maxsize = 0, dbcl;
    dbif_rec *rec = (dbif_rec *) 0;
    DBT dkey, drec;
    int seq_flag;
    xFile resf;
    /* 1. Lesen des Datensatzes zu `key' aus der Datenbank */
    if (key && !(rec = dbif_get (db, key))) { return dbc; }
    /* 2. Konstruktion des Dateinamens fr die Ergebnisdatei */
    sprintf (fnbuf, "select-%d.%d", getpid (), seq_num++);
    if (prefs.listdir[strlen (prefs.listdir) - 1] == '/') {
	resn = concat (prefs.listdir, fnbuf, NULL);
    } else {
	resn = concat (prefs.listdir, "/", fnbuf, NULL);
    }
    if (!resn) { dbif_free (rec); return dbc; }
    /* Der Name der Ergebnisdatei ist: <LIST>/select-<pid>.<seq> */
    /* 3. Erzeugung der Ergebnisdatei */
    if (!(resf = xOpen (resn, "ux", 0600))) {
	xUnlink (resn); gfree (resn);
	dbif_free (rec);
	return dbc;
    }
    /* 4. Lesen aller Datenstze aus der Datenbank; bertragen der passenden
    **    Datenstze in die Ergebnisdatei
    */
    seq_flag = R_FIRST;
    while (db->dbf->seq (db->dbf, &dkey, &drec, seq_flag) == 0) {
	size_t eml;
	int nusers, maxusers;
	dbif_rec *irec;
	char *ikey, *p, buf[2 * sizeof(int) + 1];
	/* Erzeuge einen zur DB-Schnittstelle passenden Schlssel aus `dkey' */
	eml = dkey.size;
	if (!(ikey = (char *) malloc (eml + 1))) {
	    /* Fehlgeschlagen; Resourcen freigeben und raus! */
	    free (dkey.data);
	    xClose (resf); xUnlink (resn); gfree (resn);
	    dbif_free (rec);
	    return dbc;
	}
	memcpy (ikey, dkey.data, eml);
	ikey[eml] = '\0';
	/* Erzeuge einen zur DB-Schnittstelle passenden Datensatz aus `drec' */
	if (!(irec = dbif_mkreca (drec.data))) {
	    /* Fehlgeschlagen; Resourcen freigeben und raus! */
	    free (ikey); free (drec.data); free (dkey.data);
	    xClose (resf); xUnlink (resn); gfree (resn);
	    dbif_free (rec);
	    return dbc;
	}
	p = buf;
	*p++ = dbif_rectype(irec);
	/* bertrage die Benutzerbeschrnkungen einer Domain in den Puffer
	** fr die Ausgabe.
	*/
	if (!dbif_getlim (irec, &nusers, &maxusers)) {
	    set_int ((unsigned char *) p, nusers); p += sizeof(int);
	    set_int ((unsigned char *) p, maxusers); p += sizeof(int);
	}
	/* Wenn der Datensatz den Bedingungen gengt (<- ccCheck), dann
	** bertrage ihn - mitsamt Schlssel - in die Datei
	*/
#if 0
/*	if (ccCheck (db, key, rec, ikey, irec, ccflags)) */
#else
	if (ccfunc (key, rec, ikey, irec)) {
#endif
	    size_t pwl = strlen (irec->passwd) + 1,
		   fll = strlen (irec->folder) + 1,
		   tll = p - buf;
	    xWrite (resf, ikey, eml + 1);
	    xWrite (resf, irec->passwd, pwl);
	    xWrite (resf, irec->folder, fll);
	    xWrite (resf, irec->flags, tll);
	    /* Merke die Maximallnge der bisher geschriebenen Datenstze
	    ** fr den Lesepuffer des Cursors
	    */
	    if (maxsize < eml + 1 + pwl + fll + tll) {
		maxsize = eml + 1 + pwl + fll + tll;
	    }
	    /* Trage ein, da der aktuelle Datensatz zuende ist */
	    xSet (resf);
	}
	/* Gib die alloziierten Strukturen wieder frei */
	free (ikey); free (irec); free (drec.data); free (dkey.data);
	/* Hole den Schlssel fr den nchsten Datensatz */
	seq_flag = R_NEXT;
    }
    /* 5. Umschalten der Ergebnisdatei in den Lesemodus */
    xReSet (resf);
    /* 6. Erzeugung der Datenstruktur fr den `cursor' */
    dbcl = sizeof(struct _DBIF_CURSOR_S) + strlen(resn) + maxsize + 2;
    if (!(dbc = (DBIF_CURSOR) malloc(dbcl))) {
	/* Fehlgeschlagen! Mache auch alles andere wieder rckgngig */
	xClose (resf); xUnlink (resn); gfree (resn);
	dbif_free (rec);
	return dbc;
    }
    /* Erst dann, wenn wir hier angekommen sind, ist alles erfolgreich
    ** verlaufen; belege nun den neu alloziierten - aber noch leeren
    ** Speicherbereich mit den ermittelten Werten und gib ihn danach zurck.
    */
    p = (char *) dbc + sizeof(struct _DBIF_CURSOR_S);
    /* Den Dateinamen der Cursordatei bentige ich fr die sptere Freigabe
    ** der Resourcen, d.h das Lschen dieser Datei; dieser Name wird direkt
    ** hinter der Struktur eingetragen.
    */
    dbc->res_name = p; strcpy (p, resn); p += strlen(resn) + 1;
    /* Das `xFile'-Handle wird in die Struktur eingetragen */
    dbc->results = resf;
    /* Der Zeiger auf den Lesepuffer - der direkt nach dem Namen der
    ** Cursordatei beginnt - wird eingetragen
    */
    dbc->buf = p;
    /* Die Puffergre wird eingetragen */
    dbc->bufsize = maxsize;
    /* Die Nummer des ersten Datensatzes in der Cursordatei wird eingetragen */
    dbc->recno = 0;
    /* Ich hatte weiter oben den Datensatz zu dem Argument `key' eingelesen;
    ** der Speicherplatz fr diesen mu nun wieder freigegeben werden.
    */
    dbif_free (rec);
    /* So - alles geschafft! */
    return dbc;
}

/* key = dbif_next (dbc, &userrec)
**     reads the next record the database cursor `dbc' points to.
*/
char *dbif_next (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    int ec = xSeek (dbc->results, dbc->recno), rt, nusers, maxusers;
    char *em_addr = (char *) 0, *p;
    dbif_rec *rec;
    size_t eml;
    char *ema, *pwd, *fld;
    if (ec < 0) { return em_addr; }
    if ((ec = xRead (dbc->results, dbc->buf, dbc->bufsize)) <= 0) {
	return em_addr;
    }
    /* ec ist eigentlich die Anzahl gelesener Bytes (oder < 0 fr einen Fehler);
    ** Ich bentige jedoch nur den Fehler, deshalb der Variablenname:
    ** `ec' steht fr (e)rror (c)ode. Ein Wert ec = 0 steht brigens fr ein
    ** Dateiende.
    */
    p = dbc->buf;        ema = p; eml = strlen (p);
    p += eml + 1;        pwd = p;
    p += strlen (p) + 1; fld = p;
    p += strlen (p) + 1; rt = (int) *p++;
    if (rt == RT_DOMAIN || rt == RT_NSDOMAIN) {
	nusers = get_int ((unsigned char *) p); p += sizeof (int);
	maxusers = get_int ((unsigned char *) p);
	rec = dbif_mkrec (pwd, fld, rt, nusers, maxusers);
    } else {
	rec = dbif_mkrec (pwd, fld, rt);
    }
    /* Diese Fubktion hat zwei Ergebnisse: den Schlssel und den Datensatz;
    ** der Schlssel wird als Funktionswert zurckgegeben; er fungiert
    ** zugleich als Dateiende- oder Fehlerindikator; der Datensatz wird
    ** ber den Parameter `userrec' zurckgegeben.
    */
    if (!rec) { return em_addr; }
    if (!(em_addr = malloc (eml + 1))) { free (rec); return em_addr; }
    strcpy (em_addr, ema);
    *userrec = rec;
    /* War die Leseoperation erfolgreich, dann wird der Zeiger auf das nchste
    ** Ergebnis von `dbif_select' gesetzt.
    */
    ++dbc->recno;
    return em_addr;
}

/* key = dbif_first (dbc, &userrec)
**     resets the database cursor `dbc' to the first result of the
**     `dbif_select' call which created this cursor and call `dbif_next' in
**     order to return this result.
*/
char *dbif_first (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    dbc->recno = 0;
    return dbif_next (dbc, userrec);
}

/* dbif_end (dbc)
**     release all resources which were allocated for the database cursor
**     `dbc'.
*/
int dbif_end (DBIF_CURSOR dbc)
{
    xClose (dbc->results); xUnlink (dbc->res_name);
    free (dbc);
    return 0;
}

/* Welche Tests bentige ich?
**    TRUE        - alle Datenstze (einer Domain) sollen gelesen werden
**    CC_USER     - Nur Benutzereintrge (einer Domain) sollen gelesen werden
**    CC_LINK     - Nur Verweise (einer Domain) sollen gelesen werden
**    CC_NSUSER   - Nur Nicht-Standardbenutzereintrge (einer Domain) sollen
**                  gelesen werden
**    CC_ADMIN    - Nur Benutzereintrge mit Administrationsberechtigung sollen
**                  gelesen werden
**    CC_DOMAIN   - Nur Domaineintrge sollen gelesen werden
**    CC_DOMLINK  - Nur Verweise auf die angegebene (oder alle) Domain(s)
**                  sollen gelesen werden
**    CC_NSDOMAIN - Nur Nicht-Standarddomaineintrge sollen gelesen werden
**    CC_MAILSYS  - Nur der Mailsystemadministrator soll gelesen werden.
**    CC_IP4TRANS - Nur bersetzungen IPv4 -> Domain sollen gelesen werden
**    CC_IP6TRANS - Nur bersetzungen IPv6 -> Domain sollen gelesen werden
**    CC_REDIRECT - Nur Umleitungen auf andere POP3-Server sollen gelesen
**                  werden
**    CC_IDENTITY - Nur der Datenbankeintrag, der die Datenbank als `VxMail'-
**                  Datenbank identifiziert, soll gelesen werden
**    CC_INVALID  - Nur ungltige Benutzereintrge (einer Domain) sollen
**                  gelesen werden
*/
#if 0
static int ccCheck (DBIF_FILE db,
		    char *key, dbif_rec *rec,
		    char *ikey, dbif_rec *irec,
		    int ccflags)
{
    int res = 0, irtype = dbif_rectype(irec), rtype;
    int iflg = (1 << irtype), flg;
    if (key) { rtype = dbif_rectype(rec); flg = (1 << rtype); }
    /* Wenn der Angegebene Datensatz von vorneherein nicht zu der Bedingung
    ** pat, dann kehre sofort zurck - und zwar mit einem negativen
    ** (Vergleichs-)Ergebnis.
    */
    if (!(ccflags & iflg)) { return 0; }
    /* Ich bentige folgende Spezialbehandlungen:
    ** CC_USER, CC_LINK, CC_NSUSER, CC_ADMIN, CC_INVALID
    **     - Wenn `key' != NULL ist und auf eine Domain verweist, dann mu
    **       zustzlich geprft werden, ob der Teil nach dem '@' in `ikey'
    **       mit `key' bereinstimmt.
    ** CC_LINK
    **     - Wenn `key' != NULL ist und auf einen Benutzer verweist, dann mu
    **       zustzlich geprft werden, ob `key' mit `ikey' bereinstimmt.
    **/
    if (iflg & (CC_USER|CC_LINK|CC_NSUSER|CC_ADMIN|CC_INVALID)) {
	/* Untersuche den Fall, da `key' != NULL ist */
	if (key) {
	    /* Wenn `key' und auf eine Domain verweist, dann mu der Teil nach
	    ** dem `@' in `ikey' mit dem Namen dieser Domain bereinstimmen.
	    */
	    if (flg & (CC_DOMAIN|CC_NSDOMAIN)) {
		char *p = strchr (ikey, '@') + 1;
		return !strcmp (key, p);
	    }
	    /* Wenn `ikey' auf einen LINK auf einen Benutzer verweist, und
	    ** `key' auf einen Benutzer verweist, dann mssen `irec->folder'
	    ** und `key' bereinstimmen.
	    */
	    if ((iflg & CC_LINK)
	    && (flg & (CC_USER|CC_NSUSER|CC_ADMIN|CC_INVALID))) {
		return !strcmp (key, irec->folder);
	    }
	    /* In jedem anderen Fall mssen die beiden Schlssel und die
	    ** Datensatztypen bereinstimmen
	    */
	    return (rtype == irtype) && !strcmp (key, ikey);
	}
	/* Wenn `key' == NULL ist, dann reicht bereits die ganz oben geprfte
	** Bedingung fr ein positives Ergebnis aus!
	*/
	return 1;
    }
    /* CC_DOMLINK, CC_IP4TRANS, CC_IP6TRANS
    **     - Wenn `key' != NULL ist und auf eine Domain verweist, dann
    **       mu zustzlich geprft werden, ob das `folder'-Element mit
    **       `key' bereinstimmt.
    */
    if (iflg & (CC_DOMLINK|CC_IP4TRANS|CC_IP6TRANS)) {
	/* Untersuche den Fall, da `key' != NULL ist. */
	if (key) {
	    /* Wenn `key' auf eine Domain oder eine Umleitung verweist, dann
	    ** mu berprft werden, ob `key' mit `irec->folder' bereinstimmt.
	    */
	    if (flg & (CC_DOMAIN|CC_NSDOMAIN|CC_REDIRECT)) {
		return !strcmp (key, irec->folder);
	    }
	    /* In jedem anderen Fall mssen die beiden Schlssel und die
	    ** Datensatztypen bereinstimmen
	    */
	    return (rtype == irtype) && !strcmp (key, ikey);
	}
	/* Wenn `key' == NULL ist, dann reicht bereits die ganz oben geprfte
	** Bedingung fr ein positives Ergebnis aus!
	*/
	return 1;
    }
    /* Wenn in den brigen Fllen `key' != NULL ist, dann mssen
    ** die Datensatztypen und die beiden Schlssel bereinstimmen.
    */
    if (key) { return (rtype == irtype) && !strcmp (key, ikey); }
    /* In jedem anderen Fall soll das Ergebnis positiv sein */
    return 1;
}
#endif
