/* rec.c
**
** Database Interface for the user/admin databases
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include "ifconf.h"
#include "dbif_priv.h"

/* Erzeuge einen Benutzerdatensatz aus den Daten fr Pawort, Foldername und
** Typ und - im Fall, da ein Datensatz fr eine Domain erzeugt werden soll -
** Zahl der Benutzer innerhalb der Domain und maximale Anzahl an zugelassenen
** Benutzern fr diese Domain.
** Pawort und Foldername werden mit derLnge als zustzlichem Parameter
** angegeben.
*/
dbif_rec *dbif_mkrecl (char *passwd, size_t pwl,
		       char *folder, size_t fll,
		       int rtype,
		       ...)
{
    int nusers = 0, maxusers = 0;
    dbif_rec *res;
    size_t ressz = sizeof (dbif_rec) + pwl + fll + 3;
    va_list ap;
    char *p;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	va_start (ap, rtype),
	nusers = va_arg (ap, int);
	maxusers = va_arg (ap, int);
	ressz += 2 * sizeof (int);
    }
    if (!(res = malloc (ressz))) { return res; }
    p = (char *) res + sizeof (dbif_rec);
    res->passwd = p; memcpy (p, passwd, pwl); p[pwl] = '\0';
    p += pwl + 1;
    res->folder = p; memcpy (p, folder, fll); p[fll] = '\0';
    p += fll + 1;
    res->flags = p;
    *p++ = (char) rtype;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	set_int ((unsigned char *) p, nusers);
	set_int ((unsigned char *) p + sizeof(int), maxusers);
    }
    return res;
}

/* Konvertiere einen Datensatz aus der Datenbank in einen Benutzerdatensatz */
dbif_rec *dbif_mkreca (char *p)
{
    size_t pwl, fll;
    char *passwd, *folder;
    int rtype, nusers = 0, maxusers = 0;
    pwl = get_int((unsigned char *) p); p += sizeof(int);
    fll = get_int((unsigned char *) p); p += sizeof(int);
    passwd = p; p += pwl;
    folder = p; p += fll;
    rtype = (int) *p++;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	nusers = get_int ((unsigned char *) p); p += sizeof(int);
	maxusers = get_int ((unsigned char *) p); p += sizeof(int);
    }
    return dbif_mkrecl (passwd, pwl, folder, fll, rtype, nusers, maxusers);
}

/* Erzeuge einen Benutzerdatensatz aus den Daten fr Pawort, Foldername und
** Flags; die Flags werden als NULL-terminierte Liste von Flagnamen (Zeichen-
** ketten hinter dem Foldernamen bergeben.
*/
dbif_rec *dbif_mkrec (char *passwd, char *folder, int rtype, ...)
{
    va_list ap;
    int nusers, maxusers;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	va_start (ap, rtype);
	nusers = va_arg (ap, int);
	maxusers = va_arg (ap, int);
	va_end (ap);
	return dbif_mkrecl (passwd, strlen (passwd),
			    folder, strlen (folder),
			    rtype,
			    nusers, maxusers);
    }
    return dbif_mkrecl (passwd, strlen (passwd),
			folder, strlen (folder),
			rtype);
}

/* dbif_rectype (userrec)
**     returns the type of the user record `userrec'
*/
int dbif_rectype (dbif_rec *r)
{
    return (int) *(r->flags);
}

/* dbif_setlim (domrec, users, maxusers)
**     sets the limits of a domain record; if the record is not a domain
**     record, it returns a non-zero value.
*/
int dbif_setlim (dbif_rec *r, int nusers, int maxusers)
{
    unsigned char *p = (unsigned char *) r->flags;
    if ((int) *p == RT_DOMAIN || (int) *p == RT_NSDOMAIN) {
	++p;               set_int (p, nusers);
	p += sizeof (int); set_int (p, maxusers);
	return 0;
    }
    return 1;
}

/* dbif_getlim (domrec, &users, &maxusers)
**     reads the limits of a domain record into the var arguments `users' and
**     `maxusers'; if the record is not a domain record, it returns a non-zero
**     value.
*/
int dbif_getlim (dbif_rec *r, int *nusers, int *maxusers)
{
    unsigned char *p = (unsigned char *) r->flags;
    if ((int) *p == RT_DOMAIN || (int) *p == RT_NSDOMAIN) {
	++p;               *nusers = get_int (p);
	p += sizeof (int); *maxusers = get_int (p);
	return 0;
    }
    return 1;
}

/* dbif_free (userrec)
**    releases the memory which was allocated for `userrec'.
*/
void dbif_free (dbif_rec *p)
{
    free (p);
}
