/* dbif.h
**
** Interfacedatei zu `dbif.h'
** (Database Interface for the user database)
*/
#ifndef _DBIF_H
#define _DBIF_H

#include "utypes.h"

#include <db2/db_185.h>

typedef struct {
    char *base;
    char *passwd;
    char *folder;
    char *flags;
    char *end;
} dbif_rec;

typedef struct {
    DB *dbf;
    DBTYPE dbtype;
    int seq_flag;
    int rw;
} DBIF_FILE[1];

#ifndef _DBIF_C

/* ffne eine Datenbank im Nur-Lese-Modus */
void dbif_reset (DBIF_FILE dbf, char *dbname);

/* Erzeuge eine Datenbank neu */
void dbif_create (DBIF_FILE dbf, char *dbname, int mode);

/* ffne eine Datenbank sowohl fr Schreib- als auch fr Lesebetrieb */
void dbif_rw (DBIF_FILE dbf, char *dbname);

/* Initialisiere ein DBIF handle mit leeren Werten */
void dbif_init (DBIF_FILE dbf);

/* Kopiere ein DBIF handle */
void dbif_copy (DBIF_FILE src, DBIF_FILE dst);

/* berprfe, ob eine Datenbank geffnet ist! */
int dbif_isopen (DBIF_FILE dbf);

/* Schliee eine geffnete Datenbank! */
void dbif_close (DBIF_FILE dbf);

/* Synchronize a database */
void dbif_sync (DBIF_FILE dbf);

dbif_rec dbif_resolve (DBIF_FILE db, char *em_addr);

dbif_rec dbif_get (DBIF_FILE db, char *em_addr);

int dbif_vinsert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[]);

int dbif_insert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...);

int dbif_vset (DBIF_FILE db,
	       char *em_addr,
	       char *passwd,
	       char *folder,
	       char *flags[]);

int dbif_set (DBIF_FILE db,
	      char *em_addr,
	      char *passwd,
	      char *folder,
	      ...);

const char *dbif_errormsg (void);

void dbif_seqinit (DBIF_FILE db);

char *dbif_seq (DBIF_FILE db, dbif_rec *rec);

int is_member (char *userflags, char *flagname);

void set_member (char *userflags, char *flagname);

dbif_rec dbif_mkrec (char *passwd, char *folder, ...);

#endif /* defined(_DBIF_C) */

#endif /* defined(_DBIF_H) */
