/* dbif.h
**
** Interfacedatei zu `dbif.h'
** (Database Interface for the user database)
*/
#ifndef _DBIF_H
#define _DBIF_H

/* The content of the database consists of variable length records of
** four elements:
**   - the e-mail address of the user or domain name of special key
**   - the users password or domain's maximum number of users
**   - the pathname of the user's folder / pathname of the directory
**     containing all user folders for this domain
**   - a row of flags, specifying the type of the entry
** A record of this typecontains the latter three entries; the first entry
** is the `primary key' of the database; it is only returned by `dbif_seq',
** so i didn't want it be included here.
*/
typedef struct {
    char *passwd;
    char *folder;
    char *flags;
} dbif_rec;

/* DBIF_FILE is the database handle; the name comes from the three (at the
** moment) implemented types of databases. When using a `SyBase' or
** `PostgreSQL' database, this is really not a file, but a set of tables in
** the corresponding relational database; the programs of the VxMail-system
** should nonetheless NOT recognize nor need the type of database which is
** really used; this is the reason for implementing this module as an abstract
** data type.
*/
struct _DBIF_FILE_S;
typedef struct _DBIF_FILE_S *DBIF_FILE;

struct _DBIF_CURSOR_S;
typedef struct _DBIF_CURSOR_S *DBIF_CURSOR;

#ifndef _DBIF_C

/* Connect to the user database, but restrict the access to read-only
** operations
*/
void dbif_reset (DBIF_FILE *dbf, char *dbname);

/* Setup a completely new user database; when using one of the implemented
** database modules based on DB1, DB2 or GDBM, this means, that the file(s)
** containing the database are created or truncated to file position 0;
** when using a `SyBase' or `PostgreSQL' database, this should consist of
** `DROP TABLE ...'-operations, followed by `CREATE TABLE ...'-operations on
** all tables of the user database.
*/
void dbif_create (DBIF_FILE *dbf, char *dbname, int mode);

/* Connect to the user database without restrictions; this access should
** lock the user database or parts of it; how this operation is implemented
** for `SyBase' or `PostgreSQL' databases is really not interesting for the
** VxMail-system.
*/
void dbif_rw (DBIF_FILE *dbf, char *dbname);

/* End all transactions on the user database and disconnect from it.
** When using a DB1, DB2 or GDBM library-like database, this should consist
** of a flush operation, followed by closing the corresponding database
** file(s); when using a `SyBase' or `PostgreSQL' database, this
** should consist of an `END TRANSACTION ...'-operation, followed by the
** operations which are needed for releasing the connection to the user
** database.
*/
void dbif_close (DBIF_FILE dbf);

/* commit all changes on the user database; a simple flush operation
** for DB1, DB2 or GDBM-like databases; a `COMMIT ...' or `END TRANSACTION ...'
** operation for `SyBase' or `PostgreSQL' databases.
*/
void dbif_sync (DBIF_FILE dbf);

/* follow a link-chain to it's origin and return the corresponding entry
** from the user database.
*/
dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr);

/* get a record from the user database; if there is no record corresponding to
** the key `em_addr', return a NULL-pointer instead.
** The memory allocated for the result is completely dynamically allocated;
** For freeing this memory the function `dbif_free' (see below) MUST be used.
*/
dbif_rec *dbif_get (DBIF_FILE db, char *em_addr);

/* insert a new record into the user database; for simplicity, i don't supply
** a user record; instead, all elements of the database entry are supplied as
** single arguments; the flags are supplied as a NULL-terminated vector of
** strings.
*/
int dbif_vinsert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[]);

/* insert a new record into the user database; for simplicity, i don't supply
** a user record; instead, all elements of the database entry are supplied as
** single arguments; the flags are supplied as a NULL-terminated list of
** strings.
*/
int dbif_insert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...);

/* insert a new record into the user database or update an existing one; for
** simplicity, i don't supply a user record; instead, all elements of the
** database entry are supplied as single arguments; the flags are supplied as
** a NULL-terminated vector of strings.
*/
int dbif_vset (DBIF_FILE db,
	       char *em_addr,
	       char *passwd,
	       char *folder,
	       char *flags[]);

/* insert a new record into the user database or update an existing one; for
** simplicity, i don't supply a user record; instead, all elements of the
** database entry are supplied as single arguments; the flags are supplied as
** a NULL-terminated list of strings.
*/
int dbif_set (DBIF_FILE db,
	      char *em_addr,
	      char *passwd,
	      char *folder,
	      ...);

/* delete a record from the user database; the system entries and the database
** identity can not be deleted!
** A return value of zero indicates a successful operation; a non-zero return
** value indicates an error.
*/
int dbif_del (DBIF_FILE db, char *em_addr);


/* return the error message corresponding to the last database error */
const char *dbif_errormsg (void);

/* Initialize a sequential access to the user database; i needed sequential
** access for implementing the `vxcheck' and `vxdump' programs.
*/
void dbif_seqinit (DBIF_FILE db);

/* Return the next element from the sequential access of the user database;
** the function result is the key corresponding to the user record; the
** user record itself is returned by the last parameter.
** This function retruns a NULL-pointer and does NOT change the contents of
** it's arguments if either an error occured or the last record was already
** read by a previous call.
** The database records don't have to be ordered; but the sequential access is
** the only method for DB1, DB2 or GDBM-like databases for reading any entry
** in the user database; a `SyBase' or `PostgreSQL' based database sould
** implement this function by one or more `SELECT ...' statements to the
** tables of the user database; for the VxMail-system it's only interesting
** that the complete database contents can be read.
*/
char *dbif_seq (DBIF_FILE db, dbif_rec **rec);

/* These function don't work on the database; instead they work on the flag
** element of the user database. `set_member' sets flags in a user record;
** `is_member' returns the flag value corresponding to it's name; in a later
** release, i will replace these functions by a better mechanism for
** identifying the type and usage of user records
*/
int is_member (char *userflags, char *flagname);

void set_member (char *userflags, char *flagname);

/* This function creates a user record from the arguments supplied; later,
** i will replace the user record inserting/updating functions by a single
** function and use this function for creating the corresponding elements.
*/
dbif_rec *dbif_mkrec (char *passwd, char *folder, ...);

/* This function releases the memory allocated for a user record completely;
** all elements and the user record structure itself are released.
** For implementation indepency, only this function should be used for
** releasing the dynamically allocated memory for the user record structure.
*/
void dbif_free (dbif_rec *p);

/* The following function implement a (very simple) select/cursor concept.
** I implemented this functions primarily for the programs `vxcheck' and
** `vxdump' but it seems to be usable in a more complex administration
** program, too.
*/

/* dbif_select (db, key, ccflags)
**     selects all records from the database which match the condition
**     constructed fromm the key `key' and the condition flags `ccflags'.
**     the result is a cursor which allows a sequential access to the
**     resulting records.
*/
DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, int ccflags);

/* key = dbif_first (dbc, &userrec)
**     resets the database cursor `dbc' to the first record of the selection
**     described by this cursor.
*/
char *dbif_first (DBIF_CURSOR dbc, dbif_rec **userrec);

/* key = dbif_next (dbc, &userrec)
**     reads the next record from the selection described by the database
**     cursor `dbc'.
*/
char *dbif_next (DBIF_CURSOR dbc, dbif_rec **userrec);

/* dbif_end (dbc)
**     removes the database cursor and all resources allocated for it.
*/
int dbif_end (DBIF_CURSOR dbc);

#endif /* defined(_DBIF_C) */

#endif /* defined(_DBIF_H) */
