/* hlsock.c
**
** high level operations based on the BSD socket connect/accept mechanism
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "vxerror.h"
#include "tools.h"
#include "hlsock.h"

struct _HLSock {
    int sd;	/* socket file descriptor */
    int IPver;
    char *proto;
    union {
	struct sockaddr_in v4;
    } sa;
    FILE *in, *out;
};

/* Meine Funktionen sollen `HLCreateSock' und `HLConnect' heien, wobei
** `HLCreateSock' fr die Registrierung eines Services steht, whrend mit
** `HLConnect' ein Service angefordert wird.
*/
#define MAX(a,b) ((a) > (b) ? (a) : (b))
HLSock HLCreateSock (char *proto, char *service, int maxconn)
{
    struct servent *sv;
    int sock, type, port;
    HLSock res;
    char *p, *pr;
    size_t len = strlen (proto);
    /* Socket-Typ abhngig vom Protokoll festlegen */
    if (!lccmp (proto, "tcp", MAX(len, 3))) {
	type = SOCK_STREAM;
	pr = "tcp";
    } else if (!lccmp (proto, "tcp", MAX(len, 3))) {
	type = SOCK_DGRAM;
	pr = "udp";
    } else {
	errno = ENOSYS;
	return (HLSock) 0;
    }
    /* Speicherplatz fr HLSock alloziieren */
    len = sizeof (struct _HLSock) + strlen (pr) + 1;
    res = (HLSock) galloc (__FILE__, __LINE__, len);
    /* Alloziierten Speicherplatz leeren */
    memset (res, '\0', len);
    /* Protokolltyp eintragen */
    p = (char *) res + sizeof (struct _HLSock);
    res->proto = p;
    res->in = res->out = (FILE *) 0;
    strcpy (p, pr);
    /* Service aus dem Servicenamen bestimmen */
    if (!(sv = getservbyname (service, proto))) {
	port = htons ((int) strtol (service, &p, 0));
	if (*p) {
	    gfree (res);
	    errno = -1;
	    return (HLSock) 0;
	}
    } else {
	port = sv->s_port;
    }
    if ((sock = socket (PF_INET, type, 0)) < 0) {
	gfree (res);
	return (HLSock) 0;
    }
    res->sa.v4.sin_family = AF_INET;
    res->sa.v4.sin_addr.s_addr = htonl (INADDR_ANY);
    res->sa.v4.sin_port = port;
    res->sd = sock;
    res->IPver = 4;
    if (bind(res->sd, (struct sockaddr *)&res->sa.v4, sizeof(res->sa.v4)) < 0) {
	close (res->sd);
	gfree (res);
	return (HLSock) 0;
    }
    listen (res->sd, maxconn);
    return res;
}

/* Warte auf Service-Anforderungen */
HLSock HLAccept (HLSock serv)
{
    size_t len = sizeof (struct _HLSock) + strlen (serv->proto) + 1;
    HLSock cli = galloc (__FILE__, __LINE__, len);
    char *p = (char *) cli + sizeof (struct _HLSock);
    size_t sal = sizeof (cli->sa);
    memset (cli, '\0', len);
    cli->proto = p;
    strcpy (p, serv->proto);
    cli->IPver = serv->IPver;
    cli->sd = accept(serv->sd, (struct sockaddr *) &cli->sa, SZCONV &sal);
    if (cli->sd < 0) { gfree (cli); return (HLSock) 0; }
    return cli;
}

/* Fordere einen Service an. */
HLSock HLConnect (char *proto, const char *host, char *service)
{
    struct hostent *he;
    struct servent *sv;
    int type, sock, port;
    char *p, *pr;
    HLSock res;
    size_t len = strlen (proto);
    /* Socket-Typ abhngig vom Protokoll festlegen */
    if (!lccmp (proto, "tcp", MAX(len, 3))) {
	type = SOCK_STREAM;
	pr = "tcp";
    } else if (!lccmp (proto, "tcp", MAX(len, 3))) {
	type = SOCK_DGRAM;
	pr = "udp";
    } else {
	errno = ENOSYS;
	return (HLSock) 0;
    }
    /* Speicherplatz fr HLSock alloziieren */
    len = sizeof (struct _HLSock) + strlen (pr) + 1;
    res = (HLSock) galloc (__FILE__, __LINE__, len);
    memset (res, '\0', len);
    /* Protokolltyp eintragen */
    p = (char *) res + sizeof (struct _HLSock);
    res->proto = p; strcpy (p, pr);
    res->in = res->out = (FILE *) 0;
    /* Service aus dem Servicenamen bestimmen */
    if (!(sv = getservbyname (service, proto))) {
	port = htons ((int) strtol (service, &p, 0));
	if (*p) {
	    gfree (res);
	    errno = -1;
	    return (HLSock) 0;
	}
    } else {
	port = sv->s_port;
    }
    /* Socket anlegen */
    if ((sock = socket (PF_INET, type, 0)) < 0) {
	gfree (res);
	return (HLSock) 0;
    }
    /* Bisherige Ergebnisse eintragen */
    res->sd = sock;
    res->sa.v4.sin_family = AF_INET;
    res->sa.v4.sin_port = port;
    res->IPver = 4;
    /* Adresse des Servers bestimmen */
    if (*host == '[') {
	/* In diesem Fall wurde eine (in `[' und `]' eingeschlossene)
	** IP-Adresse bergeben, die bereits den gewnschten Server
	** eindeutig identifiziert.
	*/
	char *IPADDR = gdupl (__FILE__, __LINE__, host + 1);
	IPADDR[strlen (IPADDR) - 1] = '\0';
	/* ... und eintragen */
	if (!inet_aton (IPADDR, &res->sa.v4.sin_addr)) {
	    gfree (IPADDR);
	    close (sock);
	    gfree (res);
	    return (HLSock) 0;
	}
	gfree (IPADDR);
    } else {
	/* In diesem Fall wurde ein Name angegeben, aus dem die IP-Adresse
	** mittels `gethostbyname' bestimmt wird.
	*/
	if (!(he = gethostbyname (host))) {
	    close (sock);
	    gfree (res);
	    return (HLSock) 0;
	}
	/* ... und eintragen */
	memcpy (&res->sa.v4.sin_addr.s_addr, he->h_addr, he->h_length);
    }
    /* Verbindung herstellen */
    if (connect(res->sd, (struct sockaddr *) &res->sa, sizeof (res->sa)) < 0) {
	close (res->sd);
	gfree (res);
	return (HLSock) 0;
    }
    /* Verbindung steht! */
    return res;
}

/* Liefere die IP-Adresse der Verbindung (als Zeichenkette) */
char *HLSockname (HLSock hls)
{
    char ipbuf[128];
    sprintf (ipbuf, "[%s]", inet_ntoa (hls->sa.v4.sin_addr));
    return gdupl (__FILE__, __LINE__, ipbuf);
}

unsigned long HLSockaddr (HLSock hls)
{
    char *ipbuf = HLSockname (hls);
    unsigned long addr;
    ipbuf[strlen(ipbuf) - 1] = '\0';
    *ipbuf = '\0';
    addr = inet_addr (ipbuf + 1);
    gfree (ipbuf);
    return addr;
}

/* Liefere den Socketdeskriptor einer Verbindung */
int HLSockSD (HLSock hls)
{
    return hls->sd;
}

/* Beende eine Verbindung */
int HLClose (HLSock hls)
{
    close (hls->sd);
    if (hls->in) { fclose (hls->in); }
    if (hls->out) { fclose (hls->out); }
    gfree (hls);
    return 0;
}

char *LLSockname (int sd)
{
    struct sockaddr_in sa;
    size_t sal;
    char ipbuf[128];
    if (getsockname (sd, (struct sockaddr *) &sa, SZCONV &sal) < 0) {
	return NULL;
    }
    sprintf (ipbuf, "[%s]", inet_ntoa (sa.sin_addr));
    return gdupl (__FILE__, __LINE__, ipbuf);
}

unsigned long LLSockaddr (int sd)
{
    char *ipbuf = LLSockname (sd);
    unsigned long addr;
    ipbuf[strlen(ipbuf) - 1] = '\0';
    *ipbuf = '\0';
    addr = inet_addr (ipbuf + 1);
    gfree (ipbuf);
    return addr;
}

char *LLPeername (int sd)
{
    struct sockaddr_in sa;
    size_t sal;
    char ipbuf[128];
    if (getpeername (sd, (struct sockaddr *) &sa, SZCONV &sal) < 0) {
	return NULL;
    }
    sprintf (ipbuf, "[%s]", inet_ntoa (sa.sin_addr));
    return gdupl (__FILE__, __LINE__, ipbuf);
}

unsigned long LLPeeraddr (int sd)
{
    char *ipbuf = LLPeername (sd);
    unsigned long addr;
    ipbuf[strlen(ipbuf) - 1] = '\0';
    *ipbuf = '\0';
    addr = inet_addr (ipbuf + 1);
    gfree (ipbuf);
    return addr;
}

int LLPort (char *service, char *proto)
{
    struct servent *sv = getservbyname (service, proto);
    if (!sv) { return -1; }
    return ((int) ntohs ((short) sv->s_port)) & 0xffff;
}

int HLhfInit (HLSock hls)
{
    int is, os;
    FILE *in, *out;
    if (hls->in || hls->out) { return 1; }
    if ((is = dup (hls->sd)) < 0) { return 2; }
    if (!(in = fdopen (is, "r"))) { close (is); return 3; }
    if ((os = dup (hls->sd)) < 0) { fclose (in); return 4; }
    if (!(out = fdopen (os, "w"))) { close (os); fclose (in); return 5; }
    hls->in = in; hls->out = out;
    return 0;
}

int HLPuts (const char *msg, HLSock hls)
{
    if (hls->out) {
	int x = fputs (msg, hls->out);
	if (x >= 0) { fflush (hls->out); }
	return x;
    }
    return -2;
}

char *HLGets (char *buf, size_t bfsz, HLSock hls)
{
    if (hls->in) { return fgets (buf, bfsz, hls->in); }
    return NULL;
}

#if 0
int HLFlush (HLSock hls)
{
    if (hls->out) {
	return fflush (hls->out);
    }
    return 1;
}
#endif
