/* hlsock.h
**
** Interface file for `hlsock.c'
** (high level operations based on the BSD socket connect/accept mechanism)
**
*/
#ifndef _HLSOCK_H
#define _HLSOCK_H

struct _HLSock;
typedef struct _HLSock *HLSock;

/* Create a high-level socket using the supplied arguments for the service
** the calling program intends to provide, the protocol required for this
** service and the maximum number of clients, the calling program is able to
** serve simultaneously.
*/
HLSock HLCreateSock (char *proto, char *service, int maxconn);

/* wait for service requests on a high-level socket */
HLSock HLAccept (HLSock serv);

/* Connect to a host - requesting a protocol and a service */
HLSock HLConnect (char *proto, const char *host, char *service);

/* Get the socket name (IP address) from the high-level socket `hls'
** The result is the ASCIIfied IP address in the form [x.x.x.x] -
** where each `x' represents a decimal number in the range 0..255
*/
char *HLSockname (HLSock hls);

/* Get the socket name (IP address) from the high-level socket `hls'
** and return it directly (without converting it to a string).
*/
unsigned long HLSockaddr (HLSock hls);

/* Get the socket descriptor from a high-level socket */
int HLSockSD (HLSock hls);

/* free all resources allocated for a high-level socket */
int HLClose (HLSock hls);

/* Get the socket name (IP address) from the socket descriptor `sd'
** The result is the ASCIIfied IP address in the form [x.x.x.x] -
** where each `x' represents a decimal number in the range 0..255
*/
char *LLSockname (int sd);

/* Get the socket name (IP address) from the socket descriptor `sd'
** and return it directly (without converting it to a string).
*/
unsigned long LLSockaddr (int sd);

/* Get the peer's (remote) name (IP address) from the socket descriptor `sd'
** The result is the ASCIIfied IP address in the form [x.x.x.x] -
** where each `x' represents a decimal number in the range 0..255
*/
char *LLPeername (int sd);

/* Get the peer's (remote) name (IP address) from the socket descriptor `sd'
** and return it directly (without converting it to a string).
*/
unsigned long LLPeeraddr (int sd);

/* Get the port number corresponding to the supplied `service' and
** `protocol' arguments.
*/
int LLPort (char *service, char *proto);

/* Initialize the high level socket in order to use the functions `HLPuts'
** and `HLGets' (see below).
*/
int HLhfInit (HLSock hls);

/* Send a ASCII-NUL terminated string through a high level socket */
int HLPuts (const char *msg, HLSock hls);

/* Retrieve a (`bfsz'-1 characters from) a line through a high levbel socket */
char *HLGets (char *buf, size_t bfsz, HLSock hls);

#endif /* defined(_HLSOCK_H) */
