/* tools.c
**
** Einige ntzliche Routinen
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>

#include "vxerror.h"
#define _TOOLS_C
#include "tools.h"
#include "prefs.h"

static exception_t erraction = (exception_t) 0; 

exception_t setup_exception (exception_t handle)
{
    exception_t tmp = erraction;
    erraction = handle;
    return tmp;
}

void heap_init (void)
{
}

static void *lookup (size_t sz)
{
    return malloc (sz);
}

char *sdupl (char *string)
{
    char *res = (char *) malloc (strlen (string) + 1);
    if (res) {
	strcpy (res, string);
    }
    return res;
}

char *gdupl (char *file, int line, char *string)
{
    char *res = lookup (strlen (string) + 1);
    if (!res) {
	if (erraction) {
	    erraction ("(gdupl) failed");
	}
	vxerror_at (EX_OSERR, errno, file, line, "(gdupl)");
    }
    strcpy (res, string);
    return res;
}

void *galloc (char *file, int line, size_t size)
{
    void *res = lookup (size);
    if (!res) {
	vxerror_at (EX_OSERR, errno, file, line, "(galloc)");
    }
    return res;
}

void gfree (void *ref)
{
    free (ref);
}

static void *grealloc (void *ref, size_t size)
{
    return realloc (ref, size);
}

void *gresize (char *file, int line, void *ref, size_t size)
{
    void *res = grealloc (ref, size);
    if (!res) {
	vxerror_at (EX_OSERR, errno, file, line, "(gresize)");
    }
    return res;
}

char *vconcat (char *_1st, va_list ap)
{
    size_t llen = strlen (_1st);
    char *res, *p, *q;
    va_list bp = ap;
    while ((p = va_arg (ap, char *))) { llen += strlen (p); }
    res = lookup (llen + 1);
    if (res) {
	p = res;
	strcpy (p, _1st); p += strlen (_1st);
	ap = bp;
	while ((q = va_arg (ap, char *))) { strcpy (p, q); p += strlen (q); }
    }
    return res;
}

/* concat a NULL-terminated list of string to a new string which is the
** result of this function.
*/
char *concat (char *_1st, ...)
{
    va_list ap;
    char *res;
/*    char *res, *op;
**    size_t res_sz = 0, op_sz;
*/

    va_start (ap, _1st);
    res = vconcat (_1st, ap);
    va_end (ap);
/*    op_sz = strlen (_1st);
**    if (!(res = lookup (op_sz + 1))) { return res; }
**    memcpy (res, _1st, op_sz);
**    res_sz += op_sz;
**    while ((op = va_arg (ap, char *))) {
**	char *newres;
**	op_sz = strlen (op);
**	if (!(newres = grealloc (res, res_sz + op_sz + 1))) {
**	    gfree (res);
**	    return newres;
**	}
**	res = newres;
**	memcpy (res + res_sz, op, op_sz);
**	res_sz += op_sz;
**    }
**    res[res_sz] = '\0';
*/
    return res;
}

/* concat a NULL-terminated list of string to a new string which is the
** result of this function; abort the program execution if memory for the
** function result couldn't be successfully allocated.
*/
char *gconc (char *file, int line, ...)
{
    va_list ap;
    char *_1st, *res;
    va_start (ap, line);
    _1st = va_arg (ap, char *);
    res = vconcat (_1st, (char **) ap);
    va_end (ap);
    if (!res) {
	vxerror_at (EX_OSERR, errno, file, line, "(gconc)");
    }
    return res;
}

static int noeol = 0;

void vvxlog (int ll, char *fmt, va_list av)
{
    if (loglevel >= ll) {
	if (noeol) { noeol = 0; } else { fprintf (stderr, "%5d: ", getpid()); }
	if (fmt[strlen (fmt) - 1] != '\n') { noeol = 1; }
	vfprintf (stderr, fmt, av);
	fflush (stderr);
    }
}

void vxlog (int ll, char *fmt, ...)
{
    va_list ap;
    va_start (ap, fmt);
    vvxlog (ll, fmt, ap);
    va_end (ap);
}

void npfxlog (char *fmt, ...)
{
    va_list ap;
    va_start (ap, fmt);
    vvxlog (0, fmt, ap);
    va_end (ap);
}

int lccmp (char *s1, char *s2, size_t n)
{
    while (n-- > 0) {
	if (tolower (*s1++) != tolower (*s2++)) {
	    return (int) tolower(*--s1) - (int) tolower(*--s2);
	}
    }
    return 0;
}

void lcconv(char *bf, int bfl)
{
    if (bfl < 0) {
	while (*bf) { *bf = tolower(*bf); ++bf; }
    } else {
	while (*bf && bfl-- > 0) { *bf = tolower(*bf); ++bf; }
    }
}
