/* vxopt.c
**
** scan the arguments for the VxMail-specific options
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include "tools.h"
#include "prefs.h"
#include "conf.h"

/* Welche Option sind spezifisch fr VxMail?
** -f <configfile> | --config=<configfile>
** -h              | --help
** -V              | --version
*/

static __inline__ void duplconf (void)
{
    fprintf(stderr, "%s: duplicate `--config' option ignored!\n", progname);
}

static __inline__ void argmiss (void)
{
    fprintf(stderr, "%s: missing argument for `--config' option - "
		    "option ignored!\n",
		    progname);
}

static __inline__ int optf_valid (void)
{
#if 0
    if (geteuid() != getuid() && !geteuid()) {
	fprintf(stderr, "%s: `setuid root' mode disables `--config' - "
			"option ignored!\n",
			progname);
	return 0;
    }
#endif
    return 1;
}

int vxopt (int *ac, char **av, void (*usage)(void))
{
    int ix, oc, do_version = 0, do_usage = 0, conf = 0;
    char *confname = NULL;

    /* Set the `progname' variable */
    if ((progname = strrchr (av[0], '/'))) ++progname; else progname = av[0];

    /* Evaluate the VxMail-specific options */
    for (ix = 1; ix < *ac;) {
	if (!strcmp (av[ix], "--") || *av[ix] != '-') break;
	oc = 0;
	if (!strncmp (av[ix], "-f", 2)) {
	    /* short form of configfile option */
	    if (av[ix][2]) {
		/* option argument is concatenated to the option name itself */
		if (confname) {
		    duplconf ();
		} else if (optf_valid()) {
		    confname = gdupl (__FILE__, __LINE__, &av[ix][2]);
		    conf = 1;
		}
		oc = 1;
	    } else {
		/* option argument is an extra element of `av' */
		oc = 2;
		if (confname) {
		    duplconf ();
		    if (ix + 1 >= *ac) { oc = 1; }
		} else if (ix + 1 >= *ac) {
		    argmiss (); --oc;
		} else if (optf_valid()) {
		    confname = gdupl (__FILE__, __LINE__, av[ix + 1]);
		    conf = 1;
		}
	    }
	} else if (!strncmp (av[ix], "--config", 8)) {
	    /* long form of configfile option */
	    if (!av[ix][8]) {
		/* option argument is an extra element of `av' */
		oc = 2;
		if (confname) {
		    duplconf ();
		    if (ix + 1 >= *ac) { oc = 1; }
		} else if (ix + 1 >= *ac)  {
		    argmiss (); --oc;
		} else if (optf_valid()) {
		    confname = gdupl (__FILE__, __LINE__, av[ix + 1]);
		    conf = 1;
		}
	    } else if (av[ix][8] == '=') {
		/* option argument is concatenated to the option name itself */
		if (confname) {
		    duplconf ();
		} else if (optf_valid()) {
		    confname = gdupl (__FILE__, __LINE__, &av[ix][9]);
		    conf = 1;
		}
		oc = 1;
	    } else {
		++ix; continue;
	    }
	} else if (!strcmp (av[ix], "-h") || !strcmp (av[ix], "--help")) {
	    /* `-h' or `--help' */
	    do_usage = 1;
	    oc = 1;
	} else if (!strcmp (av[ix], "-V") || !strcmp (av[ix], "--version")) {
	    /* `-V' or `--version' */
	    do_version = 1;
	    oc = 1;
	} else {
	    /* any other argument is silently ignored */
	    ++ix; continue;
	}
	/* remove the options from the argument list */
	if (oc) {
	    int jx;
	    for (jx = ix + oc; jx < *ac; ++jx) { av[jx - oc] = av[jx]; }
	    *ac -= oc;
	}
    }

    if (do_usage) {
	/* call the usage-function if required */
	usage ();
    } else if (do_version) {
	/* print the VxMail-version and exit if required */
	fprintf (stderr, "%s\n", VXMAIL_VERSION);
	exit (1);
    } else if (confname) {
	/* set the new configuration file's pathname if any was set */
	prefs.config_name = confname;
    }
    return conf;
}
